@@ -40,18 +40,13 @@
 #include <OpenIPMI/ipmiif.h>
 #include <OpenIPMI/os_handler.h>
 
-/* This represents a registration for an event handler. */
-typedef struct ipmi_ll_event_handler_id_s ipmi_ll_event_handler_id_t;
+#ifdef __cplusplus
+extern "C" {
+#endif
 
 /* Called when an IPMI response to a command comes in from the BMC. */
-typedef void (*ipmi_ll_rsp_handler_t)(ipmi_con_t   *ipmi,
-				      ipmi_addr_t  *addr,
-				      unsigned int addr_len,
-				      ipmi_msg_t   *msg,
-				      void         *rsp_data1,
-				      void         *rsp_data2,
-				      void         *rsp_data3,
-				      void         *rsp_data4);
+typedef int (*ipmi_ll_rsp_handler_t)(ipmi_con_t   *ipmi,
+				     ipmi_msgi_t  *rspi);
 
 /* Called when an IPMI event comes in from the BMC.  Note that the
    event may be NULL, meaning that an event came in but did not have
@@ -63,8 +58,7 @@ typedef void (*ipmi_ll_evt_handler_t)(ipmi_con_t   *ipmi,
 				      ipmi_addr_t  *addr,
 				      unsigned int addr_len,
 				      ipmi_event_t *event,
-				      void         *event_data,
-				      void         *data2);
+				      void         *cb_data);
 
 /* Called when an incoming command is received by the IPMI code. */
 typedef void (*ipmi_ll_cmd_handler_t)(ipmi_con_t   *ipmi,
@@ -101,10 +95,26 @@ typedef void (*ipmi_ll_ipmb_addr_cb)(ipmi_con_t   *ipmi,
 				     int          active,
 				     unsigned int hacks,
 				     void         *cb_data);
+
+/* Used to handle knowing when the connection shutdown is complete. */
+typedef void (*ipmi_ll_con_closed_cb)(ipmi_con_t *ipmi, void *cb_data);
+
 /* Set this bit in the hacks if, even though the connection is to a
    device not at 0x20, the first part of a LAN command should always
    use 0x20. */
-#define IPMI_CONN_HACK_20_AS_MAIN_ADDR	0x00000001
+#define IPMI_CONN_HACK_20_AS_MAIN_ADDR		0x00000001
+
+/* Some systems (incorrectly, according to the spec) use only the
+   bottom 4 bits or ROLE(m) for authentication in the RAKP3 message.
+   The spec says to use all 8 bits, but enabling this hack makes
+   OpenIPMI only use the bottom 4 bits. */
+#define IPMI_CONN_HACK_RAKP3_WRONG_ROLEM	0x00000002
+
+/* The spec is vague (perhaps wrong), but the default for RMCP+ seems
+   to be to use K(1) as the integrity key.  That is thus the default
+   of OpenIPMI, but this hack lets you use SIK as it says in one part
+   of the spec. */
+#define IPMI_CONN_HACK_RMCPP_INTEG_SIK		0x00000004
 
 /* The data structure representing a connection.  The low-level handler
    fills this out then calls ipmi_init_con() with the connection. */
@@ -147,12 +157,13 @@ struct ipmi_con_s
        con_changed_handler, so set that up first. */
     int (*start_con)(ipmi_con_t *ipmi);
 
-    /* Set the callback to call when the connection goes down or up.
-       There is only one of these handlers, changing it overwrites it.
-       Setting it to NULL disables it. */
-    void (*set_con_change_handler)(ipmi_con_t             *ipmi,
-				   ipmi_ll_con_changed_cb handler,
-				   void                   *cb_data);
+    /* Add a callback to call when the connection goes down or up. */
+    int (*add_con_change_handler)(ipmi_con_t             *ipmi,
+				  ipmi_ll_con_changed_cb handler,
+				  void                   *cb_data);
+    int (*remove_con_change_handler)(ipmi_con_t             *ipmi,
+				     ipmi_ll_con_changed_cb handler,
+				     void                   *cb_data);
 
     /* If OEM code discovers that an IPMB address has changed, it can
        use this to change it.  The hacks are the same as the ones in
@@ -162,10 +173,13 @@ struct ipmi_con_s
 			  int           active,
 			  unsigned int  hacks);
 
-    /* Set the handler that will be called when the IPMB address changes. */
-    void (*set_ipmb_addr_handler)(ipmi_con_t           *ipmi,
-				  ipmi_ll_ipmb_addr_cb handler,
-				  void                 *cb_data);
+    /* Add a handler that will be called when the IPMB address changes. */
+    int (*add_ipmb_addr_handler)(ipmi_con_t           *ipmi,
+				 ipmi_ll_ipmb_addr_cb handler,
+				 void                 *cb_data);
+    int (*remove_ipmb_addr_handler)(ipmi_con_t           *ipmi,
+				    ipmi_ll_ipmb_addr_cb handler,
+				    void                 *cb_data);
 
     /* This call gets the IPMB address of the connection.  It may be
        NULL if the connection does not support this.  This call may be
@@ -201,22 +215,17 @@ struct ipmi_con_s
 			unsigned int          addr_len,
 			ipmi_msg_t            *msg,
 			ipmi_ll_rsp_handler_t rsp_handler,
-			void                  *rsp_data1,
-			void                  *rsp_data2,
-			void                  *rsp_data3,
-			void                  *rsp_data4);
+			ipmi_msgi_t           *rspi);
 
-    /* Register to receive IPMI events from the interface.  Return a
-       handle that can be used for later deregistration. */
-    int (*register_for_events)(ipmi_con_t                 *ipmi,
-			       ipmi_ll_evt_handler_t      handler,
-			       void                       *event_data,
-			       void                       *data2,
-			       ipmi_ll_event_handler_id_t **id);
+    /* Register to receive IPMI events from the interface. */
+    int (*add_event_handler)(ipmi_con_t                 *ipmi,
+			     ipmi_ll_evt_handler_t      handler,
+			     void                       *cb_data);
 
-    /* Remove an event registration. */
-    int (*deregister_for_events)(ipmi_con_t                 *ipmi,
-				 ipmi_ll_event_handler_id_t *id);
+    /* Remove an event handler. */
+    int (*remove_event_handler)(ipmi_con_t                 *ipmi,
+				ipmi_ll_evt_handler_t      handler,
+				void                       *cb_data);
 
     /* Send a response message.  This is not supported on all
        interfaces, primarily only on system management interfaces.  If
@@ -248,7 +257,6 @@ struct ipmi_con_s
     /* Close an IPMI connection. */
     int (*close_connection)(ipmi_con_t *ipmi);
 
-
     /* This is set by OEM code to handle certain conditions when a
        send message fails.  It is currently only used by the IPMI LAN
        code, if a send messages response is an error, this will be
@@ -256,32 +264,56 @@ struct ipmi_con_s
        code will not do anything with the message. */
     int (*handle_send_rsp_err)(ipmi_con_t *con, ipmi_msg_t *msg);
 
-    /* Name the connection code can use for logging. */
+    /* Name the connection code can use for logging and instance names
+       for statistics.  Must be dynamically allocated with
+       ipmi_mem_alloc().  The connection code will free this.  May be
+       NULL. */
     char *name;
+
+    /* The connection code may put a string here to identify
+       itself. */
+    char *con_type;
+
+    /* The privilege level of the connection */
+    unsigned int priv_level;
+
+    /* Close an IPMI connection and report that it is closed. */
+    int (*close_connection_done)(ipmi_con_t            *ipmi,
+				 ipmi_ll_con_closed_cb handler,
+				 void                  *cb_data);
+
+    /* Hacks reported by OEM code.  This should be set by the lower
+       layer or by the user interface code. */
+    unsigned int  hacks;
+
+    /* The IPMB address as reported by the lower layer. */
+    unsigned char ipmb_addr;
+
+    /* Handle an async event for the connection reported by something
+       else. */
+    void (*handle_async_event)(ipmi_con_t   *con,
+			       ipmi_addr_t  *addr,
+			       unsigned int addr_len,
+			       ipmi_msg_t   *msg);
+
+    /* Used by the connection attribute code.  Don't do anything with
+       this yourself!.  The thing that creates this connection should
+       call ipmi_con_attr_init() when the connection is created and
+       ipmi_con_attr_cleanup() when the connection is destroyed. */
+    void *attr;
+
+    /* Statistics interfaces.  These may be NULL if the user doesn't
+       want statistics.  They pass in the user data field. */
+    int (*register_stat)(void *user_data, char *name,
+			 char *instance,  void **stat);
+    void (*add_stat)(void *user_data, void *stat, int value);
+    void (*finished_with_stat)(void *user_data, void *stat);
 };
 
 #define IPMI_CONN_NAME(c) (c->name ? c->name : "")
 
-/* Different types of low-level handlers must register themselves with
-   the IPMI code.  This is currently used so the IPMI code can validate
-   that a connection is a good working connection. */
-typedef struct ll_ipmi_s
-{
-    /* Validate that the given connection is a good connection. */
-    int (*valid_ipmi)(ipmi_con_t *ipmi);
-
-    /* Stuff below here is used internally by the IPMI code.  Set the
-       values to zero or NULL. */
-    volatile int registered;
-    struct ll_ipmi_s *next;
-} ll_ipmi_t;
-void ipmi_register_ll(ll_ipmi_t *ll);
-
-/* Called by the IPMI code in various places to validate a connection. */
-int __ipmi_validate(ipmi_con_t *ipmi);
-
 /* Initialization code for the initialization the connection code. */
-int _ipmi_conn_init(void);
+int _ipmi_conn_init(os_handler_t *os_hnd);
 void _ipmi_conn_shutdown(void);
 
 
@@ -324,4 +356,82 @@ int ipmi_conn_check_oem_handlers(ipmi_con_t               *conn,
 				 ipmi_conn_oem_check_done done,
 				 void                     *cb_data);
 
+/* Generic message handling */
+void ipmi_handle_rsp_item(ipmi_con_t            *ipmi,
+			  ipmi_msgi_t           *rspi,
+			  ipmi_ll_rsp_handler_t rsp_handler);
+
+void ipmi_handle_rsp_item_copymsg(ipmi_con_t            *ipmi,
+				  ipmi_msgi_t           *rspi,
+				  ipmi_msg_t            *msg,
+				  ipmi_ll_rsp_handler_t rsp_handler);
+
+void ipmi_handle_rsp_item_copyall(ipmi_con_t            *ipmi,
+				  ipmi_msgi_t           *rspi,
+				  ipmi_addr_t           *addr,
+				  unsigned int          addr_len,
+				  ipmi_msg_t            *msg,
+				  ipmi_ll_rsp_handler_t rsp_handler);
+
+/* You should use these for allocating and freeing mesage items.  Note
+   that if you set item->msg.data to a non-NULL value that is not
+   item->data, the system will free it with ipmi_free_msg_item_data().
+   So you should allocate it with ipmi_alloc_msg_item_data9). */
+ipmi_msgi_t *ipmi_alloc_msg_item(void);
+void ipmi_free_msg_item(ipmi_msgi_t *item);
+void *ipmi_alloc_msg_item_data(unsigned int size);
+void ipmi_free_msg_item_data(void *data);
+/* Move the data from the old message item to the new one, NULL-ing
+   out the old item's data.  This will free the new_item's original
+   data if necessary.  This will *not* copy the data items, just the
+   address and message. */
+void ipmi_move_msg_item(ipmi_msgi_t *new_item, ipmi_msgi_t *old_item);
+
+/*
+ * Connection attributes.  These are named items that code may create
+ * to attach a void data item to a connection by name.  It can then
+ * look up the data item by name.  Note that you can call
+ * ipmi_con_register_attribute multiple times.  The first time will
+ * create the item, the rest of the times will return the existing
+ * item.
+ *
+ * When the connection is destroyed, the destroy function will be
+ * called on the attribute so the memory (or anything else) can be
+ * cleaned up.
+ *
+ * This is especially for use by RMCP+ payloads so they may attach
+ * data to the connection they are associated with.
+ */
+typedef struct ipmi_con_attr_s ipmi_con_attr_t;
+
+/* Attr init function.  Return the data item in the data field.  Returns
+   an error value.  Will only be called once for the attribute.  */
+typedef int (*ipmi_con_attr_init_cb)(ipmi_con_t *con, void *cb_data,
+				     void **data);
+
+/* Called when the attribute is destroyed.  Note that this may happen
+   after connection destruction, so the connection may not exist any
+   more. */
+typedef void (*ipmi_con_attr_kill_cb)(void *cb_data, void *data);
+
+int ipmi_con_register_attribute(ipmi_con_t            *con,
+				char                  *name,
+				ipmi_con_attr_init_cb init,
+				ipmi_con_attr_kill_cb destroy,
+				void                  *cb_data,
+				ipmi_con_attr_t       **attr);
+int ipmi_con_find_attribute(ipmi_con_t      *con,
+			    char             *name,
+			    ipmi_con_attr_t **attr);
+void *ipmi_con_attr_get_data(ipmi_con_attr_t *attr);
+/* You must call the put operation of every attribute returned by
+   register or find. */
+void ipmi_con_attr_put(ipmi_con_attr_t *attr);
+int ipmi_con_attr_init(ipmi_con_t *con);
+void ipmi_con_attr_cleanup(ipmi_con_t *con);
+
+#ifdef __cplusplus
+}
+#endif
+
 #endif /* _IPMI_CONN_H */
