@@ -171,6 +171,7 @@ sel_add(emu_data_t *emu, lmc_data_t *mc, char **toks)
     int           rv;
     unsigned char record_type;
     unsigned char data[13];
+    unsigned int  r;
 
     rv = get_uchar(toks, &record_type, "record type", 0);
     if (rv)
@@ -182,9 +183,11 @@ sel_add(emu_data_t *emu, lmc_data_t *mc, char **toks)
 	    return rv;
     }
 
-    rv = ipmi_mc_add_to_sel(mc, record_type, data);
+    rv = ipmi_mc_add_to_sel(mc, record_type, data, &r);
     if (rv)
 	printf("**Unable to add to sel, error 0x%x\n", rv);
+    else
+	printf("Added record %d\n", r);
     return rv;
 }
 
@@ -565,6 +568,34 @@ mc_add(emu_data_t *emu, lmc_data_t *mc, char **toks)
     return rv;
 }
 
+static int
+mc_setchan(emu_data_t *emu, lmc_data_t *mc, char **toks)
+{
+    unsigned char channel;
+    unsigned char medium_type;
+    unsigned char protocol_type;
+    unsigned char session_support;
+    int           rv;
+
+    rv = get_uchar(toks, &channel, "Channel Number", 0);
+    if (rv)
+	return rv;
+    rv = get_uchar(toks, &medium_type, "Medium Type", 0);
+    if (rv)
+	return rv;
+    rv = get_uchar(toks, &protocol_type, "Protocol Type", 0);
+    if (rv)
+	return rv;
+    rv = get_uchar(toks, &session_support, "Session Support", 0);
+    if (rv)
+	return rv;
+    rv = ipmi_emu_set_mc_channel(mc, channel, medium_type, protocol_type,
+				 session_support);
+    if (rv)
+	printf("**Unable to set up channel, error 0x%x\n", rv);
+    return rv;
+}
+
 static int
 mc_delete(emu_data_t *emu, lmc_data_t *mc, char **toks)
 {
@@ -641,6 +672,36 @@ mc_add_fru_data(emu_data_t *emu, lmc_data_t *mc, char **toks)
     return rv;
 }
 
+static int
+mc_dump_fru_data(emu_data_t *emu, lmc_data_t *mc, char **toks)
+{
+    unsigned char *data;
+    unsigned char devid;
+    unsigned int  length;
+    int           i;
+    int           rv;
+
+    rv = get_uchar(toks, &devid, "Device ID", 0);
+    if (rv)
+	return rv;
+
+    rv = ipmi_mc_get_fru_data(mc, devid, &length, &data);
+    if (rv) {
+	printf("**Unable to dump FRU data, error 0x%x\n", rv);
+	goto out;
+    }
+
+    for (i=0; i<length; i++) {
+	if ((i > 0) && ((i % 8) == 0))
+	    printf("\n");
+	printf(" 0x%2.2x", data[i]);
+    }
+    printf("\n");
+
+ out:
+    return rv;
+}
+
 static int
 mc_setbmc(emu_data_t *emu, lmc_data_t *mc, char **toks)
 {
@@ -766,12 +827,14 @@ static struct {
     { "sensor_set_event_support", MC,   sensor_set_event_support },
     { "mc_set_power",   MC,		mc_set_power },
     { "mc_add_fru_data",MC,		mc_add_fru_data },
+    { "mc_dump_fru_data",MC,		mc_dump_fru_data },
     { "mc_set_num_leds",MC,		mc_set_num_leds },
     { "mc_add",		NOMC,		mc_add },
     { "mc_delete",	MC,		mc_delete },
     { "mc_disable",	MC,		mc_disable },
     { "mc_enable",	MC,		mc_enable },
     { "mc_setbmc",      NOMC,		mc_setbmc },
+    { "mc_setchan",	MC,		mc_setchan },
     { "atca_enable",    NOMC,	        atca_enable },
     { "atca_set_site",	NOMC,		atca_set_site },
     { "read_cmds",	NOMC,		read_cmds },
