@@ -35,13 +35,14 @@
 #include <stdio.h>
 
 #include <OpenIPMI/ipmiif.h>
-#include <OpenIPMI/ipmi_domain.h>
-#include <OpenIPMI/ipmi_mc.h>
 #include <OpenIPMI/ipmi_err.h>
-#include <OpenIPMI/ipmi_int.h>
-#include <OpenIPMI/ipmi_control.h>
-#include <OpenIPMI/ilist.h>
-#include <OpenIPMI/opq.h>
+
+#include <OpenIPMI/internal/opq.h>
+#include <OpenIPMI/internal/locked_list.h>
+#include <OpenIPMI/internal/ipmi_int.h>
+#include <OpenIPMI/internal/ipmi_domain.h>
+#include <OpenIPMI/internal/ipmi_mc.h>
+#include <OpenIPMI/internal/ipmi_control.h>
 
 struct ipmi_control_info_s
 {
@@ -53,6 +54,8 @@ struct ipmi_control_info_s
        controls. */
     int                      idx_size;
 
+    ipmi_lock_t *idx_lock;
+
     /* Total number of controls we have in this. */
     unsigned int control_count;
 
@@ -61,9 +64,11 @@ struct ipmi_control_info_s
 };
 
 #define CONTROL_ID_LEN 32
-#define CONTROL_NAME_LEN (IPMI_MAX_DOMAIN_NAME_LEN + CONTROL_ID_LEN + 5)
 struct ipmi_control_s
 {
+    unsigned int usecount;
+
+    ipmi_domain_t *domain;
     ipmi_mc_t *mc;
     unsigned char lun;
     unsigned char num;
@@ -74,8 +79,13 @@ struct ipmi_control_s
 
     int destroyed;
 
+    /* After the control is added, it will not be reported immediately.
+       Instead, it will wait until the usecount goes to zero before
+       being reported.  This marks that the add report is pending */
+    int add_pending;
+
     int type;
-    char *type_str;
+    const char *type_str;
 
     int settable;
     int readable;
@@ -96,11 +106,12 @@ struct ipmi_control_s
 
     /* A list of handlers to call when an event for the control comes
        in. */
-    ilist_t *handler_list;
+    locked_list_t *handler_list;
 
     /* For light types. */
     ipmi_control_light_t *lights;
     unsigned int         colors;
+    int                  has_local_control;
 
     /* For display types. */
     unsigned int columns;
@@ -123,16 +134,54 @@ struct ipmi_control_s
     ipmi_control_destroy_cb destroy_handler;
     void                    *destroy_handler_cb_data;
 
-    /* Name we use for reporting */
-    char name[CONTROL_NAME_LEN];
+    /* Name we use for reporting.  We add a ' ' onto the end, thus
+       the +1. */
+    char name[IPMI_CONTROL_NAME_LEN+1];
 };
 
+static void control_final_destroy(ipmi_control_t *control);
+
 /***********************************************************************
  *
  * Control ID handling.
  *
  **********************************************************************/
 
+/* Must be called with the domain entity lock held. */
+int
+_ipmi_control_get(ipmi_control_t *control)
+{
+    if (control->destroyed)
+	return EINVAL;
+    control->usecount++;
+    return 0;
+}
+
+void
+_ipmi_control_put(ipmi_control_t *control)
+{
+    _ipmi_domain_entity_lock(control->domain);
+    if (control->usecount == 1) {
+	if (control->add_pending) {
+	    control->add_pending = 0;
+	    _ipmi_domain_entity_unlock(control->domain);
+	    _ipmi_entity_call_control_handlers(control->entity,
+					       control, IPMI_ADDED);
+	    _ipmi_domain_entity_lock(control->domain);
+	}
+	if (control->destroyed
+	    && (!control->waitq
+		|| (!opq_stuff_in_progress(control->waitq))))
+	{
+	    _ipmi_domain_entity_unlock(control->domain);
+	    control_final_destroy(control);
+	    return;
+	}
+    }
+    control->usecount--;
+    _ipmi_domain_entity_unlock(control->domain);
+}
+
 ipmi_control_id_t
 ipmi_control_convert_to_id(ipmi_control_t *control)
 {
@@ -161,19 +210,48 @@ mc_cb(ipmi_mc_t *mc, void *cb_data)
     mc_cb_info_t        *info = cb_data;
     ipmi_control_info_t *controls;
     ipmi_domain_t       *domain = ipmi_mc_get_domain(mc);
+    ipmi_control_t      *control;
+    ipmi_entity_t       *entity = NULL;
     
-    ipmi_domain_entity_lock(domain);
     controls = _ipmi_mc_get_controls(mc);
-    if (info->id.lun != 4)
+    _ipmi_domain_entity_lock(domain);
+    if (info->id.lun > 4) {
 	info->err = EINVAL;
-    else if (info->id.control_num >= controls->idx_size)
+	goto out_unlock;
+    }
+
+    if (info->id.control_num >= controls->idx_size) {
 	info->err = EINVAL;
-    else if (controls->controls_by_idx[info->id.control_num] == NULL)
+	goto out_unlock;
+    }
+
+    control = controls->controls_by_idx[info->id.control_num];
+    if (!control) {
 	info->err = EINVAL;
-    else
-	info->handler(controls->controls_by_idx[info->id.control_num],
-		      info->cb_data);
-    ipmi_domain_entity_unlock(domain);
+	goto out_unlock;
+    }
+
+    info->err = _ipmi_entity_get(control->entity);
+    if (info->err)
+	goto out_unlock;
+    entity = control->entity;
+
+    info->err = _ipmi_control_get(control);
+    if (info->err)
+	goto out_unlock;
+
+    _ipmi_domain_entity_unlock(domain);
+
+    info->handler(control, info->cb_data);
+
+    _ipmi_control_put(control);
+    _ipmi_entity_put(entity);
+    return;
+
+ out_unlock:
+    _ipmi_domain_entity_unlock(domain);
+    if (entity)
+	_ipmi_entity_put(entity);
 }
 
 int
@@ -292,6 +370,11 @@ ipmi_control_find_id(ipmi_domain_id_t domain_id,
 static void
 control_final_destroy(ipmi_control_t *control)
 {
+    _ipmi_entity_get(control->entity);
+    _ipmi_entity_call_control_handlers(control->entity, control, IPMI_DELETED);
+
+    control->mc = NULL;
+
     if (control->destroy_handler)
 	control->destroy_handler(control,
 				 control->destroy_handler_cb_data);
@@ -300,10 +383,15 @@ control_final_destroy(ipmi_control_t *control)
 	control->oem_info_cleanup_handler(control, control->oem_info);
 
     if (control->handler_list)
-	ilist_twoitem_destroy(control->handler_list);
+	locked_list_destroy(control->handler_list);
 
     if (control->waitq)
 	opq_destroy(control->waitq);
+
+    if (control->entity)
+	ipmi_entity_remove_control(control->entity, control);
+
+    _ipmi_entity_put(control->entity);
     ipmi_mem_free(control);
 }
 
@@ -312,20 +400,21 @@ ipmi_control_destroy(ipmi_control_t *control)
 {
     ipmi_control_info_t *controls = _ipmi_mc_get_controls(control->mc);
 
-    if (controls->controls_by_idx[control->num] != control)
+    ipmi_lock(controls->idx_lock);
+    if (controls->controls_by_idx[control->num] != control) {
+	ipmi_unlock(controls->idx_lock);
 	return EINVAL;
+    }
 
-    if (control->entity)
-	ipmi_entity_remove_control(control->entity, control);
+    _ipmi_control_get(control);
 
     controls->control_count--;
     controls->controls_by_idx[control->num] = NULL;
 
-    control->mc = NULL;
+    ipmi_unlock(controls->idx_lock);
 
     control->destroyed = 1;
-    if (!opq_stuff_in_progress(control->waitq))
-	control_final_destroy(control);
+    _ipmi_control_put(control);
 
     return 0;
 }
@@ -334,26 +423,13 @@ static void
 control_set_name(ipmi_control_t *control)
 {
     int length;
-    int left;
 
-    control->name[0] = '(';
-    length = 1;
-    left = CONTROL_NAME_LEN - length;
-    if (control->entity) {
-	length += snprintf(control->name+length, left-3, "%s.",
-			   _ipmi_entity_name(control->entity));
-	left = CONTROL_NAME_LEN - length;
-    }
-
-    if (control->id_len > (left - 3)) {
-	memcpy(control->name+length, control->id, left-3);
-	length += left - 3;
-    } else {
-	memcpy(control->name+length, control->id, control->id_len);
-	length += control->id_len;
-    }
-    control->name[length] = ')';
+    length = ipmi_entity_get_name(control->entity, control->name,
+				  sizeof(control->name)-2);
+    control->name[length] = '.';
     length++;
+    length += snprintf(control->name+length, IPMI_CONTROL_NAME_LEN-length-2,
+		       "%s", control->id);
     control->name[length] = ' ';
     length++;
     control->name[length] = '\0';
@@ -366,6 +442,19 @@ _ipmi_control_name(ipmi_control_t *control)
     return control->name;
 }
 
+int
+ipmi_control_get_name(ipmi_control_t *control, char *name, int length)
+{
+    int rv = 0;
+
+    if (control->entity)
+	rv = ipmi_entity_get_name(control->entity, name, length);
+    if (length > control->id_len + 2)
+	length = control->id_len + 2; /* Leave space for the nil */
+    rv += snprintf(name+rv, length, ".%s", control->id);
+    return rv;
+}
+
 /***********************************************************************
  *
  * Control message handling.
@@ -408,6 +497,9 @@ ipmi_control_add_opq(ipmi_control_t         *control,
 		     ipmi_control_op_info_t *info,
 		     void                   *cb_data)
 {
+    if (control->destroyed)
+	return EINVAL;
+
     info->__control = control;
     info->__control_id = ipmi_control_convert_to_id(control);
     info->__cb_data = cb_data;
@@ -425,7 +517,18 @@ ipmi_control_opq_done(ipmi_control_t *control)
     if (!control)
 	return;
 
-    CHECK_CONTROL_LOCK(control);
+     /* This gets called on ECANCELLED error cases, if the sensor is
+	already we need to clear out the opq. */
+    if (control->destroyed) {
+	if (control->waitq) {
+	    opq_destroy(control->waitq);
+	    control->waitq = NULL;
+	}
+ 	return;
+    }
+
+    /* No check for the lock.  It will sometimes fail at destruction
+       time. */
 
     opq_op_done(control->waitq);
 }
@@ -449,13 +552,26 @@ control_rsp_handler(ipmi_mc_t  *mc,
     ipmi_control_t         *control = info->__control;
 
     if (control->destroyed) {
+	ipmi_entity_t *entity = NULL;
+
+	_ipmi_domain_entity_lock(control->domain);
+	control->usecount++;
+	_ipmi_domain_entity_unlock(control->domain);
+
+	rv = _ipmi_entity_get(control->entity);
+	if (! rv)
+	    entity = control->entity;
+
 	ipmi_log(IPMI_LOG_ERR_INFO,
 		 "%scontrol.c(control_rsp_handler): "
 		 "Control was destroyed while an operation was in progress",
 		 CONTROL_NAME(control));
 	if (info->__rsp_handler)
 	    info->__rsp_handler(control, ECANCELED, NULL, info->__cb_data);
-	control_final_destroy(control);
+
+	_ipmi_control_put(control);
+	if (entity)
+	    _ipmi_entity_put(entity);
 	return;
     }
 
@@ -479,7 +595,7 @@ control_rsp_handler(ipmi_mc_t  *mc,
 		 "Could not convert control id to a pointer",
 		 MC_NAME(mc));
 	if (info->__rsp_handler)
-	    info->__rsp_handler(NULL, rv, NULL, info->__cb_data);
+	    info->__rsp_handler(control, rv, NULL, info->__cb_data);
     }
 }
 			 
@@ -497,6 +613,9 @@ ipmi_control_send_command(ipmi_control_t         *control,
     CHECK_MC_LOCK(mc);
     CHECK_CONTROL_LOCK(control);
 
+    if (control->destroyed)
+	return EINVAL;
+
     info->__control = control;
     info->__control_id = ipmi_control_convert_to_id(control);
     info->__cb_data = cb_data;
@@ -505,15 +624,11 @@ ipmi_control_send_command(ipmi_control_t         *control,
     return rv;
 }
 
-static void
-control_addr_response_handler(ipmi_domain_t *domain,
-			      ipmi_addr_t   *addr,
-			      unsigned int  addr_len,
-			      ipmi_msg_t    *msg,
-			      void          *rsp_data1,
-			      void          *rsp_data2)
+static int
+control_addr_response_handler(ipmi_domain_t *domain, ipmi_msgi_t *rspi)
 {
-    ipmi_control_op_info_t *info = rsp_data1;
+    ipmi_msg_t             *msg = &rspi->msg;
+    ipmi_control_op_info_t *info = rspi->data1;
     int                    rv;
     ipmi_control_t         *control = info->__control;
 
@@ -524,8 +639,12 @@ control_addr_response_handler(ipmi_domain_t *domain,
 		 DOMAIN_NAME(domain));
 	if (info->__rsp_handler)
 	    info->__rsp_handler(control, ECANCELED, NULL, info->__cb_data);
-	control_final_destroy(control);
-	return;
+
+	_ipmi_domain_entity_lock(control->domain);
+	control->usecount++;
+	_ipmi_domain_entity_unlock(control->domain);
+	_ipmi_control_put(control);
+	return IPMI_MSG_ITEM_NOT_USED;
     }
 
     /* Call the next stage with the lock held. */
@@ -541,6 +660,7 @@ control_addr_response_handler(ipmi_domain_t *domain,
 	if (info->__rsp_handler)
 	    info->__rsp_handler(control, rv, NULL, info->__cb_data);
     }
+    return IPMI_MSG_ITEM_NOT_USED;
 }
 
 int
@@ -555,8 +675,8 @@ ipmi_control_send_command_addr(ipmi_domain_t          *domain,
 {
     int rv;
 
-    CHECK_MC_LOCK(bmc);
     CHECK_CONTROL_LOCK(control);
+    CHECK_MC_LOCK(control->mc);
 
     info->__control = control;
     info->__control_id = ipmi_control_convert_to_id(control);
@@ -573,6 +693,7 @@ ipmi_controls_alloc(ipmi_mc_t *mc, ipmi_control_info_t **new_controls)
     ipmi_control_info_t *controls;
     ipmi_domain_t       *domain;
     os_handler_t        *os_hnd;
+    int                 rv;
 
     CHECK_MC_LOCK(mc);
 
@@ -590,6 +711,13 @@ ipmi_controls_alloc(ipmi_mc_t *mc, ipmi_control_info_t **new_controls)
 	return ENOMEM;
     }
 
+    rv = ipmi_create_lock_os_hnd(os_hnd, &controls->idx_lock);
+    if (rv) {
+	opq_destroy(controls->control_wait_q);
+	ipmi_mem_free(controls);
+	return rv;
+    }
+
     *new_controls = controls;
     return 0;
 }
@@ -610,9 +738,8 @@ ipmi_control_alloc_nonstandard(ipmi_control_t **new_control)
 	return ENOMEM;
 
     memset(control, 0, sizeof(*control));
-    /* By default we ignore if no entity. */
-    ipmi_control_set_ignore_if_no_entity(control, 1);
 
+    control->usecount = 1;
     *new_control = control;
     return 0;
 }
@@ -629,7 +756,7 @@ ipmi_control_add_nonstandard(ipmi_mc_t               *mc,
     ipmi_domain_t       *domain;
     os_handler_t        *os_hnd;
     ipmi_control_info_t *controls = _ipmi_mc_get_controls(mc);
-    void                *link;
+    locked_list_entry_t *link;
 
     CHECK_MC_LOCK(mc);
     CHECK_ENTITY_LOCK(ent);
@@ -640,6 +767,9 @@ ipmi_control_add_nonstandard(ipmi_mc_t               *mc,
     if (num >= 256)
 	return EINVAL;
 
+    _ipmi_domain_entity_lock(domain);
+    ipmi_lock(controls->idx_lock);
+
     if (num >= controls->idx_size) {
 	ipmi_control_t **new_array;
 	unsigned int   new_size;
@@ -649,8 +779,10 @@ ipmi_control_add_nonstandard(ipmi_mc_t               *mc,
 	   too much). */
 	new_size = ((num / 16) * 16) + 16;
 	new_array = ipmi_mem_alloc(sizeof(*new_array) * new_size);
-	if (!new_array)
+	if (!new_array) {
+	    ipmi_unlock(controls->idx_lock);
 	    return ENOMEM;
+	}
 	if (controls->controls_by_idx)
 	    memcpy(new_array, controls->controls_by_idx,
 		   sizeof(*new_array) * (controls->idx_size));
@@ -663,36 +795,56 @@ ipmi_control_add_nonstandard(ipmi_mc_t               *mc,
     }
 
     control->waitq = opq_alloc(os_hnd);
-    if (! control->waitq)
-	return ENOMEM;
-
-    control->handler_list = alloc_ilist();
-    if (! control->handler_list) {
-	opq_destroy(control->waitq);
+    if (! control->waitq) {
+	ipmi_unlock(controls->idx_lock);
 	return ENOMEM;
     }
 
-    link = ipmi_entity_alloc_control_link();
+    control->handler_list = locked_list_alloc(os_hnd);
+    if (! control->handler_list) {
+	opq_destroy(control->waitq);
+	ipmi_unlock(controls->idx_lock);
+	return ENOMEM;
+    }
+
+    link = locked_list_alloc_entry();
     if (!link) {
 	opq_destroy(control->waitq);
 	control->waitq = NULL;
+	locked_list_destroy(control->handler_list);
+	control->handler_list = NULL;
+	ipmi_unlock(controls->idx_lock);
 	return ENOMEM;
     }
 
+    control->domain = domain;
     control->mc = mc;
     control->source_mc = source_mc;
     control->entity = ent;
     control->lun = 4;
     control->num = num;
-    if (! controls->controls_by_idx[num])
+    if (controls->controls_by_idx[num]) {
+	ipmi_log(IPMI_LOG_WARNING,
+		 "%scontrol.c(ipmi_control_add_nonstandard): "
+		 " Add a control at index %d, but there was already a"
+		 " control there, overwriting the old control",
+		 MC_NAME(mc), num);
+    } else {
 	controls->control_count++;
+    }
     controls->controls_by_idx[num] = control;
     control->destroy_handler = destroy_handler;
     control->destroy_handler_cb_data = destroy_handler_cb_data;
     control_set_name(control);
 
+    ipmi_unlock(controls->idx_lock);
+
+    _ipmi_domain_entity_unlock(domain);
+
     ipmi_entity_add_control(ent, control, link);
 
+    control->add_pending = 1;
+
     return 0;
 }
 
@@ -715,6 +867,8 @@ ipmi_controls_destroy(ipmi_control_info_t *controls)
 
     if (controls->control_wait_q)
 	opq_destroy(controls->control_wait_q);
+    if (controls->idx_lock)
+	ipmi_destroy_lock(controls->idx_lock);
     ipmi_mem_free(controls);
     return 0;
 }
@@ -1018,22 +1172,23 @@ typedef struct control_event_info_s
     ipmi_event_t   *event;
 } control_event_info_t;
 
-static void
-control_val_event_call_handler(void *data, void *ihandler, void *cb_data)
+static int
+control_val_event_call_handler(void *cb_data, void *item1, void *item2)
 {
-    ipmi_control_val_event_cb handler = ihandler;
-    control_event_info_t      *info = data;
+    ipmi_control_val_event_cb handler = item1;
+    control_event_info_t      *info = cb_data;
     int                       handled;
 
     handled = handler(info->control,
 		      info->valid_vals,
 		      info->vals,
-		      cb_data,
+		      item2,
 		      info->event);
     if (handled == IPMI_EVENT_HANDLED) {
 	info->handled = handled;
 	info->event = NULL;
     }
+    return LOCKED_LIST_ITER_CONTINUE;
 }
 
 void
@@ -1051,8 +1206,8 @@ ipmi_control_call_val_event_handlers(ipmi_control_t *control,
     info.event = *event;
     info.handled = IPMI_EVENT_NOT_HANDLED;
 
-    ilist_iter_twoitem(control->handler_list,
-		       control_val_event_call_handler, &info);
+    locked_list_iterate(control->handler_list,
+			control_val_event_call_handler, &info);
 
     if (handled)
 	*handled = info.handled;
@@ -1066,10 +1221,7 @@ ipmi_control_add_val_event_handler(ipmi_control_t            *control,
 {
     CHECK_CONTROL_LOCK(control);
 
-    if (ilist_twoitem_exists(control->handler_list, handler, cb_data))
-	return EADDRINUSE;
-
-    if (! ilist_add_twoitem(control->handler_list, handler, cb_data))
+    if (! locked_list_add(control->handler_list, handler, cb_data))
 	return ENOMEM;
 
     return 0;
@@ -1081,7 +1233,7 @@ int ipmi_control_remove_val_event_handler(ipmi_control_t            *control,
 {
     CHECK_CONTROL_LOCK(control);
 
-    if (! ilist_remove_twoitem(control->handler_list, handler, cb_data))
+    if (! locked_list_remove(control->handler_list, handler, cb_data))
 	return ENOENT;
 
     return 0;
@@ -1108,7 +1260,7 @@ ipmi_control_set_type(ipmi_control_t *control, int val)
     control->type_str = ipmi_get_control_type_string(val);
 }
 
-char *
+const char *
 ipmi_control_get_type_string(ipmi_control_t *control)
 {
     CHECK_CONTROL_LOCK(control);
@@ -1121,7 +1273,18 @@ ipmi_control_get_id_length(ipmi_control_t *control)
 {
     CHECK_CONTROL_LOCK(control);
 
-    return control->id_len;
+    if (control->id_type == IPMI_ASCII_STR)
+	return control->id_len+1;
+    else
+	return control->id_len;
+}
+
+enum ipmi_str_type_e
+ipmi_control_get_id_type(ipmi_control_t *control)
+{
+    CHECK_CONTROL_LOCK(control);
+
+    return control->id_type;
 }
 
 int
@@ -1158,7 +1321,8 @@ ipmi_control_set_id(ipmi_control_t *control, char *id,
     memcpy(control->id, id, length);
     control->id_type = type;
     control->id_len = length;
-    control_set_name(control);
+    if (control->entity)
+	control_set_name(control);
 }
 
 int
@@ -1192,8 +1356,6 @@ ipmi_control_set_readable(ipmi_control_t *control, int val)
 int
 ipmi_control_get_entity_id(ipmi_control_t *control)
 {
-    CHECK_CONTROL_LOCK(control);
-
     return ipmi_entity_get_entity_id(control->entity);
 }
 
@@ -1274,8 +1436,6 @@ ipmi_control_set_callbacks(ipmi_control_t *control, ipmi_control_cbs_t *cbs)
 ipmi_mc_t *
 ipmi_control_get_mc(ipmi_control_t *control)
 {
-    CHECK_CONTROL_LOCK(control);
-
     return control->mc;
 }
 
@@ -1326,15 +1486,17 @@ ipmi_control_get_num_vals(ipmi_control_t *control)
 }
 
 int
-ipmi_control_get_num_light_settings(ipmi_control_t *control,
-				    unsigned int   light)
+ipmi_control_get_num_light_values(ipmi_control_t *control,
+				  unsigned int   light)
 {
     CHECK_CONTROL_LOCK(control);
 
+    if (!control->lights)
+	return -1;
     if (light >= control->num_vals)
 	return -1;
 
-    return control->lights[light].num_settings;
+    return control->lights[light].num_values;
 }
 
 int
@@ -1344,12 +1506,14 @@ ipmi_control_get_num_light_transitions(ipmi_control_t   *control,
 {
     CHECK_CONTROL_LOCK(control);
 
+    if (!control->lights)
+	return -1;
     if (light >= control->num_vals)
 	return -1;
-    if (set >= control->lights[light].num_settings)
+    if (set >= control->lights[light].num_values)
 	return -1;
 
-    return control->lights[light].settings[set].num_transitions;
+    return control->lights[light].values[set].num_transitions;
 }
 
 int
@@ -1360,14 +1524,16 @@ ipmi_control_get_light_color(ipmi_control_t   *control,
 {
     CHECK_CONTROL_LOCK(control);
 
+    if (!control->lights)
+	return -1;
     if (light >= control->num_vals)
 	return -1;
-    if (set >= control->lights[light].num_settings)
+    if (set >= control->lights[light].num_values)
 	return -1;
-    if (num > control->lights[light].settings[set].num_transitions)
+    if (num > control->lights[light].values[set].num_transitions)
 	return -1;
 
-    return control->lights[light].settings[set].transitions[num].color;
+    return control->lights[light].values[set].transitions[num].color;
 }
 
 int
@@ -1378,14 +1544,16 @@ ipmi_control_get_light_color_time(ipmi_control_t   *control,
 {
     CHECK_CONTROL_LOCK(control);
 
+    if (!control->lights)
+	return -1;
     if (light >= control->num_vals)
 	return -1;
-    if (set >= control->lights[light].num_settings)
+    if (set >= control->lights[light].num_values)
 	return -1;
-    if (num > control->lights[light].settings[set].num_transitions)
+    if (num > control->lights[light].values[set].num_transitions)
 	return -1;
 
-    return control->lights[light].settings[set].transitions[num].time;
+    return control->lights[light].values[set].transitions[num].time;
 }
 
 int
@@ -1591,6 +1759,20 @@ ipmi_control_light_is_color_supported(ipmi_control_t *control,
     return (control->colors & (1 << color)) != 0;
 }
 
+void
+ipmi_control_light_set_has_local_control(ipmi_control_t *control, int val)
+{
+     control->has_local_control = val;
+}
+
+int
+ipmi_control_light_has_local_control(ipmi_control_t *control)
+{
+    CHECK_CONTROL_LOCK(control);
+
+    return control->has_local_control;
+}
+
 int
 ipmi_cmp_control_id(ipmi_control_id_t id1, ipmi_control_id_t id2)
 {
@@ -1688,16 +1870,42 @@ ipmi_control_set_ignore_if_no_entity(ipmi_control_t *control,
     control->ignore_if_no_entity = ignore_if_no_entity;
 }
 
+ipmi_domain_t *
+ipmi_control_get_domain(ipmi_control_t *control)
+{
+    return control->domain;
+}
+
 #ifdef IPMI_CHECK_LOCKS
 void
 __ipmi_check_control_lock(ipmi_control_t *control)
 {
-    ipmi_domain_t *domain;
-    if (!control->mc)
-	return; /* The control has really been destroyed, just go on. */
-    domain = ipmi_mc_get_domain(control->mc);
-    __ipmi_check_domain_lock(domain);
-    __ipmi_check_domain_entity_lock(domain);
+    if (!control)
+	return;
+
+    if (!DEBUG_LOCKS)
+	return;
+
+    CHECK_ENTITY_LOCK(control->entity);
+    CHECK_MC_LOCK(control->mc);
+
+    if (control->usecount == 0)
+	ipmi_report_lock_error(ipmi_domain_get_os_hnd(control->domain),
+			       "control not locked when it should have been");
 }
 #endif
 
+/***********************************************************************
+ *
+ * Crufty backwards-compatible interfaces.  Don't use these as they
+ * are deprecated.
+ *
+ **********************************************************************/
+
+int
+ipmi_control_get_num_light_settings(ipmi_control_t *control,
+				    unsigned int   light)
+{
+    return ipmi_control_get_num_light_values(control, light);
+}
+
