@@ -37,17 +37,17 @@
 #include <netdb.h>
 
 #include <OpenIPMI/os_handler.h>
-#include <OpenIPMI/ipmi_domain.h>
-#include <OpenIPMI/ipmi_mc.h>
-#include <OpenIPMI/ipmi_int.h>
 #include <OpenIPMI/ipmi_conn.h>
 #include <OpenIPMI/ipmi_err.h>
-#include <OpenIPMI/ipmi_oem.h>
 #include <OpenIPMI/ipmi_auth.h>
 #include <OpenIPMI/ipmi_lan.h>
 #include <OpenIPMI/ipmi_smi.h>
 #include <OpenIPMI/ipmi_msgbits.h>
-#include <OpenIPMI/mxp.h>
+
+#include <OpenIPMI/internal/ipmi_domain.h>
+#include <OpenIPMI/internal/ipmi_mc.h>
+#include <OpenIPMI/internal/ipmi_int.h>
+#include <OpenIPMI/internal/ipmi_oem.h>
 
 #if defined(DEBUG_MSG) || defined(DEBUG_RAWMSG)
 static void
@@ -63,35 +63,10 @@ dump_hex(void *vdata, int len)
     }
 }
 #endif
-static os_hnd_rwlock_t *global_lock;
 static os_handler_t *ipmi_os_handler;
 
 unsigned int __ipmi_log_mask = 0;
 
-void ipmi_read_lock(void)
-{
-    if (global_lock)
-	ipmi_os_handler->read_lock(ipmi_os_handler, global_lock);
-}
-
-void ipmi_read_unlock(void)
-{
-    if (global_lock)
-	ipmi_os_handler->read_unlock(ipmi_os_handler, global_lock);
-}
-
-void ipmi_write_lock(void)
-{
-    if (global_lock)
-	ipmi_os_handler->write_lock(ipmi_os_handler, global_lock);
-}
-
-void ipmi_write_unlock(void)
-{
-    if (global_lock)
-	ipmi_os_handler->write_unlock(ipmi_os_handler, global_lock);
-}
-
 os_handler_t *
 ipmi_get_global_os_handler(void)
 {
@@ -110,18 +85,6 @@ ipmi_create_lock(ipmi_domain_t *domain, ipmi_lock_t **new_lock)
     return ipmi_create_lock_os_hnd(ipmi_domain_get_os_hnd(domain), new_lock);
 }
 
-int
-ipmi_create_global_rwlock(ipmi_rwlock_t **new_lock)
-{
-    return ipmi_create_rwlock_os_hnd(ipmi_os_handler, new_lock);
-}
-
-int
-ipmi_create_rwlock(ipmi_domain_t *domain, ipmi_rwlock_t **new_lock)
-{
-    return ipmi_create_rwlock_os_hnd(ipmi_domain_get_os_hnd(domain), new_lock);
-}
-
 void
 ipmi_log(enum ipmi_log_type_e log_type, char *format, ...)
 {
@@ -135,55 +98,22 @@ ipmi_log(enum ipmi_log_type_e log_type, char *format, ...)
     va_end(ap);
 }
 
-static ll_ipmi_t *ipmi_ll = NULL;
-
-void
-ipmi_register_ll(ll_ipmi_t *ll)
-{
-    if (ll->registered)
-	return;
-
-    ipmi_write_lock();
-    if (! ll->registered) {
-	ll->registered = 1;
-	ll->next = ipmi_ll;
-	ipmi_ll = ll;
-    }
-    ipmi_write_unlock();
-}
-
-/* Must be called with the read or write lock held. */
-int
-__ipmi_validate(ipmi_con_t *ipmi)
-{
-    ll_ipmi_t *elem;
-
-    elem = ipmi_ll;
-    while (elem) {
-	if (elem->valid_ipmi(ipmi)) {
-	    return 0;
-	}
-	elem = elem->next;
-    }
-
-    return EINVAL;
-}
-
 static unsigned int
 ipmi_get_unicode(int len,
-		 unsigned char *d, int in_len,
+		 unsigned char **d, int in_len,
 		 char *out, int out_len)
 {
     if (out_len < len)
 	len = out_len;
 
-    memcpy(out, d, len);
+    memcpy(out, *d, len);
+    *d += len;
     return len;
 }
 
 static unsigned int
 ipmi_get_bcd_plus(int len,
-		  unsigned char *d, int in_len,
+		  unsigned char **d, int in_len,
 		  char *out, int out_len)
 {
     static char table[16] = {
@@ -197,7 +127,7 @@ ipmi_get_bcd_plus(int len,
     int real_length;
     char *out_s = out;
 
-    real_length = (in_len * 8) / 6;
+    real_length = (in_len * 8) / 4;
     if (len > real_length)
 	len = real_length;
     
@@ -208,25 +138,29 @@ ipmi_get_bcd_plus(int len,
     pos = 0;
     for (i=0; i<len; i++) {
 	switch (bo) {
-	    case 0:
-		val = *d & 0xf;
-		bo = 4;
-		break;
-	    case 4:
-		val = (*d >> 4) & 0xf;
-		d++;
-		bo = 0;
-		break;
+	case 0:
+	    val = **d & 0xf;
+	    bo = 4;
+	    break;
+	case 4:
+	    val = (**d >> 4) & 0xf;
+	    (*d)++;
+	    bo = 0;
+	    break;
 	}
 	*out = table[val];
 	out++;
     }
+
+    if (bo != 0)
+	(*d)++;
+
     return out - out_s;
 }
 
 static unsigned int
 ipmi_get_6_bit_ascii(int len,
-		     unsigned char *d, int in_len,
+		     unsigned char **d, int in_len,
 		     char *out, int out_len)
 {
     static char table[64] = {
@@ -257,38 +191,41 @@ ipmi_get_6_bit_ascii(int len,
     pos = 0;
     for (i=0; i<len; i++) {
 	switch (bo) {
-	    case 0:
-		val = *d & 0x3f;
-		bo = 6;
-		break;
-	    case 2:
-		val = (*d >> 2) & 0x3f;
-		d++;
-		bo = 0;
-		break;
-	    case 4:
-		val = (*d >> 4) & 0xf;
-		d++;
-		val |= (*d & 0x3) << 4;
-		bo = 2;
-		break;
-	    case 6:
-		val = (*d >> 6) & 0x3;
-		d++;
-		val |= (*d & 0xf) << 2;
-		bo = 4;
-		break;
+	case 0:
+	    val = **d & 0x3f;
+	    bo = 6;
+	    break;
+	case 2:
+	    val = (**d >> 2) & 0x3f;
+	    (*d)++;
+	    bo = 0;
+	    break;
+	case 4:
+	    val = (**d >> 4) & 0xf;
+	    (*d)++;
+	    val |= (**d & 0x3) << 4;
+	    bo = 2;
+	    break;
+	case 6:
+	    val = (**d >> 6) & 0x3;
+	    (*d)++;
+	    val |= (**d & 0xf) << 2;
+	    bo = 4;
+	    break;
 	}
 	*out = table[val];
 	out++;
     }
 
+    if (bo != 0)
+	(*d)++;
+
     return out - out_s;
 }
 
 static unsigned int
 ipmi_get_8_bit_ascii(int len,
-		     unsigned char *d, int in_len,
+		     unsigned char **d, int in_len,
 		     char *out, int out_len)
 {
     int j;
@@ -301,58 +238,77 @@ ipmi_get_8_bit_ascii(int len,
 	len = out_len;
 
     for (j=0; j<len; j++) {
-	*out = *d;
+	*out = **d;
 	out++;
-	d++;
+	(*d)++;
     }
     return len;
 };
 
 unsigned int
-ipmi_get_device_string(unsigned char        *input,
+ipmi_get_device_string(unsigned char        **pinput,
 		       unsigned int         in_len,
 		       char                 *output,
+		       int                  semantics,
 		       int                  force_unicode,
 		       enum ipmi_str_type_e *stype,
 		       unsigned int         max_out_len)
 {
-    int          type;
-    int          len;
-    unsigned int olen;
+    int           type;
+    int           len;
+    unsigned int  olen;
 
     if (max_out_len == 0)
 	return 0;
 
-    if (in_len < 2) {
+    if (in_len <= 0) {
 	*output = '\0';
 	return 0;
     }
 
-    type = (*input >> 6) & 3;
+#if 0
+    /* Note that this is technically correct, but commonly ignored.
+       0xc1 is invalid, but some FRU and SDR data still uses it.  Grr.
+       The FRU stuff has to handle the end-of-area marker c1 itself,
+       anyway, so this is relatively safe.  In a "correct" system you
+       should never see a 0xc1 here, anyway. */
+    if (**pinput == 0xc1) {
+	*output = '\0';
+	(*pinput)++;
+	return 0;
+    }
+#endif
+
+    type = (**pinput >> 6) & 3;
 
     /* Special case for FRU data, type 3 is unicode if the language is
        non-english. */
-    if ((force_unicode) && (type == 3))
+    if ((force_unicode) && (type == 3)) {
 	type = 0;
+	force_unicode = 0;
+    }
 
-    len = *input & 0x3f;
-    input++;
+    len = **pinput & 0x3f;
+    (*pinput)++;
     in_len--;
     *stype = IPMI_ASCII_STR;
     switch (type)
     {
 	case 0: /* Unicode */
-	    olen = ipmi_get_unicode(len, input, in_len, output, max_out_len);
-	    *stype = IPMI_UNICODE_STR;
+	    olen = ipmi_get_unicode(len, pinput, in_len, output, max_out_len);
+	    if (semantics == IPMI_STR_FRU_SEMANTICS)
+		*stype = IPMI_BINARY_STR;
+	    else
+		*stype = IPMI_UNICODE_STR;
 	    break;
 	case 1: /* BCD Plus */
-	    olen = ipmi_get_bcd_plus(len, input, in_len, output, max_out_len);
+	    olen = ipmi_get_bcd_plus(len, pinput, in_len, output, max_out_len);
 	    break;
 	case 2: /* 6-bit ASCII */
-	    olen=ipmi_get_6_bit_ascii(len, input, in_len, output, max_out_len);
+	    olen=ipmi_get_6_bit_ascii(len, pinput, in_len, output, max_out_len);
 	    break;
 	case 3: /* 8-bit ASCII */
-	    olen=ipmi_get_8_bit_ascii(len, input, in_len, output, max_out_len);
+	    olen=ipmi_get_8_bit_ascii(len, pinput, in_len, output, max_out_len);
 	    break;
         default:
 	    olen = 0;
@@ -437,6 +393,7 @@ static char table_6_bit[256] =
 
 static void
 ipmi_set_bcdplus(char          *input,
+		 unsigned int  in_len,
 		 unsigned char *output,
 		 int           *out_len)
 {
@@ -446,31 +403,34 @@ ipmi_set_bcdplus(char          *input,
     int  bit = 0;
     int  count = 0;
 
-    while (*s != '\0') {
-	if (pos >= len) {
-	    output[0] = (0x01 << 6) | count;
-	    return;
-	}
+    while (in_len > 0) {
 	switch(bit) {
 	    case 0:
 		pos++;
-		output[pos] = table_4_bit[(int) *s];
+		if (pos >= len)
+		    goto out_overflow;
+		output[pos] = table_4_bit[(int) *s] - 1;
 		bit = 4;
 		break;
 
 	    case 4:
-		output[pos] |= table_4_bit[(int) *s] << 4;
+		output[pos] |= (table_4_bit[(int) *s] - 1) << 4;
 		bit = 0;
 		break;
 	}
 	count++;
+	in_len--;
+	s++;
     }
+    pos++;
+ out_overflow:
     output[0] = (0x01 << 6) | count;
-    *out_len = pos+1;
+    *out_len = pos;
 }
 
 static void
 ipmi_set_6_bit_ascii(char          *input,
+		     unsigned int  in_len,
 		     unsigned char *output,
 		     int           *out_len)
 {
@@ -482,67 +442,74 @@ ipmi_set_6_bit_ascii(char          *input,
     int  cval;
     int  oval;
 
-    while (*s != '\0') {
-	if (pos >= len) {
-	    output[0] = (0x02 << 6) | count;
-	    return;
-	}
+    while (in_len > 0) {
 	cval = *s;
 	s++;
 	oval = table_6_bit[cval] - 1;
 	switch(bit) {
-	    case 0:
-		pos++;
-		output[pos] = oval;
-		bit = 6;
-		break;
+	case 0:
+	    pos++;
+	    if (pos >= len) 
+		goto out_overflow;
+	    output[pos] = oval;
+	    bit = 6;
+	    break;
 
-	    case 2:
-		output[pos] |= oval << 2;
-		bit = 0;
-		break;
+	case 2:
+	    output[pos] |= oval << 2;
+	    bit = 0;
+	    break;
 
-	    case 4:
-		output[pos] |= oval << 4;
-		pos++;
-		output[pos] = (oval >> 4) & 0x3;
-		bit = 2;
-		break;
+	case 4:
+	    output[pos] |= oval << 4;
+	    pos++;
+	    if (pos >= len) 
+		goto out_overflow;
+	    output[pos] = (oval >> 4) & 0x3;
+	    bit = 2;
+	    break;
 
-	    case 6:
-		output[pos] |= oval << 6;
-		pos++;
-		output[pos] = (oval >> 2) & 0xf;
-		bit = 4;
-		break;
+	case 6:
+	    output[pos] |= oval << 6;
+	    pos++;
+	    if (pos >= len) 
+		goto out_overflow;
+	    output[pos] = (oval >> 2) & 0xf;
+	    bit = 4;
+	    break;
 	}
 	count++;
+	in_len--;
     }
+    pos++;
+ out_overflow:
     output[0] = (0x02 << 6) | count;
-    *out_len = pos+1;
+    *out_len = pos;
 }
 
 static void
 ipmi_set_8_bit_ascii(char          *input,
+		     int           in_len,
 		     unsigned char *output,
 		     int           *out_len)
 {
-    int len = strlen(input+1);
-    if (len > (*out_len - 1))
-	len = *out_len - 1;
-    else
-	*out_len = len + 1;
-    strncpy(input, output+1, len);
-    output[0] = (0x02 << 6) | len;
-}
+    char tmp[2];
+    /* truncate if necessary. */
+    if (in_len > (*out_len - 1))
+	in_len = *out_len - 1;
 
-static void
-ipmi_set_8_bit_ascii_to_unicode(char          *input,
-				unsigned char *output,
-				int           *out_len)
-{
-    /* FIXME - this is not implemented yet. */
-    *out_len = 0;
+    /* A length of 1 is invalid, make it 2 with a nil char */
+    if (in_len == 1) {
+	tmp[0] = input[0];
+	tmp[1] = '\0';
+	input = tmp;
+	in_len++;
+    }
+
+    *out_len = in_len + 1;
+
+    memcpy(output+1, input, in_len);
+    output[0] = (0x03 << 6) | in_len;
 }
 
 void
@@ -553,45 +520,45 @@ ipmi_set_device_string(char                 *input,
 		       int                  force_unicode,
 		       int                  *out_len)
 {
-    char *s = input+1;
-    int  bsize = 0; /* Start with 4-bit. */
+    char         *s = input;
+    int          bsize = 0; /* Start with 4-bit. */
+    unsigned int i;
 
-    /* Max size is 63 (62 bytes + the type byte). */
-    if (*out_len > 63)
-	*out_len = 63;
+    /* Max size is 64 (63 bytes + the type byte). */
+    if (*out_len > 64)
+	*out_len = 64;
+    /* Truncate */
+    if (in_len > 63)
+	in_len = 63;
 
     if (type == IPMI_ASCII_STR) {
-	while (*s != '\0') {
-	    if (table_4_bit[(int) *s] == 0)
+	for (i=0; i<in_len; i++) {
+	    if (table_4_bit[(int) *s] == 0) {
 		bsize |= 1;
-	    else if (table_6_bit[(int) *s] == 0) {
-		bsize |= 2;
-		break;
+		if (table_6_bit[(int) *s] == 0) {
+		    bsize |= 2;
+		    break;
+		}
 	    }
 	    s++;
 	}
 	if (bsize == 0) {
 	    /* We can encode it in 4-bit BCD+ */
-	    ipmi_set_bcdplus(input, output, out_len);
+	    ipmi_set_bcdplus(input, in_len, output, out_len);
 	} else if (bsize == 1) {
 	    /* We can encode it in 6-bit ASCII. */
-	    ipmi_set_6_bit_ascii(input, output, out_len);
+	    ipmi_set_6_bit_ascii(input, in_len, output, out_len);
 	} else {
-	    /* Hack for FRU information, if the language is
-	       non-english and the type is ascii, it's unicode. */
-	    if (force_unicode) {
-		/* The input is ASCII and the output is unicode. */
-		ipmi_set_8_bit_ascii_to_unicode(input, output, out_len);
-	    } else {
-		/* 8-bit ASCII is required. */
-		ipmi_set_8_bit_ascii(input, output, out_len);
-	    }
+	    ipmi_set_8_bit_ascii(input, in_len, output, out_len);
 	}
     } else {
 	/* The input and output are unicode. */
 	if (in_len > *out_len-1)
 	    in_len = *out_len-1;
-	*output = in_len;
+	if ((force_unicode) && (type == IPMI_UNICODE_STR))
+	    *output = (0x3 << 6) | in_len;
+	else
+	    *output = in_len;
 	memcpy(output+1, input, in_len);
 	*out_len = in_len + 1;
     }
@@ -792,7 +759,7 @@ int ipmi_threshold_set(ipmi_thresholds_t  *th,
 	if (rv)
 	    return rv;
 	if (!val)
-	    return ENOTSUP;
+	    return ENOSYS;
     }
 
     th->vals[threshold].status = 1;
@@ -811,7 +778,7 @@ int ipmi_threshold_get(ipmi_thresholds_t  *th,
 	*value = th->vals[threshold].val;
 	return 0;
     } else {
-	return ENOTSUP;
+	return ENOSYS;
     }
 }
 
@@ -910,37 +877,40 @@ ipmi_set_threshold_out_of_range(ipmi_states_t      *states,
 void ipmi_oem_force_conn_init(void);
 int ipmi_oem_motorola_mxp_init(void);
 int ipmi_oem_intel_init(void);
-int _ipmi_pet_init(void);
-int _ipmi_conn_init(void);
 int ipmi_oem_atca_conn_init(void);
 int ipmi_oem_atca_init(void);
 int init_oem_test(void);
+int _ipmi_smi_init(os_handler_t *os_hnd);
+int _ipmi_lan_init(os_handler_t *os_hnd);
+int ipmi_malloc_init(os_handler_t *os_hnd);
+int _ipmi_rakp_init(void);
+int _ipmi_aes_cbc_init(void);
+int _ipmi_hmac_init(void);
+int _ipmi_md5_init(void);
 
 void ipmi_oem_atca_conn_shutdown(void);
 void ipmi_oem_intel_shutdown(void);
 void ipmi_oem_atca_shutdown(void);
-void _ipmi_pet_shutdown(void);
+int _ipmi_smi_shutdown(void);
+int _ipmi_lan_shutdown(void);
 
 int
 ipmi_init(os_handler_t *handler)
 {
     int rv;
 
+    ipmi_os_handler = handler;
+
+    /* Set up memory allocation first */
+    ipmi_malloc_init(handler);
+
     /* Set up logging in malloc code. */
     ipmi_malloc_log = ipmi_log;
 
-    rv = _ipmi_conn_init();
+    rv = _ipmi_conn_init(handler);
     if (rv)
 	return rv;
 
-    if (handler->create_rwlock) {
-	rv = handler->create_rwlock(handler, &global_lock);
-	if (rv)
-	    return rv;
-    } else {
-	global_lock = NULL;
-    }
-
     if (handler->create_lock) {
 	rv = handler->create_lock(handler, &seq_lock);
 	if (rv)
@@ -948,15 +918,32 @@ ipmi_init(os_handler_t *handler)
     } else {
 	seq_lock = NULL;
     }
-    ipmi_os_handler = handler;
+
+#ifdef HAVE_OPENIPMI_SMI
+    rv = _ipmi_smi_init(handler);
+    if (rv)
+	goto out_err;
+#endif
+
+    rv = _ipmi_lan_init(handler);
+    if (rv)
+	goto out_err;
+
     _ipmi_domain_init();
     _ipmi_mc_init();
-    rv = _ipmi_conn_init();
+
+    rv = _ipmi_rakp_init();
     if (rv)
-        goto out_err;
-    rv = _ipmi_pet_init();
+	goto out_err;
+    rv = _ipmi_aes_cbc_init();
     if (rv)
-        goto out_err;
+	goto out_err;
+    rv = _ipmi_hmac_init();
+    if (rv)
+	goto out_err;
+    rv = _ipmi_md5_init();
+    if (rv)
+	goto out_err;
 
     /* Call the OEM handlers. */
     ipmi_oem_force_conn_init();
@@ -969,34 +956,37 @@ ipmi_init(os_handler_t *handler)
     return 0;
 
  out_err:
+#ifdef HAVE_OPENIPMI_SMI
+    _ipmi_smi_shutdown();
+#endif
+    _ipmi_lan_shutdown();
     ipmi_oem_intel_shutdown();
     _ipmi_mc_shutdown();
     _ipmi_domain_shutdown();
-    if (global_lock)
-	handler->destroy_rwlock(ipmi_os_handler, global_lock);
     if (seq_lock)
 	handler->destroy_lock(ipmi_os_handler, seq_lock);
+    ipmi_os_handler = NULL;
     return rv;
 }
 
 void
 ipmi_shutdown(void)
 {
+    _ipmi_lan_shutdown();
+#ifdef HAVE_OPENIPMI_SMI
+    _ipmi_smi_shutdown();
+#endif
     ipmi_oem_atca_shutdown();
     ipmi_oem_atca_conn_shutdown();
     ipmi_oem_intel_shutdown();
-    _ipmi_pet_shutdown();
     _ipmi_mc_shutdown();
     _ipmi_domain_shutdown();
     _ipmi_conn_shutdown();
-    if (global_lock)
-	ipmi_os_handler->destroy_rwlock(ipmi_os_handler, global_lock);
     if (seq_lock)
 	ipmi_os_handler->destroy_lock(ipmi_os_handler, seq_lock);
-    global_lock = NULL;
 }
 
-enum con_type_e { SMI, LAN, MXP };
+enum con_type_e { SMI, LAN };
 
 struct ipmi_args_s
 {
@@ -1007,14 +997,34 @@ struct ipmi_args_s
     char            *str_addr[2];
     char            *str_port[2];
     int             num_addr;
+    int             authtype_set;
     int             authtype;
     int             privilege;
-    char            username[17];
-    char            password[17];
+    int             privilege_set;
+    int             username_set;
+    char            username[16];
+    unsigned int    username_len;
+    int             password_set;
+    char            password[20];
+    unsigned int    password_len;
+
+    int             auth_alg_set;
+    unsigned int    auth_alg;
+    int             integ_alg_set;
+    unsigned int    integ_alg;
+    int		    conf_alg_set;
+    unsigned int    conf_alg;
+    int             name_lookup_only_set;
+    int             name_lookup_only;
+    int             bmc_key_set;
+    char            bmc_key[20];
+    unsigned int    bmc_key_len;
 
     unsigned char   swid;
     struct in_addr  lan_addr[2];
     int             lan_port[2];
+
+    unsigned int    hacks;
 };
 
 #define CHECK_ARG \
@@ -1028,11 +1038,12 @@ struct ipmi_args_s
 int
 ipmi_parse_args(int         *curr_arg,
 		int         arg_count,
-		const char  **args,
+		char        * const *args,
 		ipmi_args_t **iargs)
 {
     ipmi_args_t *p;
-    int rv;
+    int         rv;
+    int         len;
     
     p = ipmi_mem_alloc(sizeof(*p));
     if (!p)
@@ -1076,6 +1087,8 @@ ipmi_parse_args(int         *curr_arg,
 	    p->authtype = IPMI_AUTHTYPE_MD5;
 	} else if (strcmp(args[*curr_arg], "straight") == 0) {
 	    p->authtype = IPMI_AUTHTYPE_STRAIGHT;
+	} else if (strcmp(args[*curr_arg], "rmcp+") == 0) {
+	    p->authtype = IPMI_AUTHTYPE_RMCP_PLUS;
 	} else if (p->num_addr == 1) {
 	    p->num_addr++;
 	    p->str_addr[1] = ipmi_strdup(args[*curr_arg]);
@@ -1096,6 +1109,7 @@ ipmi_parse_args(int         *curr_arg,
 	    goto out_err;
 	}
 	(*curr_arg)++; CHECK_ARG;
+	p->authtype_set = 1;
 
 	if (strcmp(args[*curr_arg], "callback") == 0) {
 	    p->privilege = IPMI_PRIVILEGE_CALLBACK;
@@ -1111,93 +1125,234 @@ ipmi_parse_args(int         *curr_arg,
 	    rv = EINVAL;
 	    goto out_err;
 	}
+	p->privilege_set = 1;
 	(*curr_arg)++; CHECK_ARG;
 
-	memset(p->username, 0, sizeof(p->username));
-	memset(p->password, 0, sizeof(p->password));
-	strncpy(p->username, args[*curr_arg], 16);
+	len = strlen(args[*curr_arg]);
+	if (len > 16)
+	    len = 16;
+	memcpy(p->username, args[*curr_arg], len);
 	p->username[16] = '\0';
+	p->username_len = len;
+	p->username_set = 1;
 	(*curr_arg)++; CHECK_ARG;
-	strncpy(p->password, args[*curr_arg], 16);
-	p->password[16] = '\0';
+
+	len = strlen(args[*curr_arg]);
+	if (len > 16)
+	    len = 16;
+	memcpy(p->password, args[*curr_arg], len);
+	p->password_len = len;
+	p->password_set = 1;
 	(*curr_arg)++;
-    } else if (strcmp(args[*curr_arg], "mxp") == 0) {
-	struct hostent *ent;
+    } else {
+	rv = EINVAL;
+	goto out_err;
+    }
+
+    *iargs = p;
+    return 0;
+
+ out_err:
+    ipmi_free_args(p);
+    return rv;
+}
+
+int
+ipmi_parse_args2(int         *curr_arg,
+		 int         arg_count,
+		 char        * const *args,
+		 ipmi_args_t **iargs)
+{
+    ipmi_args_t *p;
+    int rv;
+    
+    p = ipmi_mem_alloc(sizeof(*p));
+    if (!p)
+	return ENOMEM;
+    memset(p, 0, sizeof(*p));
+
+    CHECK_ARG;
+
+    if (strcmp(args[*curr_arg], "smi") == 0) {
+	(*curr_arg)++; CHECK_ARG;
+
+	p->con_type = SMI;
+
+	p->smi_intf = atoi(args[*curr_arg]);
+	(*curr_arg)++;
+    } else if (strcmp(args[*curr_arg], "lan") == 0) {
+	int i;
+	int len;
 
 	(*curr_arg)++; CHECK_ARG;
 
-	p->con_type = MXP;
+	p->con_type = LAN;
 	p->num_addr = 1;
 
-	ent = gethostbyname(args[*curr_arg]);
-	if (!ent) {
-	    rv = h_errno;
-	    goto out_err;
-	}
-	memcpy(&p->lan_addr[0],
-	       ent->h_addr_list[0],
-	       ent->h_length);
-	(*curr_arg)++; CHECK_ARG;
-	p->lan_port[0] = atoi(args[*curr_arg]);
-	(*curr_arg)++; CHECK_ARG;
+	while (*curr_arg < arg_count) {
+	    if (args[*curr_arg][0] != '-') {
+		break;
+	    }
 
-    doauth_mxp:
-	if (strcmp(args[*curr_arg], "none") == 0) {
-	    p->authtype = IPMI_AUTHTYPE_NONE;
-	} else if (strcmp(args[*curr_arg], "md2") == 0) {
-	    p->authtype = IPMI_AUTHTYPE_MD2;
-	} else if (strcmp(args[*curr_arg], "md5") == 0) {
-	    p->authtype = IPMI_AUTHTYPE_MD5;
-	} else if (strcmp(args[*curr_arg], "straight") == 0) {
-	    p->authtype = IPMI_AUTHTYPE_STRAIGHT;
-	} else if (p->num_addr == 1) {
-	    p->num_addr++;
-	    ent = gethostbyname(args[*curr_arg]);
-	    if (!ent) {
-		rv = h_errno;
+	    if (strcmp(args[*curr_arg], "-U") == 0) {
+		(*curr_arg)++; CHECK_ARG;
+		len = strlen(args[*curr_arg]);
+		if (len > 16)
+		    len = 16;
+		memcpy(p->username, args[*curr_arg], len);
+		p->username_set = 1;
+		p->username_len = len;
+	    } else if (strcmp(args[*curr_arg], "-P") == 0) {
+		(*curr_arg)++; CHECK_ARG;
+		len = strlen(args[*curr_arg]);
+		if (len > 20)
+		    len = 20;
+		memcpy(p->password, args[*curr_arg], len);
+		p->password_set = 1;
+		p->password_len = len;
+	    } else if (strcmp(args[*curr_arg], "-H") == 0) {
+		(*curr_arg)++; CHECK_ARG;
+		if (strcmp(args[*curr_arg], "intelplus") == 0)
+		    p->hacks |= IPMI_CONN_HACK_RAKP3_WRONG_ROLEM;
+		else if (strcmp(args[*curr_arg], "rakp3_wrong_rolem") == 0)
+		    p->hacks |= IPMI_CONN_HACK_RAKP3_WRONG_ROLEM;
+		else if (strcmp(args[*curr_arg], "rmcpp_integ_sik") == 0)
+		    p->hacks |= IPMI_CONN_HACK_RMCPP_INTEG_SIK;
+		/* Ignore unknown hacks. */
+	    } else if (strcmp(args[*curr_arg], "-s") == 0) {
+		p->num_addr = 2;
+	    } else if (strcmp(args[*curr_arg], "-A") == 0) {
+		(*curr_arg)++; CHECK_ARG;
+		if (strcmp(args[*curr_arg], "none") == 0) {
+		    p->authtype = IPMI_AUTHTYPE_NONE;
+		} else if (strcmp(args[*curr_arg], "md2") == 0) {
+		    p->authtype = IPMI_AUTHTYPE_MD2;
+		} else if (strcmp(args[*curr_arg], "md5") == 0) {
+		    p->authtype = IPMI_AUTHTYPE_MD5;
+		} else if (strcmp(args[*curr_arg], "straight") == 0) {
+		    p->authtype = IPMI_AUTHTYPE_STRAIGHT;
+		} else if (strcmp(args[*curr_arg], "rmcp+") == 0) {
+		    p->authtype = IPMI_AUTHTYPE_RMCP_PLUS;
+		} else {
+		    rv = EINVAL;
+		    goto out_err;
+		}
+		p->authtype_set = 1;
+	    } else if (strcmp(args[*curr_arg], "-L") == 0) {
+		(*curr_arg)++; CHECK_ARG;
+
+		if (strcmp(args[*curr_arg], "callback") == 0) {
+		    p->privilege = IPMI_PRIVILEGE_CALLBACK;
+		} else if (strcmp(args[*curr_arg], "user") == 0) {
+		    p->privilege = IPMI_PRIVILEGE_USER;
+		} else if (strcmp(args[*curr_arg], "operator") == 0) {
+		    p->privilege = IPMI_PRIVILEGE_OPERATOR;
+		} else if (strcmp(args[*curr_arg], "admin") == 0) {
+		    p->privilege = IPMI_PRIVILEGE_ADMIN;
+		} else if (strcmp(args[*curr_arg], "oem") == 0) {
+		    p->privilege = IPMI_PRIVILEGE_OEM;
+		} else {
+		    rv = EINVAL;
+		    goto out_err;
+		}
+		p->privilege_set = 1;
+	    } else if (strcmp(args[*curr_arg], "-p") == 0) {
+		(*curr_arg)++; CHECK_ARG;
+		p->str_port[0] = ipmi_strdup(args[*curr_arg]);
+		if (p->str_port[0] == NULL) {
+		    rv = ENOMEM;
+		    goto out_err;
+		}
+	    } else if (strcmp(args[*curr_arg], "-p2") == 0) {
+		(*curr_arg)++; CHECK_ARG;
+		p->str_port[2] = ipmi_strdup(args[*curr_arg]);
+		if (p->str_port[2] == NULL) {
+		    rv = ENOMEM;
+		    goto out_err;
+		}
+	    } else if (strcmp(args[*curr_arg], "-Ra") == 0) {
+		(*curr_arg)++; CHECK_ARG;
+
+		if (strcmp(args[*curr_arg], "bmcpick") == 0) {
+		    p->auth_alg = IPMI_LANP_AUTHENTICATION_ALGORITHM_BMCPICK;
+		} else if (strcmp(args[*curr_arg], "rakp_none") == 0) {
+		    p->auth_alg = IPMI_LANP_AUTHENTICATION_ALGORITHM_RAKP_NONE;
+		} else if (strcmp(args[*curr_arg], "rakp_hmac_sha1") == 0) {
+		    p->auth_alg = IPMI_LANP_AUTHENTICATION_ALGORITHM_RAKP_HMAC_SHA1;
+		} else if (strcmp(args[*curr_arg], "rakp_hmac_md5") == 0) {
+		    p->auth_alg = IPMI_LANP_AUTHENTICATION_ALGORITHM_RAKP_HMAC_MD5;
+		} else {
+		    rv = EINVAL;
+		    goto out_err;
+		}
+		p->auth_alg_set = 1;
+	    } else if (strcmp(args[*curr_arg], "-Ri") == 0) {
+		(*curr_arg)++; CHECK_ARG;
+
+		if (strcmp(args[*curr_arg], "bmcpick") == 0) {
+		    p->integ_alg = IPMI_LANP_INTEGRITY_ALGORITHM_BMCPICK;
+		} else if (strcmp(args[*curr_arg], "none") == 0) {
+		    p->integ_alg = IPMI_LANP_INTEGRITY_ALGORITHM_NONE;
+		} else if (strcmp(args[*curr_arg], "hmac_sha1") == 0) {
+		    p->integ_alg = IPMI_LANP_INTEGRITY_ALGORITHM_HMAC_SHA1_96;
+		} else if (strcmp(args[*curr_arg], "hmac_md5") == 0) {
+		    p->integ_alg = IPMI_LANP_INTEGRITY_ALGORITHM_HMAC_MD5_128;
+		} else if (strcmp(args[*curr_arg], "md5") == 0) {
+		    p->integ_alg = IPMI_LANP_INTEGRITY_ALGORITHM_MD5_128;
+		} else {
+		    rv = EINVAL;
+		    goto out_err;
+		}
+		p->integ_alg_set = 1;
+	    } else if (strcmp(args[*curr_arg], "-Rc") == 0) {
+		(*curr_arg)++; CHECK_ARG;
+
+		if (strcmp(args[*curr_arg], "bmcpick") == 0) {
+		    p->conf_alg = IPMI_LANP_CONFIDENTIALITY_ALGORITHM_BMCPICK;
+		} else if (strcmp(args[*curr_arg], "none") == 0) {
+		    p->conf_alg = IPMI_LANP_CONFIDENTIALITY_ALGORITHM_NONE;
+		} else if (strcmp(args[*curr_arg], "aes_cbc_128") == 0) {
+		    p->conf_alg = IPMI_LANP_CONFIDENTIALITY_ALGORITHM_AES_CBC_128;
+		} else if (strcmp(args[*curr_arg], "xrc4_128") == 0) {
+		    p->conf_alg = IPMI_LANP_CONFIDENTIALITY_ALGORITHM_xRC4_128;
+		} else if (strcmp(args[*curr_arg], "xrc4_40") == 0) {
+		    p->conf_alg = IPMI_LANP_CONFIDENTIALITY_ALGORITHM_xRC4_40;
+		} else {
+		    rv = EINVAL;
+		    goto out_err;
+		}
+		p->conf_alg_set = 1;
+	    } else if (strcmp(args[*curr_arg], "-Rl") == 0) {
+		p->name_lookup_only = 0;
+		p->name_lookup_only_set = 1;
+	    } else if (strcmp(args[*curr_arg], "-Rk") == 0) {
+		(*curr_arg)++; CHECK_ARG;
+		len = strlen(args[*curr_arg]);
+		if (len > 20)
+		    len = 20;
+		memcpy(p->bmc_key, args[*curr_arg], len);
+		p->bmc_key_set = 1;
+		p->bmc_key_len = len;
+	    }
+	    (*curr_arg)++;
+	}
+
+	for (i=0; i<p->num_addr; i++) {
+	    CHECK_ARG;
+	    p->str_addr[i] = ipmi_strdup(args[*curr_arg]);
+	    if (p->str_addr[0] == NULL) {
+		rv = ENOMEM;
 		goto out_err;
 	    }
-	    memcpy(&p->lan_addr[1],
-		   ent->h_addr_list[0],
-		   ent->h_length);
-	    (*curr_arg)++; CHECK_ARG;
-	    p->lan_port[1] = atoi(args[*curr_arg]);
-	    (*curr_arg)++; CHECK_ARG;
-
-	    goto doauth_mxp;
-	} else {
-	    rv = EINVAL;
-	    goto out_err;
+	    (*curr_arg)++;
+	    if (! p->str_port[i]) {
+		p->str_port[i] = ipmi_strdup("623");
+		if (p->str_port[0] == NULL) {
+		    rv = ENOMEM;
+		    goto out_err;
+		}
+	    }
 	}
-	(*curr_arg)++; CHECK_ARG;
-
-	if (strcmp(args[*curr_arg], "callback") == 0) {
-	    p->privilege = IPMI_PRIVILEGE_CALLBACK;
-	} else if (strcmp(args[*curr_arg], "user") == 0) {
-	    p->privilege = IPMI_PRIVILEGE_USER;
-	} else if (strcmp(args[*curr_arg], "operator") == 0) {
-	    p->privilege = IPMI_PRIVILEGE_OPERATOR;
-	} else if (strcmp(args[*curr_arg], "admin") == 0) {
-	    p->privilege = IPMI_PRIVILEGE_ADMIN;
-	} else if (strcmp(args[*curr_arg], "oem") == 0) {
-	    p->privilege = IPMI_PRIVILEGE_OEM;
-	} else {
-	    rv = EINVAL;
-	    goto out_err;
-	}
-	(*curr_arg)++; CHECK_ARG;
-
-	memset(p->username, 0, sizeof(p->username));
-	memset(p->password, 0, sizeof(p->password));
-	strncpy(p->username, args[*curr_arg], 16);
-	p->username[16] = '\0';
-	(*curr_arg)++; CHECK_ARG;
-	strncpy(p->password, args[*curr_arg], 16);
-	p->password[16] = '\0';
-	(*curr_arg)++; CHECK_ARG;
-
-	p->swid = strtoul(args[*curr_arg], NULL, 0);
-	(*curr_arg)++;
     } else {
 	rv = EINVAL;
 	goto out_err;
@@ -1225,46 +1380,161 @@ ipmi_free_args(ipmi_args_t *args)
     ipmi_mem_free(args);
 }
 
+int
+ipmi_parse_options(ipmi_open_option_t *option,
+		   char               *arg)
+{
+    if (strcmp(arg, "-noall") == 0) {
+	option->option = IPMI_OPEN_OPTION_ALL;
+	option->ival = 0;
+    } else if (strcmp(arg, "-all") == 0) {
+	option->option = IPMI_OPEN_OPTION_ALL;
+	option->ival = 1;
+    } else if (strcmp(arg, "-nosdrs") == 0) {
+	option->option = IPMI_OPEN_OPTION_SDRS;
+	option->ival = 0;
+    } else if (strcmp(arg, "-sdrs") == 0) {
+	option->option = IPMI_OPEN_OPTION_SDRS;
+	option->ival = 1;
+    } else if (strcmp(arg, "-nofrus") == 0) {
+	option->option = IPMI_OPEN_OPTION_FRUS;
+	option->ival = 0;
+    } else if (strcmp(arg, "-frus") == 0) {
+	option->option = IPMI_OPEN_OPTION_FRUS;
+	option->ival = 1;
+    } else if (strcmp(arg, "-nosel") == 0) {
+	option->option = IPMI_OPEN_OPTION_SEL;
+	option->ival = 0;
+    } else if (strcmp(arg, "-sel") == 0) {
+	option->option = IPMI_OPEN_OPTION_SEL;
+	option->ival = 1;
+    } else if (strcmp(arg, "-noipmbscan") == 0) {
+	option->option = IPMI_OPEN_OPTION_IPMB_SCAN;
+	option->ival = 0;
+    } else if (strcmp(arg, "-ipmbscan") == 0) {
+	option->option = IPMI_OPEN_OPTION_IPMB_SCAN;
+	option->ival = 1;
+    } else if (strcmp(arg, "-nooeminit") == 0) {
+	option->option = IPMI_OPEN_OPTION_OEM_INIT;
+	option->ival = 0;
+    } else if (strcmp(arg, "-oeminit") == 0) {
+	option->option = IPMI_OPEN_OPTION_OEM_INIT;
+	option->ival = 1;
+    } else if (strcmp(arg, "-noseteventrcvr") == 0) {
+	option->option = IPMI_OPEN_OPTION_SET_EVENT_RCVR;
+	option->ival = 0;
+    } else if (strcmp(arg, "-seteventrcvr") == 0) {
+	option->option = IPMI_OPEN_OPTION_SET_EVENT_RCVR;
+	option->ival = 1;
+    } else if (strcmp(arg, "-noactivate") == 0) {
+	option->option = IPMI_OPEN_OPTION_ACTIVATE_IF_POSSIBLE;
+	option->ival = 0;
+    } else if (strcmp(arg, "-activate") == 0) {
+	option->option = IPMI_OPEN_OPTION_ACTIVATE_IF_POSSIBLE;
+	option->ival = 1;
+    } else if (strcmp(arg, "-nosetseltime") == 0) {
+	option->option = IPMI_OPEN_OPTION_SET_SEL_TIME;
+	option->ival = 0;
+    } else if (strcmp(arg, "-setseltime") == 0) {
+	option->option = IPMI_OPEN_OPTION_SET_SEL_TIME;
+	option->ival = 1;
+    } else
+	return EINVAL;
+
+    return 0;
+}
+
 int
 ipmi_args_setup_con(ipmi_args_t  *args,
 		    os_handler_t *handlers,
 		    void         *user_data,
 		    ipmi_con_t   **con)
 {
+    int rv;
+
     switch(args->con_type) {
+#ifdef HAVE_OPENIPMI_SMI
     case SMI:
-	return ipmi_smi_setup_con(args->smi_intf, handlers, user_data, con);
+	rv = ipmi_smi_setup_con(args->smi_intf, handlers, user_data, con);
+	if (!rv)
+	    (*con)->hacks = args->hacks;
+	break;
+#endif
 
     case LAN:
-	return ipmi_ip_setup_con(args->str_addr,
-				 args->str_port,
-				 args->num_addr,
-				 args->authtype,
-				 args->privilege,
-				 args->username,
-				 strlen(args->username),
-				 args->password,
-				 strlen(args->password),
-				 handlers, user_data,
-				 con);
+    {
+	int              i;
+	ipmi_lanp_parm_t parms[11];
 
-    case MXP:
-	return mxp_lan_setup_con(args->lan_addr,
-				 args->lan_port,
-				 args->num_addr,
-				 args->authtype,
-				 args->privilege,
-				 args->username,
-				 strlen(args->username),
-				 args->password,
-				 strlen(args->password),
-				 handlers, user_data,
-				 args->swid,
-				 con);
+	i = 0;
+	parms[i].parm_id = IPMI_LANP_PARMID_ADDRS;
+	parms[i].parm_data = args->str_addr;
+	parms[i].parm_data_len = args->num_addr;
+	i++;
+	parms[i].parm_id = IPMI_LANP_PARMID_PORTS;
+	parms[i].parm_data = args->str_port;
+	parms[i].parm_data_len = args->num_addr;
+	i++;
+	if (args->authtype_set) {
+	    parms[i].parm_id = IPMI_LANP_PARMID_AUTHTYPE;
+	    parms[i].parm_val = args->authtype;
+	    i++;
+	}
+	if (args->privilege_set) {
+	    parms[i].parm_id = IPMI_LANP_PARMID_PRIVILEGE;
+	    parms[i].parm_val = args->privilege;
+	    i++;
+	}
+	if (args->username_set) {
+	    parms[i].parm_id = IPMI_LANP_PARMID_USERNAME;
+	    parms[i].parm_data = args->username;
+	    parms[i].parm_data_len = args->username_len;
+	    i++;
+	}
+	if (args->password_set) {
+	    parms[i].parm_id = IPMI_LANP_PARMID_PASSWORD;
+	    parms[i].parm_data = args->password;
+	    parms[i].parm_data_len = args->password_len;
+	    i++;
+	}
+	if (args->auth_alg_set) {
+	    parms[i].parm_id = IPMI_LANP_AUTHENTICATION_ALGORITHM;
+	    parms[i].parm_val = args->auth_alg;
+	    i++;
+	}
+	if (args->integ_alg_set) {
+	    parms[i].parm_id = IPMI_LANP_INTEGRITY_ALGORITHM;
+	    parms[i].parm_val = args->integ_alg;
+	    i++;
+	}
+	if (args->conf_alg_set) {
+	    parms[i].parm_id = IPMI_LANP_CONFIDENTIALITY_ALGORITHM;
+	    parms[i].parm_val = args->conf_alg;
+	    i++;
+	}
+	if (args->name_lookup_only_set) {
+	    parms[i].parm_id = IPMI_LANP_NAME_LOOKUP_ONLY;
+	    parms[i].parm_val = args->name_lookup_only;
+	    i++;
+	}
+	if (args->bmc_key_set) {
+	    parms[i].parm_id = IPMI_LANP_BMC_KEY;
+	    parms[i].parm_data = args->bmc_key;
+	    parms[i].parm_data_len = args->bmc_key_len;
+	    i++;
+	}
+	rv = ipmi_lanp_setup_con(parms, i, handlers, user_data, con);
+	if (!rv)
+	    (*con)->hacks = args->hacks;
+	break;
+    }
 
     default:
-	return EINVAL;
+	rv = EINVAL;
+	break;
     }
+
+    return rv;
 }
 
 /* This is the number of seconds between 1/1/70 (IPMI event date) and
@@ -1288,7 +1558,7 @@ ipmi_handle_snmp_trap_data(void            *src_addr,
 	dump_hex(src_addr, src_addr_len);
 	ipmi_log(IPMI_LOG_DEBUG_CONT, "\n data is:\n  ");
 	dump_hex(data, data_len);
-	ipmi_log(IPMI_LOG_DEBUG_END, "\n");
+	ipmi_log(IPMI_LOG_DEBUG_END, " ");
     }
 
     if (data_len < 46)
@@ -1385,3 +1655,131 @@ ipmi_openipmi_version(void)
 {
     return OPENIPMI_VERSION;
 }
+
+ipmi_msgi_t *
+ipmi_alloc_msg_item(void)
+{
+    ipmi_msgi_t *rv;
+
+    rv = ipmi_mem_alloc(sizeof(ipmi_msgi_t));
+    if (!rv)
+	return NULL;
+    memset(rv, 0, sizeof(rv));
+    rv->msg.data = rv->data;
+    return rv;
+}
+
+void
+ipmi_free_msg_item(ipmi_msgi_t *item)
+{
+    if (item->msg.data && (item->msg.data != item->data))
+	ipmi_free_msg_item_data(item->msg.data);
+    ipmi_mem_free(item);
+}
+
+void *
+ipmi_alloc_msg_item_data(unsigned int size)
+{
+    return ipmi_mem_alloc(size);
+}
+
+void
+ipmi_free_msg_item_data(void *data)
+{
+    ipmi_mem_free(data);
+}
+
+void
+ipmi_move_msg_item(ipmi_msgi_t *new_item, ipmi_msgi_t *old_item)
+{
+    if (new_item->msg.data && (new_item->msg.data != new_item->data))
+	ipmi_free_msg_item_data(new_item->msg.data);
+    memcpy(&new_item->addr, &old_item->addr, old_item->addr_len);
+    new_item->addr_len = old_item->addr_len;
+    new_item->msg = old_item->msg;
+
+    if (!old_item->msg.data) {
+	/* Nothing to do */
+    } else if (old_item->msg.data != old_item->data) {
+	/* Copied the actual data pointer. */
+	old_item->msg.data = NULL;
+    } else {
+	memcpy(new_item->data, old_item->data, old_item->msg.data_len);
+	new_item->msg.data = new_item->data;
+    }
+}
+
+void
+ipmi_handle_rsp_item_copyall(ipmi_con_t            *ipmi,
+			     ipmi_msgi_t           *rspi,
+			     ipmi_addr_t           *addr,
+			     unsigned int          addr_len,
+			     ipmi_msg_t            *msg,
+			     ipmi_ll_rsp_handler_t rsp_handler)
+{
+    int used = IPMI_MSG_ITEM_NOT_USED;
+
+    memcpy(&rspi->addr, addr, addr_len);
+    rspi->addr_len = addr_len;
+    rspi->msg = *msg;
+    memcpy(rspi->data, msg->data, msg->data_len);
+    rspi->msg.data = rspi->data;
+
+    /* call the user handler. */
+    if (rsp_handler)
+	used = rsp_handler(ipmi, rspi);
+
+    if (!used)
+	ipmi_free_msg_item(rspi);
+}
+
+void
+ipmi_handle_rsp_item_copymsg(ipmi_con_t            *ipmi,
+			     ipmi_msgi_t           *rspi,
+			     ipmi_msg_t            *msg,
+			     ipmi_ll_rsp_handler_t rsp_handler)
+{
+    int used = IPMI_MSG_ITEM_NOT_USED;
+
+    rspi->msg = *msg;
+    memcpy(rspi->data, msg->data, msg->data_len);
+    rspi->msg.data = rspi->data;
+
+    /* call the user handler. */
+    if (rsp_handler)
+	used = rsp_handler(ipmi, rspi);
+
+    if (!used)
+	ipmi_free_msg_item(rspi);
+}
+
+void
+ipmi_handle_rsp_item(ipmi_con_t            *ipmi,
+		     ipmi_msgi_t           *rspi,
+		     ipmi_ll_rsp_handler_t rsp_handler)
+{
+    int used = IPMI_MSG_ITEM_NOT_USED;
+
+    /* call the user handler. */
+    if (rsp_handler)
+	used = rsp_handler(ipmi, rspi);
+
+    if (!used)
+	ipmi_free_msg_item(rspi);
+}
+
+os_handler_t *
+ipmi_alloc_os_handler(void)
+{
+    os_handler_t *rv = ipmi_mem_alloc(sizeof(*rv));
+    if (rv)
+	memset(rv, 0, sizeof(*rv));
+    return rv;
+}
+
+void
+ipmi_free_os_handler(os_handler_t *handler)
+{
+    ipmi_mem_free(handler);
+}
+
