@@ -31,6 +31,8 @@
  *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
+#ifdef HAVE_OPENIPMI_SMI
+
 #include <sys/ioctl.h>
 #include <sys/socket.h>
 #include <sys/types.h>
@@ -44,12 +46,16 @@
 
 #include <linux/ipmi.h>
 #include <net/af_ipmi.h>
+
 #include <OpenIPMI/ipmi_conn.h>
-#include <OpenIPMI/ipmi_event.h>
 #include <OpenIPMI/ipmi_msgbits.h>
-#include <OpenIPMI/ipmi_int.h>
 #include <OpenIPMI/ipmi_smi.h>
 #include <OpenIPMI/ipmi_err.h>
+#include <OpenIPMI/ipmi_auth.h>
+
+#include <OpenIPMI/internal/ipmi_event.h>
+#include <OpenIPMI/internal/ipmi_int.h>
+#include <OpenIPMI/internal/locked_list.h>
 
 /* We time the SMI messages, but we have a long timer. */
 #define SMI_TIMEOUT 60000
@@ -86,8 +92,7 @@ typedef struct pending_cmd_s
     ipmi_addr_t           addr;
     unsigned int          addr_len;
     ipmi_ll_rsp_handler_t rsp_handler;
-    void                  *rsp_data;
-    void                  *data2, *data3, *data4;
+    ipmi_msgi_t           *rsp_item;
     int                   use_orig_addr;
     ipmi_addr_t           orig_addr;
     unsigned int          orig_addr_len;
@@ -105,18 +110,10 @@ typedef struct cmd_handler_s
     struct cmd_handler_s *next, *prev;
 } cmd_handler_t;
 
-struct ipmi_ll_event_handler_id_s
-{
-    ipmi_con_t            *ipmi;
-    ipmi_ll_evt_handler_t handler;
-    void                  *event_data;
-    void                  *data2;
-
-    ipmi_ll_event_handler_id_t *next, *prev;
-};
-
 typedef struct smi_data_s
 {
+    int                        refcount;
+
     ipmi_con_t                 *ipmi;
     int                        using_socket;
     int                        fd;
@@ -126,23 +123,25 @@ typedef struct smi_data_s
     cmd_handler_t              *cmd_handlers;
     ipmi_lock_t                *cmd_handlers_lock;
     os_hnd_fd_id_t             *fd_wait_id;
-    ipmi_ll_event_handler_id_t *event_handlers;
-    ipmi_lock_t                *event_handlers_lock;
+    ipmi_lock_t                *smi_lock;
+    locked_list_t              *event_handlers;
 
     unsigned char              slave_addr;
 
     os_hnd_timer_id_t          *audit_timer;
     audit_timer_info_t         *audit_info;
 
-    ipmi_ll_con_changed_cb     con_change_handler;
-    void                       *con_change_cb_data;
+    /* Handles connection shutdown reporting. */
+    ipmi_ll_con_closed_cb close_done;
+    void                  *close_cb_data;
 
-    ipmi_ll_ipmb_addr_cb ipmb_addr_handler;
-    void                 *ipmb_addr_cb_data;
+    locked_list_t          *con_change_handlers;
+    locked_list_t          *ipmb_change_handlers;
 
     struct smi_data_s *next, *prev;
 } smi_data_t;
 
+static ipmi_lock_t *smi_list_lock = NULL;
 static smi_data_t *smi_list = NULL;
 
 /* Must be called with the ipmi read or write lock. */
@@ -150,14 +149,129 @@ static int smi_valid_ipmi(ipmi_con_t *ipmi)
 {
     smi_data_t *elem;
 
+    ipmi_lock(smi_list_lock);
     elem = smi_list;
     while ((elem) && (elem->ipmi != ipmi)) {
 	elem = elem->next;
     }
+    if (elem)
+	elem->refcount++;
+    ipmi_unlock(smi_list_lock);
 
     return (elem != NULL);
 }
 
+static void
+smi_cleanup(ipmi_con_t *ipmi)
+{
+    smi_data_t    *smi;
+    pending_cmd_t *cmd, *next_cmd;
+    cmd_handler_t *hnd_to_free, *next_hnd;
+    int           rv;
+
+    /* First order of business is to remove it from the SMI list. */
+    smi = (smi_data_t *) ipmi->con_data;
+
+    ipmi_lock(smi_list_lock);
+    if (smi->next)
+	smi->next->prev = smi->prev;
+    if (smi->prev)
+	smi->prev->next = smi->next;
+    else
+	smi_list = smi->next;
+    ipmi_unlock(smi_list_lock);
+
+    if (smi->close_done)
+	smi->close_done(ipmi, smi->close_cb_data);
+
+    cmd = smi->pending_cmds;
+    smi->pending_cmds = NULL;
+    while (cmd) {
+	ipmi_addr_t   *addr;
+	unsigned int  addr_len;
+	unsigned char data[1];
+	next_cmd = cmd->next;
+	if (cmd->rsp_handler) {
+	    if (cmd->use_orig_addr) {
+		addr = &cmd->orig_addr;
+		addr_len = cmd->orig_addr_len;
+	    } else {
+		addr = &cmd->addr;
+		addr_len = cmd->addr_len;
+	    }
+	    data[0] = IPMI_UNKNOWN_ERR_CC;
+	    
+	    cmd->msg.netfn |= 1;
+	    cmd->msg.data = data;
+	    cmd->msg.data_len = 1;
+	    ipmi_handle_rsp_item_copyall(ipmi, cmd->rsp_item,
+					 addr, addr_len, &cmd->msg,
+					 cmd->rsp_handler);
+	}
+	ipmi_mem_free(cmd);
+	cmd = next_cmd;
+    }
+
+    hnd_to_free = smi->cmd_handlers;
+    smi->cmd_handlers = NULL;
+    while (hnd_to_free) {
+	next_hnd = hnd_to_free->next;
+	ipmi_mem_free(hnd_to_free);
+	hnd_to_free = next_hnd;
+    }
+
+    if (smi->audit_info) {
+	rv = ipmi->os_hnd->stop_timer(ipmi->os_hnd, smi->audit_timer);
+	if (rv)
+	    smi->audit_info->cancelled = 1;
+	else {
+	    ipmi->os_hnd->free_timer(ipmi->os_hnd, smi->audit_timer);
+	    ipmi_mem_free(smi->audit_info);
+	}
+    }
+
+    if (ipmi->oem_data_cleanup)
+	ipmi->oem_data_cleanup(ipmi);
+    ipmi_con_attr_cleanup(ipmi);
+    if (smi->smi_lock)
+	ipmi_destroy_lock(smi->smi_lock);
+    if (smi->cmd_handlers_lock)
+	ipmi_destroy_lock(smi->cmd_handlers_lock);
+    if (smi->cmd_lock)
+	ipmi_destroy_lock(smi->cmd_lock);
+    if (smi->fd_wait_id)
+	ipmi->os_hnd->remove_fd_to_wait_for(ipmi->os_hnd, smi->fd_wait_id);
+    if (smi->con_change_handlers)
+	locked_list_destroy(smi->con_change_handlers);
+    if (smi->event_handlers)
+	locked_list_destroy(smi->event_handlers);
+    if (smi->ipmb_change_handlers)
+	locked_list_destroy(smi->ipmb_change_handlers);
+
+    /* Close the fd after we have deregistered it. */
+    close(smi->fd);
+
+    ipmi_mem_free(smi);
+    if (ipmi->name)
+	ipmi_mem_free(ipmi->name);
+    ipmi_mem_free(ipmi);
+}
+
+static void
+smi_put(ipmi_con_t *ipmi)
+{
+    smi_data_t *elem = ipmi->con_data;
+    int        done;
+
+    ipmi_lock(smi_list_lock);
+    elem->refcount--;
+    done = elem->refcount == 0;
+    ipmi_unlock(smi_list_lock);
+
+    if (done)
+	smi_cleanup(ipmi);
+}
+
 /* Must be called with cmd_lock held. */
 static void
 add_cmd(ipmi_con_t    *ipmi,
@@ -193,33 +307,6 @@ remove_cmd(ipmi_con_t    *ipmi,
 	smi->pending_cmds = cmd->next;
 }
 
-/* Must be called with event_lock held. */
-static void
-add_event_handler(ipmi_con_t                 *ipmi,
-		  smi_data_t                 *smi,
-		  ipmi_ll_event_handler_id_t *event)
-{
-    event->ipmi = ipmi;
-
-    event->next = smi->event_handlers;
-    event->prev = NULL;
-    if (smi->event_handlers)
-	smi->event_handlers->prev = event;
-    smi->event_handlers = event;
-}
-
-static void
-remove_event_handler(smi_data_t                 *smi,
-		     ipmi_ll_event_handler_id_t *event)
-{
-    if (event->next)
-	event->next->prev = event->prev;
-    if (event->prev)
-	event->prev->next = event->next;
-    else
-	smi->event_handlers = event->next;
-}
-
 static int
 add_cmd_registration(ipmi_con_t            *ipmi,
 		     unsigned char         netfn,
@@ -344,17 +431,19 @@ smi_send(smi_data_t   *smi,
     ipmi_addr_t myaddr;
 
     if (DEBUG_MSG) {
+	char buf1[32], buf2[32];
 	ipmi_log(IPMI_LOG_DEBUG_START, "outgoing msgid=%08lx\n addr =", msgid);
 	dump_hex((unsigned char *) addr, addr_len);
         ipmi_log(IPMI_LOG_DEBUG_CONT,
                  "\n msg  = netfn=%s cmd=%s data_len=%d.",
-		 ipmi_get_netfn_string(msg->netfn),
-                 ipmi_get_command_string(msg->netfn, msg->cmd), msg->data_len);
+		 ipmi_get_netfn_string(msg->netfn, buf1, 32),
+                 ipmi_get_command_string(msg->netfn, msg->cmd, buf2, 32),
+		 msg->data_len);
 	if ( msg->data_len ) {
 	        ipmi_log(IPMI_LOG_DEBUG_CONT, "\n data =\n  ");
 	        dump_hex((unsigned char *)msg->data, msg->data_len);
 	}
-	ipmi_log(IPMI_LOG_DEBUG_END, "\n");
+	ipmi_log(IPMI_LOG_DEBUG_END, " ");
     }
 
     if ((addr->addr_type == IPMI_IPMB_BROADCAST_ADDR_TYPE)
@@ -368,7 +457,7 @@ smi_send(smi_data_t   *smi,
     }
 
     if (msg->data_len > IPMI_MAX_MSG_LENGTH)
-	return EMSGSIZE;
+	return EBADF;
 
     if (smi->using_socket) {
 	struct sockaddr_ipmi     saddr;
@@ -422,6 +511,67 @@ set_ipmb_in_dev(smi_data_t *smi)
     }
 }
 
+typedef struct call_ipmb_change_handler_s
+{
+    smi_data_t  *smi;
+    int          err;
+    unsigned int ipmb_addr;
+    int          active;
+    unsigned int hacks;
+} call_ipmb_change_handler_t;
+
+static int
+call_ipmb_change_handler(void *cb_data, void *item1, void *item2)
+{
+    call_ipmb_change_handler_t *info = cb_data;
+    ipmi_ll_ipmb_addr_cb       handler = item1;
+
+    handler(info->smi->ipmi, info->err, info->ipmb_addr, info->active,
+	    info->hacks, item2);
+    return LOCKED_LIST_ITER_CONTINUE;
+}
+
+static void
+call_ipmb_change_handlers(smi_data_t *smi, int err, unsigned int ipmb_addr,
+			  int active, unsigned int hacks)
+{
+    call_ipmb_change_handler_t info;
+
+    info.smi = smi;
+    info.err = err;
+    info.ipmb_addr = ipmb_addr;
+    info.active = active;
+    info.hacks = hacks;
+    locked_list_iterate(smi->ipmb_change_handlers, call_ipmb_change_handler,
+			&info);
+}
+
+static int
+smi_add_ipmb_addr_handler(ipmi_con_t           *ipmi,
+			  ipmi_ll_ipmb_addr_cb handler,
+			  void                 *cb_data)
+{
+    smi_data_t *smi = (smi_data_t *) ipmi->con_data;
+
+    if (locked_list_add(smi->ipmb_change_handlers, handler, cb_data))
+	return 0;
+    else
+	return ENOMEM;
+}
+
+static int
+smi_remove_ipmb_addr_handler(ipmi_con_t           *ipmi,
+			     ipmi_ll_ipmb_addr_cb handler,
+			     void                 *cb_data)
+{
+    smi_data_t *smi = (smi_data_t *) ipmi->con_data;
+
+    if (locked_list_remove(smi->ipmb_change_handlers, handler, cb_data))
+	return 0;
+    else
+	return EINVAL;
+}
+
 static void
 ipmb_handler(ipmi_con_t   *ipmi,
 	     int          err,
@@ -437,25 +587,12 @@ ipmb_handler(ipmi_con_t   *ipmi,
 
     if (ipmb != smi->slave_addr) {
 	smi->slave_addr = ipmb;
-	if (smi->ipmb_addr_handler)
-	    smi->ipmb_addr_handler(ipmi, err, ipmb, active, 0,
-				   smi->ipmb_addr_cb_data);
-
+	ipmi->ipmb_addr = ipmb;
+	call_ipmb_change_handlers(smi, err, ipmb, active, 0);
 	set_ipmb_in_dev(smi);
     }
 }
 
-static void
-smi_set_ipmb_addr_handler(ipmi_con_t           *ipmi,
-			  ipmi_ll_ipmb_addr_cb handler,
-			  void                 *cb_data)
-{
-    smi_data_t *smi = (smi_data_t *) ipmi->con_data;
-
-    smi->ipmb_addr_handler = handler;
-    smi->ipmb_addr_cb_data = cb_data;
-}
-
 static void
 audit_timeout_handler(void              *cb_data,
 		      os_hnd_timer_id_t *id)
@@ -473,10 +610,8 @@ audit_timeout_handler(void              *cb_data,
 	goto out_done;
     }
 
-    ipmi_read_lock();
-
     if (!smi_valid_ipmi(ipmi)) {
-	goto out_unlock_done;
+	goto out_done;
     }
 
     smi = ipmi->con_data;
@@ -497,7 +632,7 @@ audit_timeout_handler(void              *cb_data,
 	si.lun = 0;
 	ipmi->send_command(ipmi,
 			   (ipmi_addr_t *) &si, sizeof(si),
-			   &msg, NULL, NULL, NULL, NULL, NULL);
+			   &msg, NULL, NULL);
     }
 
     timeout.tv_sec = SMI_AUDIT_TIMEOUT / 1000000;
@@ -511,8 +646,8 @@ audit_timeout_handler(void              *cb_data,
     /* Make sure the timer info doesn't get freed. */
     info = NULL;
 
- out_unlock_done:
-    ipmi_read_unlock();
+    smi_put(ipmi);
+
  out_done:
     if (info) {
 	ipmi->os_hnd->free_timer(ipmi->os_hnd, id);
@@ -527,11 +662,7 @@ handle_response(ipmi_con_t *ipmi, struct ipmi_recv *recv)
     smi_data_t            *smi = (smi_data_t *) ipmi->con_data;
     pending_cmd_t         *cmd, *finder;
     ipmi_ll_rsp_handler_t rsp_handler;
-    void                  *rsp_data;
-    void                  *data2, *data3, *data4;
-    ipmi_addr_t           *addr;
-    unsigned int          addr_len;
-    ipmi_addr_t           tmp_addr;
+    ipmi_msgi_t           *rspi;
 
     cmd = (pending_cmd_t *) recv->msgid;
     
@@ -551,10 +682,7 @@ handle_response(ipmi_con_t *ipmi, struct ipmi_recv *recv)
 
     /* Extract everything we need from the command here. */
     rsp_handler = cmd->rsp_handler;
-    rsp_data = cmd->rsp_data;
-    data2 = cmd->data2;
-    data3 = cmd->data3;
-    data4 = cmd->data4;
+    rspi = cmd->rsp_item;
 
     remove_cmd(ipmi, smi, cmd);
 
@@ -563,67 +691,126 @@ handle_response(ipmi_con_t *ipmi, struct ipmi_recv *recv)
     if (cmd->use_orig_addr) {
 	/* We did an address translation, make sure the address is the one
 	   that was previously provided. */
-	addr = &tmp_addr;
-	addr_len = cmd->orig_addr_len;
-	memcpy(addr, &cmd->orig_addr, addr_len);
+	memcpy(&rspi->addr, &cmd->orig_addr, cmd->orig_addr_len);
+	rspi->addr_len = cmd->orig_addr_len;
     } else {
-	addr = (ipmi_addr_t *) recv->addr;
-	addr_len = recv->addr_len;
+	memcpy(&rspi->addr, (ipmi_addr_t *) recv->addr, recv->addr_len);
+	rspi->addr_len = recv->addr_len;
     }
 
     ipmi_mem_free(cmd);
     cmd = NULL; /* It's gone after this point. */
 
-    /* call the user handler. */
-    if (rsp_handler)
-	rsp_handler(ipmi, addr, addr_len, &(recv->msg),
-		    rsp_data, data2, data3, data4);
+    ipmi_handle_rsp_item_copymsg(ipmi, rspi, &recv->msg, rsp_handler);
+
     return;
 
  out_unlock:
     ipmi_unlock(smi->cmd_lock);
 }
 
+typedef struct call_event_handler_s
+{
+    smi_data_t   *smi;
+    ipmi_addr_t  *addr;
+    unsigned int addr_len;
+    ipmi_event_t *event;
+} call_event_handler_t;
+
+static int
+call_event_handler(void *cb_data, void *item1, void *item2)
+{
+    call_event_handler_t  *info = cb_data;
+    ipmi_ll_evt_handler_t handler = item1;
+
+    handler(info->smi->ipmi, info->addr, info->addr_len, info->event, item2);
+    return LOCKED_LIST_ITER_CONTINUE;
+}
+
+static int
+smi_add_event_handler(ipmi_con_t            *ipmi,
+		      ipmi_ll_evt_handler_t handler,
+		      void                  *cb_data)
+{
+    smi_data_t *smi = (smi_data_t *) ipmi->con_data;
+    int        rv = 0;
+
+    ipmi_lock(smi->smi_lock);
+    if (! locked_list_add(smi->event_handlers, handler, cb_data))
+	rv = ENOMEM;
+    if (!rv) {
+	if (locked_list_num_entries(smi->event_handlers) == 1) {
+	    int val = 1;
+	    if (smi->using_socket)
+		rv = ioctl(smi->fd, SIOCIPMIGETEVENT, &val);
+	    else
+		rv = ioctl(smi->fd, IPMICTL_SET_GETS_EVENTS_CMD, &val);
+	    if (rv == -1) {
+		locked_list_remove(smi->event_handlers, handler, cb_data);
+		rv = errno;
+	    }
+	}
+    }
+    ipmi_unlock(smi->smi_lock);
+    return rv;
+}
+
+static int
+smi_remove_event_handler(ipmi_con_t            *ipmi,
+			 ipmi_ll_evt_handler_t handler,
+			 void                  *cb_data)
+{
+    smi_data_t *smi = (smi_data_t *) ipmi->con_data;
+    int        rv = 0;
+
+    ipmi_lock(smi->smi_lock);
+    if (! locked_list_remove(smi->event_handlers, handler, cb_data))
+	rv = EINVAL;
+    if (locked_list_num_entries(smi->event_handlers) == 0) {
+	int val = 0;
+	if (smi->using_socket)
+	    ioctl(smi->fd, SIOCIPMIGETEVENT, &val);
+	else
+	    ioctl(smi->fd, IPMICTL_SET_GETS_EVENTS_CMD, &val);
+    }
+    ipmi_unlock(smi->smi_lock);
+    return rv;
+}
+
 static ipmi_mcid_t invalid_mcid = IPMI_MCID_INVALID;
 
 static void
-handle_async_event(ipmi_con_t *ipmi, struct ipmi_recv *recv)
+handle_async_event(ipmi_con_t  *ipmi,
+		   ipmi_addr_t *addr,
+		   unsigned int addr_len,
+		   ipmi_msg_t   *msg)
 {
-    smi_data_t                 *smi = (smi_data_t *) ipmi->con_data;
-    ipmi_ll_event_handler_id_t *elem, *next;
-    ipmi_event_t               *event;
-    ipmi_time_t                timestamp;
-    unsigned int               type = recv->msg.data[2];
-    unsigned int               record_id = ipmi_get_uint16(recv->msg.data);
+    smi_data_t           *smi = (smi_data_t *) ipmi->con_data;
+    ipmi_event_t         *event;
+    ipmi_time_t          timestamp;
+    unsigned int         type = msg->data[2];
+    unsigned int         record_id = ipmi_get_uint16(msg->data);
+    call_event_handler_t info;
 
     if (type < 0xe0)
-	timestamp = ipmi_seconds_to_time(ipmi_get_uint32(recv->msg.data+3));
+	timestamp = ipmi_seconds_to_time(ipmi_get_uint32(msg->data+3));
     else
 	timestamp = -1;
     event = ipmi_event_alloc(invalid_mcid,
 			     record_id,
 			     type,
 			     timestamp,
-			     recv->msg.data+3, 13);
+			     msg->data+3, 13);
     if (!event)
 	/* We missed it here, but the SEL fetch should catch it later. */
 	return;
 
-    ipmi_lock(smi->event_handlers_lock);
-    elem = smi->event_handlers;
-    while (elem != NULL) {
-	/* Fetch the next element now, so the user can delete the
-           current one. */
-	next = elem->next;
+    info.smi = smi;
+    info.addr = addr;
+    info.addr_len = addr_len;
+    info.event = event;
+    locked_list_iterate(smi->event_handlers, call_event_handler, &info);
 
-	/* call the user handler. */
-	elem->handler(ipmi,
-		      (ipmi_addr_t *) recv->addr, recv->addr_len,
-		      event, elem->event_data, elem->data2);
-
-	elem = next;
-    }
-    ipmi_unlock(smi->event_handlers_lock);
     ipmi_event_free(event);
 }
 
@@ -673,17 +860,22 @@ static void
 gen_recv_msg(ipmi_con_t *ipmi, struct ipmi_recv *recv)
 {
     if (DEBUG_MSG) {
+	char buf1[32], buf2[32], buf3[32];
 	ipmi_log(IPMI_LOG_DEBUG_START, "incoming msgid=%08lx\n addr =",
 		 recv->msgid);
 	dump_hex((unsigned char *) recv->addr, recv->addr_len);
         ipmi_log(IPMI_LOG_DEBUG_CONT,
                  "\n msg  = netfn=%s cmd=%s data_len=%d. cc=%s",
-		 ipmi_get_netfn_string(recv->msg.netfn),
-                 ipmi_get_command_string(recv->msg.netfn, recv->msg.cmd), recv->msg.data_len,
-		 ipmi_get_cc_string(recv->msg.data[0]));
-	ipmi_log(IPMI_LOG_DEBUG_CONT, "\n data =\n  ");
-	dump_hex(recv->msg.data, recv->msg.data_len);
-	ipmi_log(IPMI_LOG_DEBUG_END, "\n");
+		 ipmi_get_netfn_string(recv->msg.netfn, buf1, 32),
+                 ipmi_get_command_string(recv->msg.netfn, recv->msg.cmd,
+					 buf2, 32),
+		 recv->msg.data_len,
+		 ipmi_get_cc_string(recv->msg.data[0], buf3, 32));
+	if (recv->msg.data_len) {
+	    ipmi_log(IPMI_LOG_DEBUG_CONT, "\n data =\n  ");
+	    dump_hex(recv->msg.data, recv->msg.data_len);
+	}
+	ipmi_log(IPMI_LOG_DEBUG_END, " ");
     }
 
     switch (recv->recv_type) {
@@ -692,7 +884,8 @@ gen_recv_msg(ipmi_con_t *ipmi, struct ipmi_recv *recv)
 	    break;
 
 	case IPMI_ASYNC_EVENT_RECV_TYPE:
-	    handle_async_event(ipmi, recv);
+	    handle_async_event(ipmi, (ipmi_addr_t *) recv->addr,
+			       recv->addr_len, &recv->msg);
 	    break;
 
 	case IPMI_CMD_RECV_TYPE:
@@ -715,12 +908,10 @@ ipmi_dev_data_handler(int            fd,
     struct ipmi_recv recv;
     int              rv;
 
-    ipmi_read_lock();
-
     if (!smi_valid_ipmi(ipmi)) {
 	/* We can have due to a race condition, just return and
            everything should be fine. */
-	goto out_unlock2;
+	return;
     }
 
     recv.msg.data = data;
@@ -734,13 +925,13 @@ ipmi_dev_data_handler(int            fd,
 	    data[0] = IPMI_REQUESTED_DATA_LENGTH_EXCEEDED_CC;
 	    rv = 0;
 	} else
-	    goto out_unlock2;
+	    goto out;
     }
 
     gen_recv_msg(ipmi, &recv);
 
- out_unlock2:
-    ipmi_read_unlock();
+ out:
+    smi_put(ipmi);
 }
 
 static void
@@ -756,12 +947,10 @@ ipmi_sock_data_handler(int            fd,
     int                  rv;
     struct ipmi_recv     recv;
 
-    ipmi_read_lock();
-
     if (!smi_valid_ipmi(ipmi)) {
 	/* We can have due to a race condition, just return and
            everything should be fine. */
-	goto out_unlock2;
+	return;
     }
 
     addr_len = sizeof(addr);
@@ -770,38 +959,40 @@ ipmi_sock_data_handler(int            fd,
     if (rv == -1) {
 	/* FIXME - no handling for EMSGSIZE. */
 	if (errno == EINTR) {
-	    goto out_unlock2; /* Try again later. */
+	    goto out; /* Try again later. */
 	} else {
 	    ipmi_log(IPMI_LOG_SEVERE,
 		     "%sipmi_smi.c(ipmi_sock_data_handler): "
-		     "Error receiving message: %s\n",
+		     "Error receiving message: %s",
 		     IPMI_CONN_NAME(ipmi), strerror(errno));
-	    goto out_unlock2;
+	    goto out;
 	}
     }
     if (DEBUG_MSG) {
 	ipmi_log(IPMI_LOG_DEBUG_START, "incoming\n addr(%d.) = ", addr_len);
 	dump_hex((unsigned char *) &addr, MIN(addr_len,sizeof(addr)));
-	ipmi_log(IPMI_LOG_DEBUG_CONT, "\n data(%d.) =\n  ", rv);
-	dump_hex(data, rv);
-	ipmi_log(IPMI_LOG_DEBUG_END, "\n");
+	if (rv) {
+	    ipmi_log(IPMI_LOG_DEBUG_CONT, "\n data(%d.) =\n  ", rv);
+	    dump_hex(data, rv);
+	}
+	ipmi_log(IPMI_LOG_DEBUG_END, " ");
     }
 
     if (rv < sizeof(*smsg)) {
 	ipmi_log(IPMI_LOG_SEVERE,
 		 "%sipmi_smi.c(ipmi_sock_data_handler): "
-		 "Undersized socket message: %d bytes\n",
+		 "Undersized socket message: %d bytes",
 		 IPMI_CONN_NAME(ipmi), rv);
-	goto out_unlock2;
+	goto out;
     }
 
     smsg = (struct ipmi_sock_msg *) data;
     if (rv < (sizeof(*smsg) + smsg->data_len)) {
 	ipmi_log(IPMI_LOG_SEVERE,
 		 "%sipmi_smi.c(ipmi_sock_data_handler): "
-		 "Got subsized msg, size was %d, data_len was %d\n",
+		 "Got subsized msg, size was %d, data_len was %d",
 		 IPMI_CONN_NAME(ipmi), rv, smsg->data_len);
-	return;
+	goto out;
     }
 
     recv.recv_type = smsg->recv_type;
@@ -815,8 +1006,8 @@ ipmi_sock_data_handler(int            fd,
 
     gen_recv_msg(ipmi, &recv);
 
- out_unlock2:
-    ipmi_read_unlock();
+ out:
+    smi_put(ipmi);
 }
 
 static int
@@ -825,22 +1016,27 @@ smi_send_command(ipmi_con_t            *ipmi,
 		 unsigned int          addr_len,
 		 ipmi_msg_t            *msg,
 		 ipmi_ll_rsp_handler_t rsp_handler,
-		 void                  *rsp_data,
-		 void                  *data2,
-		 void                  *data3,
-		 void                  *data4)
+		 ipmi_msgi_t           *trspi)
 {
     pending_cmd_t *cmd;
     smi_data_t    *smi;
     int           rv;
     ipmi_addr_t   tmp_addr;
+    ipmi_msgi_t   *rspi = trspi;
 
 
+    if (addr_len > sizeof(ipmi_addr_t))
+	return EINVAL;
+
+    if (msg->data_len > IPMI_MAX_MSG_LENGTH)
+	return EINVAL;
+
     smi = (smi_data_t *) ipmi->con_data;
 
-    if (addr_len > sizeof(ipmi_addr_t)) {
-	rv = EINVAL;
-	goto out_unlock2;
+    if (!rspi) {
+	rspi = ipmi_alloc_msg_item();
+	if (!rspi)
+	    return ENOMEM;
     }
 
     cmd = ipmi_mem_alloc(sizeof(*cmd));
@@ -878,10 +1074,7 @@ smi_send_command(ipmi_con_t            *ipmi,
     /* Put it in the list first. */
     cmd->msg = *msg;
     cmd->rsp_handler = rsp_handler;
-    cmd->rsp_data = rsp_data;
-    cmd->data2 = data2;
-    cmd->data3 = data3;
-    cmd->data4 = data4;
+    cmd->rsp_item = rspi;
 
     ipmi_lock(smi->cmd_lock);
     add_cmd(ipmi, addr, addr_len, msg, smi, cmd);
@@ -889,101 +1082,18 @@ smi_send_command(ipmi_con_t            *ipmi,
     rv = smi_send(smi, smi->fd, addr, addr_len, msg, (long) cmd);
     if (rv) {
 	remove_cmd(ipmi, smi, cmd);
+	ipmi_mem_free(cmd);
 	goto out_unlock;
     }
 
  out_unlock:
     ipmi_unlock(smi->cmd_lock);
  out_unlock2:
-    return rv;
-}
-
-static int
-smi_register_for_events(ipmi_con_t                 *ipmi,
-			ipmi_ll_evt_handler_t      handler,
-			void                       *event_data,
-			void                       *data2,
-			ipmi_ll_event_handler_id_t **id)
-{
-    smi_data_t                 *smi;
-    int                        rv = 0;
-    int                        was_empty;
-    ipmi_ll_event_handler_id_t *entry;
-
-    smi = (smi_data_t *) ipmi->con_data;
-
-    entry = ipmi_mem_alloc(sizeof(*entry));
-    if (!entry) {
-	rv = ENOMEM;
-	goto out_unlock2;
+    if (rv) {
+	/* If we allocated an rspi, free it. */
+	if (!trspi && rspi)
+	    ipmi_free_msg_item(rspi);
     }
-
-    entry->handler = handler;
-    entry->event_data = event_data;
-    entry->data2 = data2;
-
-    ipmi_lock(smi->event_handlers_lock);
-    was_empty = smi->event_handlers == NULL;
-
-    add_event_handler(ipmi, smi, entry);
-
-    if (was_empty) {
-	int val = 1;
-	if (smi->using_socket)
-	    rv = ioctl(smi->fd, SIOCIPMIGETEVENT, &val);
-	else
-	    rv = ioctl(smi->fd, IPMICTL_SET_GETS_EVENTS_CMD, &val);
-	if (rv == -1) {
-	    remove_event_handler(smi, entry);
-	    rv = errno;
-	    goto out_unlock;
-	}
-    }
-
-    if (id)
-	*id = entry;
-
- out_unlock:
-    ipmi_unlock(smi->event_handlers_lock);
- out_unlock2:
-    return rv;
-}
-
-static int
-smi_deregister_for_events(ipmi_con_t                 *ipmi,
-			  ipmi_ll_event_handler_id_t *id)
-{
-    smi_data_t *smi;
-    int        rv = 0;
-
-    smi = (smi_data_t *) ipmi->con_data;
-
-    if (id->ipmi != ipmi) {
-	rv = EINVAL;
-	goto out_unlock2;
-    }
-
-    ipmi_lock(smi->event_handlers_lock);
-
-    remove_event_handler(smi, id);
-    id->ipmi = NULL;
-
-    if (smi->event_handlers == NULL) {
-	int val = 0;
-	if (smi->using_socket)
-	    rv = ioctl(smi->fd, SIOCIPMIGETEVENT, &val);
-	else
-	    rv = ioctl(smi->fd, IPMICTL_SET_GETS_EVENTS_CMD, &val);
-	if (rv == -1) {
-	    rv = errno;
-	    goto out_unlock;
-	}
-    }
-
- out_unlock:
-    ipmi_unlock(smi->event_handlers_lock);
- out_unlock2:
-
     return rv;
 }
 
@@ -1052,7 +1162,7 @@ smi_deregister_for_command(ipmi_con_t    *ipmi,
     reg.netfn = netfn;
     reg.cmd = cmd;
     if (smi->using_socket)
-	rv = ioctl(smi->fd, SIOCIPMIREGCMD, &reg);
+	rv = ioctl(smi->fd, SIOCIPMIUNREGCMD, &reg);
     else
 	rv = ioctl(smi->fd, IPMICTL_UNREGISTER_FOR_CMD, &reg);
     if (rv == -1) {
@@ -1068,110 +1178,30 @@ smi_deregister_for_command(ipmi_con_t    *ipmi,
 }
 
 static int
-smi_close_connection(ipmi_con_t *ipmi)
+smi_close_connection_done(ipmi_con_t *ipmi,
+			  ipmi_ll_con_closed_cb handler,
+			  void                  *cb_data)
 {
-    smi_data_t                 *smi;
-    pending_cmd_t              *cmd, *next_cmd;
-    cmd_handler_t              *hnd_to_free, *next_hnd;
-    ipmi_ll_event_handler_id_t *evt_to_free, *next_evt;
-    int                        rv;
+    smi_data_t *smi;
 
     if (! smi_valid_ipmi(ipmi)) {
 	return EINVAL;
     }
 
-    /* First order of business is to remove it from the SMI list. */
     smi = (smi_data_t *) ipmi->con_data;
+    smi->close_done = handler;
+    smi->close_cb_data = cb_data;
 
-    if (smi->next)
-	smi->next->prev = smi->prev;
-    if (smi->prev)
-	smi->prev->next = smi->next;
-    else
-	smi_list = smi->next;
-
-    /* After this point no other operations can occur on this ipmi
-       interface, so it's safe. */
-
-    cmd = smi->pending_cmds;
-    smi->pending_cmds = NULL;
-    while (cmd) {
-	ipmi_addr_t   *addr;
-	unsigned int  addr_len;
-	unsigned char data[1];
-	next_cmd = cmd->next;
-	if (cmd->rsp_handler) {
-	    if (cmd->use_orig_addr) {
-		addr = &cmd->orig_addr;
-		addr_len = cmd->orig_addr_len;
-	    } else {
-		addr = &cmd->addr;
-		addr_len = cmd->addr_len;
-	    }
-	    data[0] = IPMI_UNKNOWN_ERR_CC;
-	    
-	    cmd->msg.netfn |= 1;
-	    cmd->msg.data = data;
-	    cmd->msg.data_len = 1;
-	    cmd->rsp_handler(NULL, addr, addr_len, &cmd->msg, cmd->rsp_data,
-			     cmd->data2, cmd->data3, cmd->data4);
-	}
-	ipmi_mem_free(cmd);
-	cmd = next_cmd;
-    }
-
-    hnd_to_free = smi->cmd_handlers;
-    smi->cmd_handlers = NULL;
-    while (hnd_to_free) {
-	next_hnd = hnd_to_free->next;
-	ipmi_mem_free(hnd_to_free);
-	hnd_to_free = next_hnd;
-    }
-
-    evt_to_free = smi->event_handlers;
-    smi->event_handlers = NULL;
-    while (evt_to_free) {
-	evt_to_free->ipmi = NULL;
-	next_evt = evt_to_free->next;
-	ipmi_mem_free(evt_to_free);
-	evt_to_free = next_evt;
-    }
-
-    if (smi->audit_info) {
-	rv = ipmi->os_hnd->stop_timer(ipmi->os_hnd, smi->audit_timer);
-	if (rv)
-	    smi->audit_info->cancelled = 1;
-	else {
-	    ipmi->os_hnd->free_timer(ipmi->os_hnd, smi->audit_timer);
-	    ipmi_mem_free(smi->audit_info);
-	}
-    }
-
-    if (ipmi->oem_data_cleanup)
-	ipmi->oem_data_cleanup(ipmi);
-    if (smi->event_handlers_lock)
-	ipmi_destroy_lock(smi->event_handlers_lock);
-    if (smi->cmd_handlers_lock)
-	ipmi_destroy_lock(smi->cmd_handlers_lock);
-    if (smi->cmd_lock)
-	ipmi_destroy_lock(smi->cmd_lock);
-    if (smi->fd_wait_id)
-	ipmi->os_hnd->remove_fd_to_wait_for(ipmi->os_hnd, smi->fd_wait_id);
-
-    /* Close the fd after we have deregistered it. */
-    close(smi->fd);
-
-    ipmi_mem_free(smi);
-    ipmi_mem_free(ipmi);
-
+    smi_put(ipmi);
+    smi_put(ipmi);
     return 0;
 }
 
-static ll_ipmi_t smi_ll_ipmi =
+static int
+smi_close_connection(ipmi_con_t *ipmi)
 {
-    .valid_ipmi = smi_valid_ipmi,
-    .registered = 0
-};
+    return smi_close_connection_done(ipmi, NULL, NULL);
+}
 
 static void
 cleanup_con(ipmi_con_t *ipmi)
@@ -1180,12 +1210,17 @@ cleanup_con(ipmi_con_t *ipmi)
     os_handler_t *handlers = ipmi->os_hnd;
 
     if (ipmi) {
+	ipmi_con_attr_cleanup(ipmi);
 	ipmi_mem_free(ipmi);
+	if (ipmi->name) {
+	    ipmi_mem_free(ipmi->name);
+	    ipmi->name = NULL;
+	}
     }
 
     if (smi) {
-	if (smi->event_handlers_lock)
-	    ipmi_destroy_lock(smi->event_handlers_lock);
+	if (smi->smi_lock)
+	    ipmi_destroy_lock(smi->smi_lock);
 	if (smi->cmd_handlers_lock)
 	    ipmi_destroy_lock(smi->cmd_handlers_lock);
 	if (smi->cmd_lock)
@@ -1194,20 +1229,72 @@ cleanup_con(ipmi_con_t *ipmi)
 	    close(smi->fd);
 	if (smi->fd_wait_id)
 	    handlers->remove_fd_to_wait_for(ipmi->os_hnd, smi->fd_wait_id);
+	if (smi->con_change_handlers)
+	    locked_list_destroy(smi->con_change_handlers);
+	if (smi->event_handlers)
+	    locked_list_destroy(smi->event_handlers);
+	if (smi->ipmb_change_handlers)
+	    locked_list_destroy(smi->ipmb_change_handlers);
 	ipmi_mem_free(smi);
     }
 }
 
+typedef struct call_con_change_handler_s
+{
+    smi_data_t  *smi;
+    int          err;
+    unsigned int port;
+    int          any_port_up;
+} call_con_change_handler_t;
+
+static int
+call_con_change_handler(void *cb_data, void *item1, void *item2)
+{
+    call_con_change_handler_t  *info = cb_data;
+    ipmi_ll_con_changed_cb     handler = item1;
+
+    handler(info->smi->ipmi, info->err, info->port, info->any_port_up, item2);
+    return LOCKED_LIST_ITER_CONTINUE;
+}
+
 static void
-smi_set_con_change_handler(ipmi_con_t             *ipmi,
+call_con_change_handlers(smi_data_t *smi, int err, unsigned int port,
+			 int any_port_up)
+{
+    call_con_change_handler_t info;
+
+    info.smi = smi;
+    info.err = err;
+    info.port = port;
+    info.any_port_up = any_port_up;
+    locked_list_iterate(smi->con_change_handlers, call_con_change_handler,
+			&info);
+}
+
+static int
+smi_add_con_change_handler(ipmi_con_t             *ipmi,
 			   ipmi_ll_con_changed_cb handler,
 			   void                   *cb_data)
 {
-    smi_data_t *smi = ipmi->con_data;
+    smi_data_t *smi = (smi_data_t *) ipmi->con_data;
 
-    smi->con_change_handler = handler;
-    smi->con_change_cb_data = cb_data;
-    return;
+    if (locked_list_add(smi->con_change_handlers, handler, cb_data))
+	return 0;
+    else
+	return ENOMEM;
+}
+
+static int
+smi_remove_con_change_handler(ipmi_con_t             *ipmi,
+			      ipmi_ll_con_changed_cb handler,
+			      void                   *cb_data)
+{
+    smi_data_t *smi = (smi_data_t *) ipmi->con_data;
+
+    if (locked_list_remove(smi->con_change_handlers, handler, cb_data))
+	return 0;
+    else
+	return EINVAL;
 }
 
 static void
@@ -1218,8 +1305,7 @@ finish_start_con(void *cb_data, os_hnd_timer_id_t *id)
 
     ipmi->os_hnd->free_timer(ipmi->os_hnd, id);
 
-    if (smi->con_change_handler)
-	smi->con_change_handler(ipmi, 0, 1, 1, smi->con_change_cb_data);
+    call_con_change_handlers(smi, 0, 1, 1);
 }
 
 static void
@@ -1232,10 +1318,8 @@ smi_set_ipmb_addr(ipmi_con_t    *ipmi,
 
     if (smi->slave_addr != ipmb) {
 	smi->slave_addr = ipmb;
-	if (smi->ipmb_addr_handler)
-	    smi->ipmb_addr_handler(ipmi, 0, ipmb, active, 0,
-				   smi->ipmb_addr_cb_data);
-
+	ipmi->ipmb_addr = ipmb;
+	call_ipmb_change_handlers(smi, 0, ipmb, active, 0);
 	set_ipmb_in_dev(smi);
     }
 }
@@ -1268,8 +1352,7 @@ finish_connection(ipmi_con_t *ipmi, smi_data_t *smi)
     return;
 
  out_err:
-    if (smi->con_change_handler)
-	smi->con_change_handler(ipmi, err, 0, 0, smi->con_change_cb_data);
+    call_con_change_handlers(smi, err, 0, 0);
 }
 
 static void
@@ -1283,30 +1366,21 @@ handle_ipmb_addr(ipmi_con_t   *ipmi,
     smi_data_t *smi = (smi_data_t *) ipmi->con_data;
 
     if (err) {
-	if (smi->con_change_handler)
-	    smi->con_change_handler(ipmi, err, 0, 0, smi->con_change_cb_data);
+	call_con_change_handlers(smi, err, 0, 0);
 	return;
     }
 
     smi->slave_addr = ipmb_addr;
+    ipmi->ipmb_addr = ipmb_addr;
     finish_connection(ipmi, smi);
-    if (smi->ipmb_addr_handler)
-	smi->ipmb_addr_handler(ipmi, err, ipmb_addr, active, 0,
-			       smi->ipmb_addr_cb_data);
-
+    call_ipmb_change_handlers(smi, err, ipmb_addr, active, 0);
     set_ipmb_in_dev(smi);
 }
 
-static void
-handle_dev_id(ipmi_con_t   *ipmi,
-	      ipmi_addr_t  *addr,
-	      unsigned int addr_len,
-	      ipmi_msg_t   *msg,
-	      void         *rsp_data1,
-	      void         *rsp_data2,
-	      void         *rsp_data3,
-	      void         *rsp_data4)
+static int
+handle_dev_id(ipmi_con_t *ipmi, ipmi_msgi_t *msgi)
 {
+    ipmi_msg_t        *msg = &msgi->msg;
     smi_data_t        *smi = (smi_data_t *) ipmi->con_data;
     int               err;
     unsigned int      manufacturer_id;
@@ -1338,11 +1412,11 @@ handle_dev_id(ipmi_con_t   *ipmi,
 	    goto out_err;
     } else
 	finish_connection(ipmi, smi);
-    return;
+    return IPMI_MSG_ITEM_NOT_USED;
 
  out_err:
-    if (smi->con_change_handler)
-	smi->con_change_handler(ipmi, err, 0, 0, smi->con_change_cb_data);
+    call_con_change_handlers(smi, err, 0, 0);
+    return IPMI_MSG_ITEM_NOT_USED;
 }
 
 static void
@@ -1362,10 +1436,9 @@ smi_oem_done(ipmi_con_t *ipmi, void *cb_data)
     si.channel = 0xf;
     si.lun = 0;
     rv = smi_send_command(ipmi, (ipmi_addr_t *) &si, sizeof(si), &msg,
-			  handle_dev_id, NULL, NULL, NULL, NULL);
+			  handle_dev_id, NULL);
     if (rv) {
-	if (smi->con_change_handler)
-	    smi->con_change_handler(ipmi, rv, 0, 0, smi->con_change_cb_data);
+	call_con_change_handlers(smi, rv, 0, 0);
     }
 }
 
@@ -1419,9 +1492,6 @@ setup(int          if_num,
     smi_data_t *smi = NULL;
     int        rv;
 
-    /* Make sure we register before anything else. */
-    ipmi_register_ll(&smi_ll_ipmi);
-
     /* Keep things sane. */
     if (if_num >= 100)
 	return EINVAL;
@@ -1433,6 +1503,13 @@ setup(int          if_num,
 
     ipmi->user_data = user_data;
     ipmi->os_hnd = handlers;
+    ipmi->con_type = "smi";
+    ipmi->priv_level = IPMI_PRIVILEGE_ADMIN; /* Always admin privilege. */
+    ipmi->ipmb_addr = 0x20; /* Assume this until told otherwise */
+
+    rv = ipmi_con_attr_init(ipmi);
+    if (rv)
+	goto out_err;
 
     smi = ipmi_mem_alloc(sizeof(*smi));
     if (!smi) {
@@ -1443,15 +1520,9 @@ setup(int          if_num,
 
     ipmi->con_data = smi;
 
+    smi->refcount = 1;
     smi->ipmi = ipmi;
     smi->slave_addr = 0x20; /* Assume this until told otherwise. */
-    smi->pending_cmds = NULL;
-    smi->cmd_lock = NULL;
-    smi->cmd_handlers = NULL;
-    smi->cmd_handlers_lock = NULL;
-    smi->event_handlers = NULL;
-    smi->event_handlers_lock = NULL;
-    smi->fd_wait_id = NULL;
 
     smi->fd = open_smi_fd(if_num, &smi->using_socket);
     if (smi->fd == -1) {
@@ -1459,6 +1530,24 @@ setup(int          if_num,
 	goto out_err;
     }
 
+    smi->con_change_handlers = locked_list_alloc(handlers);
+    if (!smi->con_change_handlers) {
+	rv = ENOMEM;
+	goto out_err;
+    }
+
+    smi->event_handlers = locked_list_alloc(handlers);
+    if (!smi->event_handlers) {
+	rv = ENOMEM;
+	goto out_err;
+    }
+
+    smi->ipmb_change_handlers = locked_list_alloc(handlers);
+    if (!smi->ipmb_change_handlers) {
+	rv = ENOMEM;
+	goto out_err;
+    }
+
     /* Create the locks if they are available. */
     rv = ipmi_create_lock_os_hnd(handlers, &smi->cmd_lock);
     if (rv)
@@ -1468,7 +1557,7 @@ setup(int          if_num,
     if (rv)
 	goto out_err;
 
-    rv = ipmi_create_lock_os_hnd(handlers, &smi->event_handlers_lock);
+    rv = ipmi_create_lock_os_hnd(handlers, &smi->smi_lock);
     if (rv)
 	goto out_err;
 
@@ -1476,15 +1565,19 @@ setup(int          if_num,
 
     ipmi->start_con = smi_start_con;
     ipmi->set_ipmb_addr = smi_set_ipmb_addr;
-    ipmi->set_ipmb_addr_handler = smi_set_ipmb_addr_handler;
-    ipmi->set_con_change_handler = smi_set_con_change_handler;
+    ipmi->add_ipmb_addr_handler = smi_add_ipmb_addr_handler;
+    ipmi->remove_ipmb_addr_handler = smi_remove_ipmb_addr_handler;
+    ipmi->add_con_change_handler = smi_add_con_change_handler;
+    ipmi->remove_con_change_handler = smi_remove_con_change_handler;
     ipmi->send_command = smi_send_command;
-    ipmi->register_for_events = smi_register_for_events;
-    ipmi->deregister_for_events = smi_deregister_for_events;
+    ipmi->add_event_handler = smi_add_event_handler;
+    ipmi->remove_event_handler = smi_remove_event_handler;
     ipmi->send_response = smi_send_response;
     ipmi->register_for_command = smi_register_for_command;
     ipmi->deregister_for_command = smi_deregister_for_command;
     ipmi->close_connection = smi_close_connection;
+    ipmi->close_connection_done = smi_close_connection_done;
+    ipmi->handle_async_event = handle_async_event;
 
     if (smi->using_socket) {
 	rv = handlers->add_fd_to_wait_for(ipmi->os_hnd,
@@ -1506,13 +1599,13 @@ setup(int          if_num,
     }
 
     /* Now it's valid, add it to the smi list. */
-    ipmi_write_lock();
+    ipmi_lock(smi_list_lock);
     if (smi_list)
 	smi_list->prev = smi;
     smi->next = smi_list;
     smi->prev = NULL;
     smi_list = smi;
-    ipmi_write_unlock();
+    ipmi_unlock(smi_list_lock);
 
     *new_con = ipmi;
 
@@ -1524,12 +1617,12 @@ setup(int          if_num,
 }
 
 int
-ipmi_smi_setup_con(int               if_num,
-		   os_handler_t      *handlers,
-		   void              *user_data,
-		   ipmi_con_t        **new_con)
+ipmi_smi_setup_con(int          if_num,
+		   os_handler_t *handlers,
+		   void         *user_data,
+		   ipmi_con_t   **new_con)
 {
-    int                          err;
+    int err;
 
     if (!handlers->add_fd_to_wait_for
 	|| !handlers->remove_fd_to_wait_for
@@ -1540,3 +1633,25 @@ ipmi_smi_setup_con(int               if_num,
     err = setup(if_num, handlers, user_data, new_con);
     return err;
 }
+
+int
+_ipmi_smi_init(os_handler_t *os_hnd)
+{
+    int rv;
+
+    rv = ipmi_create_global_lock(&smi_list_lock);
+    if (rv)
+	return rv;
+    return 0;
+}
+
+void
+_ipmi_smi_shutdown(void)
+{
+    if (smi_list_lock) {
+	ipmi_destroy_lock(smi_list_lock);
+	smi_list_lock = NULL;
+    }
+}
+
+#endif /* HAVE_OPENIPMI_SMI */
