@@ -41,15 +41,19 @@
 #include <errno.h>
 #include <unistd.h>
 #include <string.h>
+#include <stdio.h>
 
 #include <OpenIPMI/os_handler.h>
-#include <OpenIPMI/ipmi_int.h>
 #include <OpenIPMI/ipmi_err.h>
 #include <OpenIPMI/ipmi_pet.h>
 #include <OpenIPMI/ipmi_pef.h>
 #include <OpenIPMI/ipmi_lanparm.h>
 #include <OpenIPMI/ipmi_msgbits.h>
-#include <OpenIPMI/ipmi_domain.h>
+
+#include <OpenIPMI/internal/ipmi_int.h>
+#include <OpenIPMI/internal/locked_list.h>
+#include <OpenIPMI/internal/ipmi_domain.h>
+#include <OpenIPMI/internal/ipmi_mc.h>
 
 /* Recheck the PET config every 10 minutes. */
 #define PET_TIMEOUT_SEC 600
@@ -60,7 +64,8 @@
 /* Alerts get retried this many times. */
 #define IPMI_LANPARM_DEFAULT_ALERT_RETRIES 3
 
-static ipmi_rwlock_t  *pet_lock = NULL;
+#define IPMI_PET_ATTR_NAME "ipmi_pet"
+
 
 /* This data structure defines a data/mask setting for a parameter,
    either from the LAN or PEF parms. */
@@ -78,7 +83,9 @@ typedef struct parm_check_s
    held write, we read-lock the pet timer to avoid locking problem. */
 typedef struct pet_timer_s {
     int          cancelled;
+    int          running;
     os_handler_t *os_hnd;
+    ipmi_lock_t  *lock; /* Lock is here because we need it in the timer. */
     ipmi_pet_t   *pet;
     int          err;
 } pet_timer_t;
@@ -89,22 +96,30 @@ typedef struct pet_timer_s {
 struct ipmi_pet_s
 {
     int destroyed;
+    int in_list;
+
+    unsigned int refcount;
+
+    char name[IPMI_PET_NAME_LEN];
 
     /* Configuration parameters */
+    ipmi_mcid_t      mc;
     ipmi_domain_id_t domain;
     struct in_addr   ip_addr;
+    char             mac_addr[6];
+    unsigned int     policy_num;
     unsigned int     eft_sel;
     unsigned int     apt_sel;
     unsigned int     lan_dest_sel;
 
-    unsigned int connection;
     unsigned int channel;
     ipmi_pet_t   *pet;
     int          pef_err;
     int          pef_lock_broken;
     int          lanparm_err;
     int          lanparm_lock_broken;
-    int          changed;
+    int          changed_lanparm;
+    int          changed_pef;
 
     int            lanparm_check_pos;
     ipmi_lanparm_t *lanparm;
@@ -121,8 +136,6 @@ struct ipmi_pet_s
     ipmi_pet_done_cb destroy_done;
     void             *destroy_cb_data;
 
-    ipmi_lock_t      *lock;
-
     int              in_progress;
 
     /* The LAN configuration parameters to check. */
@@ -134,35 +147,136 @@ struct ipmi_pet_s
     /* Timer to check the configuration periodically. */
     pet_timer_t       *timer_info;
     os_hnd_timer_id_t *timer;
-
-    /* Used so we know when MCs are added and can check to see if we
-       need to scan them. */
-    ipmi_domain_mc_upd_t *mc_upd;
-
-    ipmi_pet_t       *next;
-    ipmi_pet_t       *prev;
-};
-
-/* The list head. */
-static ipmi_pet_t pet_list =
-{
-    .next = &pet_list,
-    .prev = &pet_list
 };
 
 static void rescan_pet(void *cb_data, os_hnd_timer_id_t *id);
 
+static void
+pet_lock(ipmi_pet_t *pet)
+{
+    ipmi_lock(pet->timer_info->lock);
+}
+
+static void
+pet_unlock(ipmi_pet_t *pet)
+{
+    ipmi_unlock(pet->timer_info->lock);
+}
+
 static void
 internal_pet_destroy(ipmi_pet_t *pet)
 {
-    /* Eventually, we could disable the PET entries, but the PEF
-       entries are possibley shared, and there doesn't seem to be any
-       way to disable the LAN entry. */
-    ipmi_unlock(pet->lock);
-    ipmi_destroy_lock(pet->lock);
+    os_handler_t *os_hnd = pet->timer_info->os_hnd;
+
+    if (pet->in_list) {
+	ipmi_domain_attr_t *attr;
+	locked_list_t      *pets;
+	int                rv;
+	rv = ipmi_domain_id_find_attribute(pet->domain,
+					   IPMI_PET_ATTR_NAME, &attr);
+	if (!rv) {
+	    pet->refcount++;
+	    pet->in_list = 0;
+	    pet_unlock(pet);
+	
+	    pets = ipmi_domain_attr_get_data(attr);
+
+	    locked_list_remove(pets, pet, NULL);
+	    ipmi_domain_attr_put(attr);
+	    pet_lock(pet);
+	    /* While we were unlocked, someone may have come in and
+	       grabbed the PET by iterating the list of PETs.  That's
+	       ok, we just let them handle the destruction since this
+	       code will not be entered again. */
+	    if (pet->refcount != 1) {
+		pet->refcount--;
+		pet_unlock(pet);
+		return;
+	    }
+	}
+    }
+    pet_unlock(pet);
+
+    if (os_hnd->stop_timer(os_hnd, pet->timer) == 0) {
+	ipmi_destroy_lock(pet->timer_info->lock);
+	os_hnd->free_timer(os_hnd, pet->timer);
+	ipmi_mem_free(pet->timer_info);
+    } else {
+	pet->timer_info->cancelled = 1;
+    }
+
+    if (pet->destroy_done) {
+	pet->destroy_done(pet, 0, pet->destroy_cb_data);
+    }
+
     ipmi_mem_free(pet);
 }
 
+static void
+pet_get(ipmi_pet_t *pet)
+{
+    pet_lock(pet);
+    pet->refcount++;
+    pet_unlock(pet);
+}
+
+static void
+pet_put(ipmi_pet_t *pet)
+{
+    pet_lock(pet);
+    pet->refcount--;
+    if (pet->refcount == 0) {
+	internal_pet_destroy(pet);
+	return;
+    }
+    pet_unlock(pet);
+}
+
+void
+ipmi_pet_ref(ipmi_pet_t *pet)
+{
+    pet_get(pet);
+}
+
+void
+ipmi_pet_deref(ipmi_pet_t *pet)
+{
+    pet_put(pet);
+}
+
+static int
+pet_attr_init(ipmi_domain_t *domain, void *cb_data, void **data)
+{
+    locked_list_t *pets;
+    
+    pets = locked_list_alloc(ipmi_domain_get_os_hnd(domain));
+    if (!pets)
+	return ENOMEM;
+
+    *data = pets;
+    return 0;
+}
+
+static int
+destroy_pet(void *cb_data, void *item1, void *item2)
+{
+    ipmi_pet_t *pet = item1;
+
+    pet_lock(pet);
+    pet->in_list = 0;
+    pet_unlock(pet);
+    return LOCKED_LIST_ITER_CONTINUE;
+}
+
+static void
+pet_attr_destroy(void *cb_data, void *data)
+{
+    locked_list_t *pets = data;
+
+    locked_list_iterate(pets, destroy_pet, NULL);
+    locked_list_destroy(pets);
+}
+
 /* Must be called locked, this will unlock the PET. */
 static void
 pet_op_done(ipmi_pet_t *pet)
@@ -173,9 +287,23 @@ pet_op_done(ipmi_pet_t *pet)
     pet->in_progress--;
 
     if (pet->in_progress == 0) {
+	if (pet->lanparm) {
+	    ipmi_lanparm_destroy(pet->lanparm, NULL, NULL);
+	    pet->lanparm = NULL;
+	}
+
 	if (pet->done) {
-	    pet->done(pet, 0, pet->cb_data);
+	    ipmi_pet_done_cb done = pet->done;
+	    void             *cb_data = pet->cb_data;
 	    pet->done = NULL;
+	    pet_unlock(pet);
+	    done(pet, 0, cb_data);
+	    pet_lock(pet);
+	}
+
+	if (pet->destroyed) {
+	    internal_pet_destroy(pet);
+	    return;
 	}
 
 	/* Restart the timer */
@@ -183,14 +311,11 @@ pet_op_done(ipmi_pet_t *pet)
 	timeout.tv_usec = 0;
 	os_hnd->start_timer(os_hnd, pet->timer, &timeout, rescan_pet,
 			    pet->timer_info);
+	pet->timer_info->running = 1;
 
-	if (pet->destroyed) {
-	    internal_pet_destroy(pet);
-	    return;
-	}
     }
 
-    ipmi_unlock(pet->lock);
+    pet_unlock(pet);
 }
 
 static void
@@ -200,7 +325,7 @@ lanparm_unlocked(ipmi_lanparm_t *lanparm,
 {
     ipmi_pet_t *pet = cb_data;
 
-    ipmi_lock(pet->lock);
+    pet_lock(pet);
     ipmi_lanparm_destroy(pet->lanparm, NULL, NULL);
     pet->lanparm = NULL;
     pet_op_done(pet);
@@ -215,10 +340,8 @@ lanparm_commited(ipmi_lanparm_t *lanparm,
     int           rv;
     unsigned char data[1];
 
-    ipmi_lock(pet->lock);
+    pet_lock(pet);
     if (pet->destroyed) {
-	ipmi_lanparm_destroy(pet->lanparm, NULL, NULL);
-	pet->lanparm = NULL;
 	pet_op_done(pet);
 	goto out;
     }
@@ -236,7 +359,7 @@ lanparm_commited(ipmi_lanparm_t *lanparm,
 	pet_op_done(pet);
 	goto out;
     }
-    ipmi_unlock(pet->lock);
+    pet_unlock(pet);
  out:
     return;
 }
@@ -247,17 +370,19 @@ lanparm_op_done(ipmi_pet_t *pet, int err)
 {
     int           rv;
 
+    /* Cheap hack, -1 means stop. */
+    if (err == -1)
+	err = 0;
+
     pet->lanparm_err = err;
     if (pet->lanparm_lock_broken) {
 	/* Locking is not supported. */
-	ipmi_lanparm_destroy(pet->lanparm, NULL, NULL);
-	pet->lanparm = NULL;
 	pet_op_done(pet);
 	goto out;
     } else {
 	unsigned char data[1];
 
-	if (!pet->lanparm_err && pet->changed) {
+	if (!pet->lanparm_err && pet->changed_lanparm) {
 	    /* Don't commit if an error occurred. */
 	    data[0] = 2; /* commit */
 	    rv = ipmi_lanparm_set_parm(pet->lanparm, 0, data, 1,
@@ -276,7 +401,7 @@ lanparm_op_done(ipmi_pet_t *pet, int err)
 	    goto out;
 	}
     }
-    ipmi_unlock(pet->lock);
+    pet_unlock(pet);
  out:
     return;
 }
@@ -296,7 +421,7 @@ lanparm_next_config(ipmi_pet_t *pet)
     pet->lanparm_check_pos++;
     if (pet->lanparm_check_pos >= NUM_LANPARM_SETTINGS) {
 	/* Return non-zero, to end the operation. */
-	return 1;
+	return -1;
     }
 
     check = &(pet->lanparm_check[pet->lanparm_check_pos]);
@@ -321,7 +446,7 @@ lanparm_set_config(ipmi_lanparm_t *lanparm,
     ipmi_pet_t    *pet = cb_data;
     int           rv;
 
-    ipmi_lock(pet->lock);
+    pet_lock(pet);
     if (pet->destroyed) {
 	lanparm_op_done(pet, ECANCELED);
 	goto out;
@@ -340,7 +465,7 @@ lanparm_set_config(ipmi_lanparm_t *lanparm,
 	lanparm_op_done(pet, rv);
 	goto out;
     }
-    ipmi_unlock(pet->lock);
+    pet_unlock(pet);
  out:
     return;
 }
@@ -360,7 +485,7 @@ lanparm_got_config(ipmi_lanparm_t *lanparm,
     int           check_failed = 0;
     int           i;
 
-    ipmi_lock(pet->lock);
+    pet_lock(pet);
     if (pet->destroyed) {
 	lanparm_op_done(pet, ECANCELED);
 	goto out;
@@ -418,6 +543,7 @@ lanparm_got_config(ipmi_lanparm_t *lanparm,
 	    lanparm_op_done(pet, rv);
 	    goto out;
 	}
+	pet->changed_lanparm = 1;
     } else {
 	rv = lanparm_next_config(pet);
 	if (rv) {
@@ -426,7 +552,7 @@ lanparm_got_config(ipmi_lanparm_t *lanparm,
 	}
     }
 
-    ipmi_unlock(pet->lock);
+    pet_unlock(pet);
  out:
     return;
 }
@@ -438,7 +564,7 @@ pef_unlocked(ipmi_pef_t    *pef,
 {
     ipmi_pet_t *pet = cb_data;
 
-    ipmi_lock(pet->lock);
+    pet_lock(pet);
     ipmi_pef_destroy(pet->pef, NULL, NULL);
     pet->pef = NULL;
     pet_op_done(pet);
@@ -453,7 +579,7 @@ pef_commited(ipmi_pef_t    *pef,
     int           rv;
     unsigned char data[1];
 
-    ipmi_lock(pet->lock);
+    pet_lock(pet);
     if (pet->destroyed) {
 	ipmi_pef_destroy(pet->pef, NULL, NULL);
 	pet->pef = NULL;
@@ -474,7 +600,7 @@ pef_commited(ipmi_pef_t    *pef,
 	pet_op_done(pet);
 	goto out;
     }
-    ipmi_unlock(pet->lock);
+    pet_unlock(pet);
  out:
     return;
 }
@@ -485,6 +611,10 @@ pef_op_done(ipmi_pet_t *pet, int err)
 {
     int           rv;
 
+    /* Cheap hack, -1 means stop. */
+    if (err == -1)
+	err = 0;
+
     pet->pef_err = err;
     if (pet->pef_lock_broken) {
 	/* Locking is not supported. */
@@ -495,7 +625,7 @@ pef_op_done(ipmi_pet_t *pet, int err)
     } else {
 	unsigned char data[1];
 
-	if (!pet->pef_err && pet->changed) {
+	if (!pet->pef_err && pet->changed_pef) {
 	    /* Don't commit if an error occurred. */
 	    data[0] = 2; /* commit */
 	    rv = ipmi_pef_set_parm(pet->pef, 0, data, 1, pef_commited, pet);
@@ -512,7 +642,7 @@ pef_op_done(ipmi_pet_t *pet, int err)
 	    goto out;
 	}
     }
-    ipmi_unlock(pet->lock);
+    pet_unlock(pet);
  out:
     return;
 }
@@ -532,7 +662,7 @@ pef_next_config(ipmi_pet_t *pet)
     pet->pef_check_pos++;
     if (pet->pef_check_pos >= NUM_PEF_SETTINGS) {
 	/* Return non-zero, to end the operation. */
-	return 1;
+	return -1;
     }
 
     check = &(pet->pef_check[pet->pef_check_pos]);
@@ -555,7 +685,7 @@ pef_set_config(ipmi_pef_t    *pef,
     ipmi_pet_t    *pet = cb_data;
     int           rv;
 
-    ipmi_lock(pet->lock);
+    pet_lock(pet);
     if (pet->destroyed) {
 	pef_op_done(pet, ECANCELED);
 	goto out;
@@ -574,7 +704,7 @@ pef_set_config(ipmi_pef_t    *pef,
 	pef_op_done(pet, rv);
 	goto out;
     }
-    ipmi_unlock(pet->lock);
+    pet_unlock(pet);
  out:
     return;
 }
@@ -594,7 +724,7 @@ pef_got_config(ipmi_pef_t    *pef,
     int           check_failed = 0;
     int           i;
 
-    ipmi_lock(pet->lock);
+    pet_lock(pet);
     if (pet->destroyed) {
 	pef_op_done(pet, ECANCELED);
 	goto out;
@@ -643,6 +773,7 @@ pef_got_config(ipmi_pef_t    *pef,
 	    pef_op_done(pet, rv);
 	    goto out;
 	}
+	pet->changed_pef = 1;
     } else {
 	rv = pef_next_config(pet);
 	if (rv) {
@@ -651,7 +782,7 @@ pef_got_config(ipmi_pef_t    *pef,
 	}
     }
 
-    ipmi_unlock(pet->lock);
+    pet_unlock(pet);
  out:
     return;
 }
@@ -664,7 +795,7 @@ pef_locked(ipmi_pef_t *pef,
     ipmi_pet_t    *pet = cb_data;
     int           rv;
 
-    ipmi_lock(pet->lock);
+    pet_lock(pet);
     if (pet->destroyed) {
 	pef_op_done(pet, ECANCELED);
 	goto out;
@@ -691,7 +822,7 @@ pef_locked(ipmi_pef_t *pef,
 	goto out;
     }
 
-    ipmi_unlock(pet->lock);
+    pet_unlock(pet);
  out:
     return;
 }
@@ -703,7 +834,7 @@ pef_alloced(ipmi_pef_t *pef, int err, void *cb_data)
     unsigned char data[1];
     int           rv;
 
-    ipmi_lock(pet->lock);
+    pet_lock(pet);
     if (pet->destroyed) {
 	pef_op_done(pet, ECANCELED);
 	goto out;
@@ -727,46 +858,40 @@ pef_alloced(ipmi_pef_t *pef, int err, void *cb_data)
 	goto out;
     }
 
-    ipmi_unlock(pet->lock);
+    pet_unlock(pet);
  out:
     return;
 }
 
 static int
-start_pet_setup(ipmi_domain_t *domain,
-		ipmi_pet_t    *pet)
+start_pet_setup(ipmi_mc_t  *mc,
+		ipmi_pet_t *pet)
 {
-    int                          rv = 0;
-    ipmi_system_interface_addr_t si;
-    ipmi_mc_t                    *mc;
+    int  rv = 0;
+
+    pet_lock(pet);
 
     if (pet->in_progress) {
-	rv = EAGAIN;
-	goto out;
+	pet_unlock(pet);
+	return EAGAIN;
     }
 
-    si.addr_type = IPMI_SYSTEM_INTERFACE_ADDR_TYPE;
-    si.channel = pet->connection;
-    si.lun = 0;
-
-    mc = _ipmi_find_mc_by_addr(domain, (ipmi_addr_t *) &si, sizeof(si));
-    if (!mc)
-	return EINVAL;
-
     pet->pet = pet;
     pet->pef_lock_broken = 0;
     pet->pef_err = 0;
+    pet->changed_pef = 0;
     pet->lanparm_lock_broken = 0;
     pet->lanparm_err = 0;
+    pet->changed_lanparm = 0;
 
     pet->pef_check_pos = 0;
+    pet->in_progress++;
     rv = ipmi_pef_alloc(mc, pef_alloced, pet, &pet->pef);
     if (rv) {
+	pet->in_progress--;
 	ipmi_log(IPMI_LOG_WARNING,
 		 "start_pet_setup: Unable to allocate pef: 0x%x", rv);
 	goto out;
-    } else {
-	pet->in_progress++;
     }
 
     /* Now that we have the channel, set up the lan parms. */
@@ -777,6 +902,7 @@ start_pet_setup(ipmi_domain_t *domain,
 		 "start_pet_setup: Unable to allocate lanparm: 0x%x",
 		 rv);
     } else {
+	pet->in_progress++;
 	rv = ipmi_lanparm_get_parm(pet->lanparm,
 				   IPMI_LANPARM_DEST_TYPE,
 				   pet->lan_dest_sel,
@@ -784,53 +910,72 @@ start_pet_setup(ipmi_domain_t *domain,
 				   lanparm_got_config,
 				   pet);
 	if (rv) {
+	    pet->in_progress--;
 	    ipmi_log(IPMI_LOG_WARNING,
 		     "start_pet_setup: Unable to get dest type: 0x%x",
 		     rv);
 	    ipmi_lanparm_destroy(pet->lanparm, NULL, NULL);
 	    pet->lanparm = NULL;
-	} else {
-	    pet->in_progress++;
 	}
     }
     rv = 0; /* We continue with the PEF run, even if the lanparm fails. */
 
  out:
+    pet_unlock(pet);
     return rv;
 }
 
 int
-ipmi_pet_create(ipmi_domain_t    *domain,
-		unsigned int     connection,
-		unsigned int     channel,
-		struct in_addr   ip_addr,
-		unsigned char    mac_addr[6],
-		unsigned int     eft_sel,
-		unsigned int     policy_num,
-		unsigned int     apt_sel,
-		unsigned int     lan_dest_sel,
-		ipmi_pet_done_cb done,
-		void             *cb_data,
-		ipmi_pet_t       **ret_pet)
+ipmi_pet_create_mc(ipmi_mc_t        *mc,
+		   unsigned int     channel,
+		   struct in_addr   ip_addr,
+		   unsigned char    mac_addr[6],
+		   unsigned int     eft_sel,
+		   unsigned int     policy_num,
+		   unsigned int     apt_sel,
+		   unsigned int     lan_dest_sel,
+		   ipmi_pet_done_cb done,
+		   void             *cb_data,
+		   ipmi_pet_t       **ret_pet)
 {
-    ipmi_pet_t     *pet;
-    int            rv;
-    os_handler_t   *os_hnd;
+    ipmi_pet_t         *pet;
+    int                rv;
+    os_handler_t       *os_hnd;
+    char               domain_name[IPMI_MC_NAME_LEN];
+    ipmi_domain_t      *domain = ipmi_mc_get_domain(mc);
+    ipmi_domain_attr_t *attr;
+    locked_list_t      *pets;
+
+    rv = ipmi_domain_register_attribute(domain, IPMI_PET_ATTR_NAME,
+					pet_attr_init, pet_attr_destroy, NULL,
+					&attr);
+    if (rv)
+	return rv;
+    pets = ipmi_domain_attr_get_data(attr);
 
     pet = ipmi_mem_alloc(sizeof(*pet));
-    if (!pet)
+    if (!pet) {
+	ipmi_domain_attr_put(attr);
 	return ENOMEM;
+    }
     memset(pet, 0, sizeof(*pet));
 
+    ipmi_domain_get_name(domain, domain_name, sizeof(domain_name));
+    snprintf(pet->name, sizeof(pet->name), "%s.%d", domain_name,
+	     ipmi_domain_get_unique_num(domain));
+    pet->refcount = 1;
+    pet->in_list = 1;
+    pet->mc = ipmi_mc_convert_to_id(mc);
     pet->domain = ipmi_domain_convert_to_id(domain);
-    pet->connection = connection;
     pet->channel = channel;
     pet->ip_addr = ip_addr;
+    pet->policy_num = policy_num;
     pet->eft_sel = eft_sel;
     pet->apt_sel = apt_sel;
     pet->lan_dest_sel = lan_dest_sel;
     pet->done = done;
     pet->cb_data = cb_data;
+    memcpy(pet->mac_addr, mac_addr, sizeof(pet->mac_addr));
     pet->in_progress = 0;
 
     /* Set up all the data we want in the PEF and LANPARMs
@@ -898,42 +1043,50 @@ ipmi_pet_create(ipmi_domain_t    *domain,
 
     os_hnd = ipmi_domain_get_os_hnd(domain);
     pet->os_hnd = os_hnd;
-    rv = ipmi_create_lock_os_hnd(os_hnd, &pet->lock);
-    if (rv) {
-	ipmi_mem_free(pet);
-	return rv;
-    }
 
-    ipmi_rwlock_write_lock(pet_lock);
     /* Start a timer for this PET to periodically check it. */
     pet->timer_info = ipmi_mem_alloc(sizeof(*(pet->timer_info)));
     if (!pet->timer_info) {
 	rv = ENOMEM;
-	goto out_unlock_err;
+	goto out_err;
     }
     pet->timer_info->cancelled = 0;
     pet->timer_info->os_hnd = os_hnd;
     pet->timer_info->pet = pet;
+    pet->timer_info->running = 0;
+    pet->timer_info->lock = NULL;
     rv = os_hnd->alloc_timer(os_hnd, &pet->timer);
     if (rv)
-	goto out_unlock_err;
+	goto out_err;
 
-    rv = start_pet_setup(domain, pet);
+    rv = ipmi_create_lock_os_hnd(os_hnd, &pet->timer_info->lock);
     if (rv)
-	goto out_unlock_err;
+	goto out_err;
 
-    pet->next = &pet_list;
-    pet->prev = pet_list.prev;
-    pet_list.prev->next = pet;
-    pet_list.prev = pet;
+    if (! locked_list_add(pets, pet, NULL)) {
+	rv = ENOMEM;
+	goto out_err;
+    }
+
+    ipmi_domain_attr_put(attr);
+
+    rv = start_pet_setup(mc, pet);
+    if (rv)
+	goto out_err;
+
+    if (ret_pet)
+	*ret_pet = pet;
 
-    ipmi_rwlock_write_unlock(pet_lock);
     return 0;
 
- out_unlock_err:
+ out_err:
+    locked_list_remove(pets, pet, NULL);
+    ipmi_domain_attr_put(attr);
     if (pet->timer_info) {
 	if (pet->timer) {
 	    if (os_hnd->stop_timer(os_hnd, pet->timer) == 0) {
+		if (pet->timer_info->lock)
+		    ipmi_destroy_lock(pet->timer_info->lock);
 		os_hnd->free_timer(os_hnd, pet->timer);
 		ipmi_mem_free(pet->timer_info);
 	    } else {
@@ -942,18 +1095,67 @@ ipmi_pet_create(ipmi_domain_t    *domain,
 	} else
 	    ipmi_mem_free(pet->timer_info);
     }
-    ipmi_rwlock_write_unlock(pet_lock);
     ipmi_mem_free(pet);
     return rv;
 }
 
+int
+ipmi_pet_create(ipmi_domain_t    *domain,
+		unsigned int     connection,
+		unsigned int     channel,
+		struct in_addr   ip_addr,
+		unsigned char    mac_addr[6],
+		unsigned int     eft_sel,
+		unsigned int     policy_num,
+		unsigned int     apt_sel,
+		unsigned int     lan_dest_sel,
+		ipmi_pet_done_cb done,
+		void             *cb_data,
+		ipmi_pet_t       **ret_pet)
+{
+    ipmi_system_interface_addr_t si;
+    ipmi_mc_t                    *mc;
+    int                          rv;
+
+    si.addr_type = IPMI_SYSTEM_INTERFACE_ADDR_TYPE;
+    si.channel = connection;
+    si.lun = 0;
+
+    mc = _ipmi_find_mc_by_addr(domain, (ipmi_addr_t *) &si, sizeof(si));
+    if ((!mc)  && (connection == 0)) {
+	/* If the specific connection doesn't exist and the connection
+	   is 0, use the BMC channel. */
+	si.channel = IPMI_BMC_CHANNEL;
+	mc = _ipmi_find_mc_by_addr(domain, (ipmi_addr_t *) &si,
+				   sizeof(si));
+    }
+
+    if (!mc)
+	return EINVAL;
+
+    rv = ipmi_pet_create_mc(mc,
+			    channel,
+			    ip_addr,
+			    mac_addr,
+			    eft_sel,
+			    policy_num,
+			    apt_sel,
+			    lan_dest_sel,
+			    done,
+			    cb_data,
+			    ret_pet);
+    _ipmi_mc_put(mc);
+
+    return rv;
+}
+
 static void
-rescan_pet_domain(ipmi_domain_t *domain, void *cb_data)
+rescan_pet_mc(ipmi_mc_t *mc, void *cb_data)
 {
     pet_timer_t *timer_info = cb_data;
     ipmi_pet_t  *pet = timer_info->pet;
 
-    timer_info->err = start_pet_setup(domain, pet);
+    timer_info->err = start_pet_setup(mc, pet);
 }
 
 static void
@@ -964,19 +1166,20 @@ rescan_pet(void *cb_data, os_hnd_timer_id_t *id)
     int            rv;
     struct timeval timeout;
 
-    ipmi_rwlock_read_lock(pet_lock);
-
+    ipmi_lock(timer_info->lock);
     if (timer_info->cancelled) {
+	ipmi_unlock(timer_info->lock);
 	timer_info->os_hnd->free_timer(timer_info->os_hnd, id);
+	ipmi_destroy_lock(timer_info->lock);
 	ipmi_mem_free(timer_info);
-	ipmi_rwlock_read_unlock(pet_lock);
 	return;
     }
-
     pet = timer_info->pet;
-    ipmi_lock(pet->lock);
+    pet->timer_info->running = 0;
+    pet_get(pet);
+
     timer_info->err = 0;
-    rv = ipmi_domain_pointer_cb(pet->domain, rescan_pet_domain, timer_info);
+    rv = ipmi_mc_pointer_cb(pet->mc, rescan_pet_mc, timer_info);
     if (!rv)
 	rv = timer_info->err;
 
@@ -987,10 +1190,10 @@ rescan_pet(void *cb_data, os_hnd_timer_id_t *id)
 	timeout.tv_usec = 0;
 	os_hnd->start_timer(os_hnd, pet->timer, &timeout, rescan_pet,
 			    pet->timer_info);
+	pet->timer_info->running = 1;
     }
 
-    ipmi_unlock(pet->lock);
-    ipmi_rwlock_read_unlock(pet_lock);
+    ipmi_unlock(timer_info->lock);
 }
 
 int
@@ -999,75 +1202,149 @@ ipmi_pet_destroy(ipmi_pet_t       *pet,
 		 void             *cb_data)
 
 {
-    ipmi_pet_t *e;
-    int        rv = 0;
+    pet_lock(pet);
+    if (pet->in_list) {
+	ipmi_domain_attr_t *attr;
+	locked_list_t      *pets;
+	int                rv;
 
-    ipmi_rwlock_write_lock(pet_lock);
+	pet->in_list = 0;
+	rv = ipmi_domain_id_find_attribute(pet->domain,
+					   IPMI_PET_ATTR_NAME, &attr);
+	if (!rv) {
+	    pet_unlock(pet);
+	
+	    pets = ipmi_domain_attr_get_data(attr);
 
-    if (pet->timer_info) {
-	os_handler_t *os_hnd = pet->timer_info->os_hnd;
-	if (pet->timer) {
-	    if (os_hnd->stop_timer(os_hnd, pet->timer) == 0) {
-		os_hnd->free_timer(os_hnd, pet->timer);
-		ipmi_mem_free(pet->timer_info);
-	    } else {
-		pet->timer_info->cancelled = 1;
-	    }
-	} else
-	    ipmi_mem_free(pet->timer_info);
+	    locked_list_remove(pets, pet, NULL);
+	    ipmi_domain_attr_put(attr);
+	    pet_lock(pet);
+	}
     }
 
-    e = pet_list.next;
-    while (e != &pet_list) {
-	if (e == pet)
-	    break;
-	e = e->next;
-    }
-    if (e == &pet_list) {
-	rv = EINVAL;
-	goto out_unlock_err;
-    }
-
-    pet->next->prev = pet->prev;
-    pet->prev->next = pet->next;
-
-    ipmi_lock(pet->lock);
     pet->destroyed = 1;
     pet->destroy_done = done;
     pet->destroy_cb_data = cb_data;
+    pet_unlock(pet);
 
-    if (! pet->in_progress)
-	internal_pet_destroy(pet);
-    else
-	ipmi_unlock(pet->lock);
-
- out_unlock_err:
-    ipmi_rwlock_write_unlock(pet_lock);
-    return rv;
-}
-
-int
-_ipmi_pet_init(void)
-{
-    int rv;
-
-    rv = ipmi_create_global_rwlock(&pet_lock);
-    if (rv)
-	return rv;
-    
+    pet_put(pet);
     return 0;
 }
 
-void
-_ipmi_pet_shutdown(void)
+int
+ipmi_pet_get_name(ipmi_pet_t *pet, char *name, int length)
 {
-    ipmi_pet_t *e, *n;
+    int  slen;
 
-    e = pet_list.next;
-    while (e != &pet_list) {
-	n = e->next;
-	ipmi_pet_destroy(e, NULL, NULL);
-	e = n;
+    if (length <= 0)
+	return 0;
+
+    /* Never changes, no lock needed. */
+    slen = strlen(pet->name);
+    if (slen == 0) {
+	if (name)
+	    *name = '\0';
+	goto out;
     }
-    ipmi_destroy_rwlock(pet_lock);
+
+    if (name) {
+	memcpy(name, pet->name, slen);
+	name[slen] = '\0';
+    }
+ out:
+    return slen;
+}
+
+typedef struct iterate_pets_info_s
+{
+    ipmi_pet_ptr_cb handler;
+    void            *cb_data;
+} iterate_pets_info_t;
+
+static int
+pets_handler(void *cb_data, void *item1, void *item2)
+{
+    iterate_pets_info_t *info = cb_data;
+    info->handler(item1, info->cb_data);
+    pet_put(item1);
+    return LOCKED_LIST_ITER_CONTINUE;
+}
+
+static int
+pets_prefunc(void *cb_data, void *item1, void *item2)
+{
+    pet_get(item1);
+    return LOCKED_LIST_ITER_CONTINUE;
+}
+
+void
+ipmi_pet_iterate_pets(ipmi_domain_t   *domain,
+		      ipmi_pet_ptr_cb handler,
+		      void            *cb_data)
+{
+    iterate_pets_info_t info;
+    ipmi_domain_attr_t  *attr;
+    locked_list_t       *pets;
+    int                 rv;
+
+    rv = ipmi_domain_find_attribute(domain, IPMI_PET_ATTR_NAME,
+				    &attr);
+    if (rv)
+	return;
+    pets = ipmi_domain_attr_get_data(attr);
+
+    info.handler = handler;
+    info.cb_data = cb_data;
+    locked_list_iterate_prefunc(pets, pets_prefunc, pets_handler, &info);
+    ipmi_domain_attr_put(attr);
+}
+
+ipmi_mcid_t
+ipmi_pet_get_mc_id(ipmi_pet_t *pet)
+{
+    return pet->mc;
+}
+
+unsigned int
+ipmi_pet_get_channel(ipmi_pet_t *pet)
+{
+    return pet->channel;
+}
+
+struct in_addr *
+ipmi_pet_get_ip_addr(ipmi_pet_t *pet, struct in_addr *ip_addr)
+{
+    memcpy(ip_addr, &pet->ip_addr, sizeof(*ip_addr));
+    return ip_addr;
+}
+
+unsigned char *
+ipmi_pet_get_mac_addr(ipmi_pet_t *pet, unsigned char mac_addr[6])
+{
+    memcpy(mac_addr, pet->mac_addr, sizeof(mac_addr));
+    return mac_addr;
+}
+
+unsigned int
+ipmi_pet_get_eft_sel(ipmi_pet_t *pet)
+{
+    return pet->eft_sel;
+}
+
+unsigned int
+ipmi_pet_get_policy_num(ipmi_pet_t *pet)
+{
+    return pet->policy_num;
+}
+
+unsigned int
+ipmi_pet_get_apt_sel(ipmi_pet_t *pet)
+{
+    return pet->apt_sel;
+}
+
+unsigned int
+ipmi_pet_get_lan_dest_sel(ipmi_pet_t *pet)
+{
+    return pet->lan_dest_sel;
 }
