@@ -32,19 +32,24 @@
  */
 
 #include <string.h>
+#include <stdio.h>
 
 #include <OpenIPMI/ipmiif.h>
 #include <OpenIPMI/ipmi_sdr.h>
 #include <OpenIPMI/ipmi_msgbits.h>
-#include <OpenIPMI/ipmi_domain.h>
-#include <OpenIPMI/ipmi_mc.h>
 #include <OpenIPMI/ipmi_err.h>
-#include <OpenIPMI/ipmi_int.h>
-#include <OpenIPMI/opq.h>
-#include <OpenIPMI/ilist.h>
 
-/* Max bytes to try to get at a time. */
-#define MAX_SDR_FETCH 20
+#include <OpenIPMI/internal/opq.h>
+#include <OpenIPMI/internal/ilist.h>
+#include <OpenIPMI/internal/ipmi_domain.h>
+#include <OpenIPMI/internal/ipmi_mc.h>
+#include <OpenIPMI/internal/ipmi_int.h>
+
+/* Max bytes to try to get at a time, the minimum allowed, and the
+   amount to decrement between tries. */
+#define MAX_SDR_FETCH_BYTES 28
+#define MIN_SDR_FETCH_BYTES 10
+#define SDR_FETCH_BYTES_DECR 6
 
 /* Do up to this many retries when the reservation is lost. */
 #define MAX_SDR_FETCH_RETRIES 10
@@ -71,7 +76,7 @@ typedef struct fetch_info_s
     unsigned int idx;
     unsigned int offset;
     unsigned int read_len;
-    unsigned char data[MAX_SDR_FETCH+2];
+    unsigned char data[MAX_SDR_FETCH_BYTES+2];
 
     ilist_item_t link;
 } fetch_info_t;
@@ -130,12 +135,17 @@ struct ipmi_sdr_info_s
     /* Are we currently fetching SDRs? */
     enum fetch_state_e fetch_state;
 
+    /* Are we currently fetching database data? */
+    int                db_fetching;
+
     /* When fetching the data in event-driven mode, these are the
        variables that track what is going on. */
     int                    curr_rec_id;
     int                    read_offset; /* Next data to read, -1 if
                                            the header */
 
+    int                    fetch_size;
+
     int                    curr_read_rec_id;
     int                    next_read_rec_id;
     int                    curr_read_idx;
@@ -177,8 +187,12 @@ struct ipmi_sdr_info_s
     unsigned int num_sdrs;
     unsigned int sdr_array_size;
     ipmi_sdr_t *sdrs;
+
+    char db_key[32+5];
+    int  db_key_set;
 };
 
+static void internal_destroy_sdr_info(ipmi_sdr_info_t *sdrs);
 
 static inline void sdr_lock(ipmi_sdr_info_t *sdrs)
 {
@@ -214,6 +228,119 @@ cleanup_fetch_items(ipmi_sdr_info_t *sdrs)
     ilist_iter(sdrs->outstanding_fetch, cancel_fetch, NULL);
 }
 
+static void
+process_db_data(ipmi_sdr_info_t *sdrs,
+		unsigned char   *db_data,
+		unsigned int    len)
+{
+    int           num;
+    unsigned char *d;
+
+    if (len < 9)
+	goto no_db;
+
+    /* Format# is the last byte. */
+    d = db_data + len - 1;
+    if (*d != 1)
+	goto no_db;
+
+    /* timestamps are the 8 bytes before the format#. */
+    d -= 8;
+    sdrs->last_addition_timestamp = ipmi_get_uint32(d);
+    d += 4;
+    sdrs->last_erase_timestamp = ipmi_get_uint32(d);
+    d += 4;
+    len -= 9;
+    num = len / sizeof(ipmi_sdr_t);
+    /* Allocate 9 extra bytes for storing the timestamps and
+     * format#. */
+    sdrs->sdrs = ipmi_mem_alloc((sizeof(ipmi_sdr_t) * num) + 9);
+    if (!sdrs->sdrs)
+	goto no_db;
+    memcpy(sdrs->sdrs, db_data, sizeof(ipmi_sdr_t) * num);
+    sdrs->num_sdrs = num;
+    sdrs->sdr_array_size = num;
+    sdrs->fetched = 1;
+
+ no_db:
+    sdrs->os_hnd->database_free(sdrs->os_hnd, db_data);
+}
+
+static void
+db_fetched(void          *cb_data,
+	   int           err,
+	   unsigned char *db_data,
+	   unsigned int  db_data_len)
+{
+    ipmi_sdr_info_t *sdrs = cb_data;
+
+    sdr_lock(sdrs);
+    if (sdrs->destroyed) {
+	internal_destroy_sdr_info(sdrs);
+	return;
+    }
+
+    /* Note that since this is run from the opq, there is no reason to
+       check to see if another fetch is going on and has finished.  We
+       are guaranteed that this works. */
+    if (!err)
+	process_db_data(sdrs, db_data, db_data_len);
+
+    sdrs->db_fetching = 0;
+    sdr_unlock(sdrs);
+    if (!err)
+	sdrs->os_hnd->database_free(sdrs->os_hnd, db_data);
+    opq_op_done(sdrs->sdr_wait_q);
+}
+
+static void
+start_db_fetch(void *cb_data, int shutdown)
+{
+    ipmi_sdr_info_t *sdrs = cb_data;
+    int             rv = ENOSYS;
+
+    if (shutdown)
+	return;
+
+    sdr_lock(sdrs);
+    if (sdrs->destroyed) {
+	internal_destroy_sdr_info(sdrs);
+	return;
+    }
+
+    /* Go ahead and do the database fetch here if we have support. */
+    if (sdrs->os_hnd->database_find && sdrs->db_key_set)
+    {
+	unsigned char *db_data;
+	unsigned int  db_data_len;
+	int           data_fetched = 0;
+
+	rv = sdrs->os_hnd->database_find(sdrs->os_hnd,
+					 sdrs->db_key,
+					 &data_fetched,
+					 &db_data,
+					 &db_data_len,
+					 db_fetched,
+					 sdrs);
+	/* If the above fails, no problem, the db_data will be NULL. */
+	if (!rv) {
+	    if (data_fetched) {
+		process_db_data(sdrs, db_data, db_data_len);
+		rv = -1; /* Just mark it as done */
+	    }
+	}
+    } else {
+	rv = -1; /* Just mark it as done */
+    }
+
+    if (rv) {
+	sdrs->db_fetching = 0;
+	sdr_unlock(sdrs);
+	opq_op_done(sdrs->sdr_wait_q);
+    } else
+	sdr_unlock(sdrs);
+}
+
 int
 ipmi_sdr_info_alloc(ipmi_domain_t   *domain,
 		    ipmi_mc_t       *mc,
@@ -253,6 +380,7 @@ ipmi_sdr_info_alloc(ipmi_domain_t   *domain,
     sdrs->lun = lun;
     sdrs->sensor = sensor;
     sdrs->sdr_wait_q = NULL;
+    sdrs->fetch_size = MAX_SDR_FETCH_BYTES;
 
     /* Assume we have a dynamic population until told otherwise. */
     sdrs->dynamic_population = 1;
@@ -321,6 +449,12 @@ ipmi_sdr_info_alloc(ipmi_domain_t   *domain,
     return rv;
 }
 
+void
+ipmi_sdr_set_mc(ipmi_sdr_info_t *sdrs, ipmi_mc_t *mc)
+{
+    sdrs->mc = ipmi_mc_convert_to_id(mc);
+}
+
 static void
 internal_destroy_sdr_info(ipmi_sdr_info_t *sdrs)
 {
@@ -354,6 +488,16 @@ internal_destroy_sdr_info(ipmi_sdr_info_t *sdrs)
     ipmi_mem_free(sdrs);
 }
 
+void
+ipmi_sdr_clean_out_sdrs(ipmi_sdr_info_t *sdrs)
+{
+    if (sdrs->sdrs)
+	ipmi_mem_free(sdrs->sdrs);
+    sdrs->sdrs = NULL;
+    sdrs->dynamic_population = 1;
+    sdrs->fetched = 0;
+}
+
 int
 ipmi_sdr_info_destroy(ipmi_sdr_info_t      *sdrs,
 		      ipmi_sdr_destroyed_t handler,
@@ -369,7 +513,7 @@ ipmi_sdr_info_destroy(ipmi_sdr_info_t      *sdrs,
     sdrs->destroyed = 1;
     sdrs->destroy_handler = handler;
     sdrs->destroy_cb_data = cb_data;
-    if (sdrs->fetch_state != IDLE) {
+    if ((sdrs->fetch_state != IDLE) || sdrs->db_fetching) {
 	/* It's currently in fetch state, so let it be destroyed in
            the handler, since we can't cancel the handler or
            operation. */
@@ -408,6 +552,24 @@ fetch_complete(ipmi_sdr_info_t *sdrs, int err)
 	sdrs->working_sdrs = NULL;
 	if (to_free)
 	    ipmi_mem_free(to_free);
+
+	if (sdrs->sdrs && sdrs->db_key_set && sdrs->os_hnd->database_store) {
+	    unsigned int  len = sdrs->num_sdrs * sizeof(ipmi_sdr_t);
+	    unsigned char *d = ((unsigned char *) sdrs->sdrs) + len;
+
+	    /* We always allocate 9 extra bytes in the SDR data to put
+	       the timestamps and format at the end. */
+	    ipmi_set_uint32(d, sdrs->last_addition_timestamp);
+	    d += 4;
+	    ipmi_set_uint32(d, sdrs->last_erase_timestamp);
+	    d += 4;
+	    *d = 1; /* format # */
+	    len += 9;
+	    sdrs->os_hnd->database_store(sdrs->os_hnd,
+					 sdrs->db_key,
+					 (unsigned char *) sdrs->sdrs,
+					 len);
+	}
     }
     sdrs->fetch_state = HANDLERS;
     sdr_unlock(sdrs);
@@ -450,7 +612,7 @@ handle_reservation_check(ipmi_mc_t  *mc,
     if (!mc) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
 		 "MC went away while SDR fetch was in progress(1)");
-	fetch_complete(sdrs, ENXIO);
+	fetch_complete(sdrs, ECANCELED);
 	goto out;
     }
 	
@@ -463,7 +625,7 @@ handle_reservation_check(ipmi_mc_t  *mc,
 	    ipmi_log(IPMI_LOG_ERR_INFO,
 		     "Lost reservation too many times trying to"
 		     " fetch the SDRs");
-	    fetch_complete(sdrs, EBUSY);
+	    fetch_complete(sdrs, EAGAIN);
 	    goto out;
 	} else {
 	    if (sdrs->working_sdrs) {
@@ -701,7 +863,7 @@ handle_sdr_data(ipmi_mc_t  *mc,
 
 	ipmi_log(IPMI_LOG_ERR_INFO,
 		 "MC went away while SDR fetch was in progress(2)");
-	fetch_complete(sdrs, ENXIO);
+	fetch_complete(sdrs, ECANCELED);
 	goto out;
     }
 
@@ -745,12 +907,12 @@ handle_sdr_data(ipmi_mc_t  *mc,
 	    ipmi_log(IPMI_LOG_ERR_INFO,
 		     "To many retries trying to fetch SDRs");
 
-	    sdrs->fetch_err = EBUSY;
+	    sdrs->fetch_err = EAGAIN;
 
 	    if (!ilist_empty(sdrs->outstanding_fetch))
 		goto out_unlock;
 
-	    fetch_complete(sdrs, EBUSY);
+	    fetch_complete(sdrs, EAGAIN);
 	    goto out;
 	}
 
@@ -777,12 +939,12 @@ handle_sdr_data(ipmi_mc_t  *mc,
 	    ipmi_log(IPMI_LOG_ERR_INFO,
 		     "Lost reservation too many times trying to fetch SDRs");
 
-	    sdrs->fetch_err = EBUSY;
+	    sdrs->fetch_err = EAGAIN;
 
 	    if (!ilist_empty(sdrs->outstanding_fetch))
 		goto out_unlock;
 
-	    fetch_complete(sdrs, EBUSY);
+	    fetch_complete(sdrs, EAGAIN);
 	    goto out;
 	} else {
 	    if (sdrs->working_sdrs) {
@@ -819,6 +981,37 @@ handle_sdr_data(ipmi_mc_t  *mc,
 	goto out;
     }
 
+    if (rsp->data[0] == IPMI_CANNOT_RETURN_REQ_LENGTH_CC) {
+	/* It's more than the system can return in a single messages,
+	   decrease the size. */
+	ilist_add_tail(sdrs->free_fetch, info, &info->link);
+
+	sdrs->fetch_size -= SDR_FETCH_BYTES_DECR;
+	if (sdrs->fetch_size < MIN_SDR_FETCH_BYTES) {
+	    ipmi_log(IPMI_LOG_ERR_INFO,
+		     "SDR target chould not support the minimum fetch size");
+
+	    sdrs->fetch_err = IPMI_IPMI_ERR_VAL(rsp->data[0]);
+
+	    if (!ilist_empty(sdrs->outstanding_fetch))
+		goto out_unlock;
+
+	    fetch_complete(sdrs, IPMI_IPMI_ERR_VAL(rsp->data[0]));
+	    goto out;
+	} else {
+	    /* Cancel any current or newer pending operations. */
+	    cancel_same_or_newer(sdrs, info->idx);
+
+	    /* Re-start the fetch on this SDR. */
+	    sdrs->next_read_offset = -1;
+	    sdrs->read_size = -1;
+	    sdrs->next_read_rec_id = info->sdr_rec;
+	    sdrs->curr_read_idx = info->idx-1;
+
+	    goto out_nextmsg;
+	}
+    }
+
     if (rsp->data[0] != 0) {
 	ilist_add_tail(sdrs->free_fetch, info, &info->link);
 	sdrs->fetch_retry_count = MAX_SDR_FETCH_RETRIES+1;
@@ -931,8 +1124,9 @@ handle_sdr_data(ipmi_mc_t  *mc,
 		    int new_num_sdrs = sdrs->working_num_sdrs + 10;
 		    ipmi_sdr_t *new_sdrs;
 
-		    new_sdrs = ipmi_mem_alloc(sizeof(ipmi_sdr_t)
-					      * new_num_sdrs);
+		    /* Allocate 9 extra bytes for the db info. */
+		    new_sdrs = ipmi_mem_alloc((sizeof(ipmi_sdr_t)
+					       * new_num_sdrs) + 9);
 		    if (!new_sdrs) {
 			ipmi_log(IPMI_LOG_ERR_INFO,
 				 "SDR respository had more SDRs than"
@@ -973,8 +1167,8 @@ handle_sdr_data(ipmi_mc_t  *mc,
 	    info->read_len = SDR_HEADER_SIZE;
 	} else {
 	    info->read_len = sdrs->read_size - sdrs->next_read_offset;
-	    if (info->read_len > MAX_SDR_FETCH)
-		info->read_len = MAX_SDR_FETCH;
+	    if (info->read_len > sdrs->fetch_size)
+		info->read_len = sdrs->fetch_size;
 	    info->offset = sdrs->next_read_offset;
 	    sdrs->next_read_offset += info->read_len;
 	}
@@ -1015,6 +1209,11 @@ initial_sdr_fetch(ipmi_sdr_info_t *sdrs, ipmi_mc_t *mc)
     }
     info->sdr_rec = sdrs->curr_rec_id;
     info->offset = 0;
+    /* If all systems were implemented correctly, we could do a big
+       fetch here and if it was too big then they would just return
+       what was available.  Some systems, though, are picky about the
+       sizes being exactly right.  So we fetch the header first so we
+       can get the size. */
     info->read_len = SDR_HEADER_SIZE;
     info->fetch_retry_num = sdrs->fetch_retry_count;
     info->idx = sdrs->curr_read_idx;
@@ -1042,7 +1241,7 @@ handle_reservation(ipmi_mc_t  *mc,
     if (!mc) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
 		 "MC went away while SDR fetch was in progress(3)");
-	fetch_complete(sdrs, ENXIO);
+	fetch_complete(sdrs, ECANCELED);
 	goto out;
     }
 	
@@ -1109,7 +1308,7 @@ handle_sdr_info(ipmi_mc_t  *mc,
     if (!mc) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
 		 "MC went away while SDR fetch was in progress(4)");
-	fetch_complete(sdrs, ENXIO);
+	fetch_complete(sdrs, ECANCELED);
 	goto out;
     }
 	
@@ -1218,8 +1417,9 @@ handle_sdr_info(ipmi_mc_t  *mc,
 	goto out;
     }
 
-    sdrs->working_sdrs = ipmi_mem_alloc(sizeof(ipmi_sdr_t)
-					* sdrs->working_num_sdrs);
+    /* Allocate 9 extra bytes for the db info. */
+    sdrs->working_sdrs = ipmi_mem_alloc((sizeof(ipmi_sdr_t)
+					* sdrs->working_num_sdrs) + 9);
     if (!sdrs->working_sdrs) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
 		 "Could not allocate working SDR information");
@@ -1403,12 +1603,14 @@ sdr_fetch_cb(ipmi_mc_t *mc, void *cb_data)
     sdr_fetch_info_t    *info = cb_data;
     ipmi_sdr_info_t     *sdrs = info->sdrs;
     sdr_fetch_handler_t *elem;
+    unsigned char       guid[16];
 
     elem = ipmi_mem_alloc(sizeof(*elem));
     if (!elem) {
 	info->rv = ENOMEM;
 	return;
     }
+    memset(elem, 0, sizeof(*elem));
 
     elem->sdrs = sdrs;
     elem->handler = info->handler;
@@ -1427,6 +1629,26 @@ sdr_fetch_cb(ipmi_mc_t *mc, void *cb_data)
     }
 
     sdr_lock(sdrs);
+    if (!sdrs->fetched) {
+	/* Look in the database before the first fetch. */
+	if (ipmi_mc_get_guid(mc, guid) == 0) {
+	    char *s;
+	    int  i;
+
+	    s = sdrs->db_key;
+	    s += sprintf(s, "sdr-");
+	    for (i=0; i<16; i++)
+		s += sprintf(s, "%2.2x", guid[i]);
+	    sdrs->db_key_set = 1;
+	}
+
+	sdrs->db_fetching = 1;
+	sdr_unlock(sdrs);
+	if (!opq_new_op(sdrs->sdr_wait_q, start_db_fetch, sdrs, 0))
+	    sdrs->db_fetching = 0;
+    } else
+	sdr_unlock(sdrs);
+
     if (! opq_new_op_with_done(sdrs->sdr_wait_q,
 			       handle_start_fetch,
 			       sdrs,
@@ -1435,7 +1657,6 @@ sdr_fetch_cb(ipmi_mc_t *mc, void *cb_data)
     {
 	info->rv = ENOMEM;
     }
-    sdr_unlock(sdrs);
 
  out:
     if (info->rv)
@@ -1768,7 +1989,9 @@ ipmi_sdr_add(ipmi_sdr_info_t *sdrs,
     sdr_lock(sdrs);
     if (sdrs->num_sdrs >= sdrs->sdr_array_size) {
 	ipmi_sdr_t *new_array;
-	new_array = ipmi_mem_alloc(sizeof(ipmi_sdr_t) * sdrs->sdr_array_size + 10);
+	/* Allocate 9 extra bytes for the db info. */
+	new_array = ipmi_mem_alloc((sizeof(ipmi_sdr_t)
+				    * (sdrs->sdr_array_size + 10)) + 9);
 	if (!new_array) {
 	    rv = ENOMEM;
 	    goto out_unlock;
@@ -1855,7 +2078,7 @@ start_sdr_write(ipmi_sdr_info_t *sdrs,
     cmd_msg.data[8] = sdr->major_version | (sdr->minor_version << 4);
     cmd_msg.data[9] = sdr->type;
     cmd_msg.data[10] = sdr->length;
-    if (sdr->length >= (MAX_SDR_FETCH - 5)) {
+    if (sdr->length >= (sdrs->fetch_size - 5)) {
 	cmd_msg.data[5] = 1;
 	memcpy(cmd_msg.data+11, sdr->data, sdr->length);
 	cmd_msg.data_len = 11 + sdr->length;
@@ -1863,9 +2086,9 @@ start_sdr_write(ipmi_sdr_info_t *sdrs,
 				    handle_sdr_write_done, sdr);
     } else {
 	cmd_msg.data[5] = 0;
-	memcpy(cmd_msg.data+11, sdr->data, (MAX_SDR_FETCH - 5));
-	cmd_msg.data_len = 11 + (MAX_SDR_FETCH - 5);
-	sdrs->sdr_data_write = MAX_SDR_FETCH - 5;
+	memcpy(cmd_msg.data+11, sdr->data, (sdrs->fetch_size - 5));
+	cmd_msg.data_len = 11 + (sdrs->fetch_size - 5);
+	sdrs->sdr_data_write = sdrs->fetch_size - 5;
 	return ipmi_mc_send_command(mc, sdrs->lun, &cmd_msg,
 				    handle_sdr_write, sdr);
     }
@@ -1895,7 +2118,7 @@ handle_sdr_write(ipmi_mc_t  *mc,
     if (!mc) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
 		 "MC went away while SDR fetch was in progress(5)");
-	save_complete(sdrs, ENXIO);
+	save_complete(sdrs, ECANCELED);
 	goto out;
     }
 	
@@ -1905,7 +2128,7 @@ handle_sdr_write(ipmi_mc_t  *mc,
 	if (sdrs->fetch_retry_count > MAX_SDR_FETCH_RETRIES) {
 	    ipmi_log(IPMI_LOG_ERR_INFO,
 		     "handle_sdr_write: Lost reservation too many times");
-	    save_complete(sdrs, EBUSY);
+	    save_complete(sdrs, EAGAIN);
 	    goto out;
 	} else {
 	    rv = start_save(sdrs, mc);
@@ -1934,7 +2157,7 @@ handle_sdr_write(ipmi_mc_t  *mc,
     ipmi_set_uint16(cmd_msg.data+2, sdrs->curr_rec_id);
     cmd_msg.data[4] = sdrs->sdr_data_write;
     wleft = sdr->length - sdrs->sdr_data_write;
-    if (wleft >= MAX_SDR_FETCH) {
+    if (wleft >= sdrs->fetch_size) {
 	cmd_msg.data[5] = 1;
 	memcpy(cmd_msg.data+6, sdr->data+sdrs->sdr_data_write, wleft);
 	cmd_msg.data_len = 6 + wleft;
@@ -1942,9 +2165,10 @@ handle_sdr_write(ipmi_mc_t  *mc,
 				  handle_sdr_write_done, sdr);
     } else {
 	cmd_msg.data[5] = 0;
-	memcpy(cmd_msg.data+6, sdr->data+sdrs->sdr_data_write, MAX_SDR_FETCH);
-	cmd_msg.data_len = 6 + MAX_SDR_FETCH;
-	sdrs->sdr_data_write += MAX_SDR_FETCH;
+	memcpy(cmd_msg.data+6, sdr->data+sdrs->sdr_data_write,
+	       sdrs->fetch_size);
+	cmd_msg.data_len = 6 + sdrs->fetch_size;
+	sdrs->sdr_data_write += sdrs->fetch_size;
 	rv = ipmi_mc_send_command(mc, sdrs->lun, &cmd_msg,
 				  handle_sdr_write, sdr);
     }
@@ -1982,7 +2206,7 @@ handle_sdr_write_done(ipmi_mc_t  *mc,
     if (!mc) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
 		 "MC went away while SDR fetch was in progress(6)");
-	save_complete(sdrs, ENXIO);
+	save_complete(sdrs, ECANCELED);
 	goto out;
     }
 	
@@ -1992,7 +2216,7 @@ handle_sdr_write_done(ipmi_mc_t  *mc,
 	if (sdrs->fetch_retry_count > MAX_SDR_FETCH_RETRIES) {
 	    ipmi_log(IPMI_LOG_ERR_INFO,
 		     "handle_sdr_write_done: Lost reservation too many times");
-	    save_complete(sdrs, EBUSY);
+	    save_complete(sdrs, EAGAIN);
 	    goto out;
 	} else {
 	    rv = start_save(sdrs, mc);
@@ -2056,7 +2280,7 @@ handle_sdr_clear(ipmi_mc_t  *mc,
     if (!mc) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
 		 "MC went away while SDR fetch was in progress(7)");
-	save_complete(sdrs, ENXIO);
+	save_complete(sdrs, ECANCELED);
 	goto out;
     }
 	
@@ -2110,7 +2334,7 @@ handle_save_reservation(ipmi_mc_t  *mc,
     if (!mc) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
 		 "MC went away while SDR fetch was in progress(8)");
-	save_complete(sdrs, ENXIO);
+	save_complete(sdrs, ECANCELED);
 	goto out;
     }
 	
@@ -2263,7 +2487,6 @@ sdr_save_cb(ipmi_mc_t *mc, void *cb_data)
     sdr_unlock(sdrs);
 
  out:
-    ipmi_read_unlock();
     if (info->rv)
 	ipmi_mem_free(elem);
 }
