@@ -7,7 +7,7 @@
  *         Corey Minyard <minyard@mvista.com>
  *         source@mvista.com
  *
- * Copyright 2002,2003 MontaVista Software Inc.
+ * Copyright 2002,2003,2004,2005 MontaVista Software Inc.
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public License
@@ -34,15 +34,16 @@
 #include <string.h>
 
 #include <OpenIPMI/ipmiif.h>
-#include <OpenIPMI/ipmi_event.h>
-#include <OpenIPMI/ipmi_sel.h>
 #include <OpenIPMI/ipmi_msgbits.h>
-#include <OpenIPMI/ipmi_domain.h>
-#include <OpenIPMI/ipmi_mc.h>
 #include <OpenIPMI/ipmi_err.h>
-#include <OpenIPMI/ipmi_int.h>
-#include <OpenIPMI/opq.h>
-#include <OpenIPMI/ilist.h>
+
+#include <OpenIPMI/internal/opq.h>
+#include <OpenIPMI/internal/ilist.h>
+#include <OpenIPMI/internal/ipmi_int.h>
+#include <OpenIPMI/internal/ipmi_event.h>
+#include <OpenIPMI/internal/ipmi_sel.h>
+#include <OpenIPMI/internal/ipmi_domain.h>
+#include <OpenIPMI/internal/ipmi_mc.h>
 
 #define MAX_SEL_FETCH_RETRIES 10
 
@@ -63,6 +64,8 @@ typedef struct sel_event_holder_s
     ipmi_event_t *event;
 } sel_event_holder_t;
 
+#define SEL_NAME_LEN (IPMI_DOMAIN_NAME_LEN + 32)
+
 struct ipmi_sel_info_s
 {
     ipmi_mcid_t mc;
@@ -75,6 +78,14 @@ struct ipmi_sel_info_s
     uint16_t entries;
     uint32_t last_addition_timestamp;
     uint32_t last_erase_timestamp;
+
+    /* These are here to store the timestamps until the operation
+       completes.  Successfully.  Otherwise, if we restart the fetch,
+       it will have the timestamps set wrong and won't do the
+       fetch. */
+    uint32_t curr_addition_timestamp;
+    uint32_t curr_erase_timestamp;
+
     uint16_t free_bytes;
     unsigned int overflow : 1;
     unsigned int supports_delete_sel : 1;
@@ -129,6 +140,23 @@ struct ipmi_sel_info_s
 
     ipmi_sel_new_event_handler_cb new_event_handler;
     void                          *new_event_cb_data;
+
+    char name[SEL_NAME_LEN];
+
+    ipmi_domain_stat_t *sel_good_scans;
+    ipmi_domain_stat_t *sel_scan_lost_reservation;
+    ipmi_domain_stat_t *sel_fail_scan_lost_reservation;
+    ipmi_domain_stat_t *sel_received_events;
+    ipmi_domain_stat_t *sel_fetch_errors;
+
+    ipmi_domain_stat_t *sel_good_clears;
+    ipmi_domain_stat_t *sel_clear_lost_reservation;
+    ipmi_domain_stat_t *sel_clear_errors;
+
+    ipmi_domain_stat_t *sel_good_deletes;
+    ipmi_domain_stat_t *sel_delete_lost_reservation;
+    ipmi_domain_stat_t *sel_fail_delete_lost_reservation;
+    ipmi_domain_stat_t *sel_delete_errors;
 };
 
 static inline void sel_lock(ipmi_sel_info_t *sel)
@@ -229,6 +257,8 @@ ipmi_sel_alloc(ipmi_mc_t       *mc,
     }
     memset(sel, 0, sizeof(*sel));
 
+    strncpy(sel->name, _ipmi_mc_name(mc), sizeof(sel->name));
+
     sel->events = alloc_ilist();
     if (!sel->events) {
 	rv = ENOMEM;
@@ -273,6 +303,42 @@ ipmi_sel_alloc(ipmi_mc_t       *mc,
 	    ipmi_mem_free(sel);
 	}
     } else {
+	ipmi_domain_stat_register(domain, "sel_good_scans",
+				  _ipmi_mc_name(mc),
+				  &sel->sel_good_scans);
+	ipmi_domain_stat_register(domain, "sel_scan_lost_reservation",
+				  _ipmi_mc_name(mc),
+				  &sel->sel_scan_lost_reservation);
+	ipmi_domain_stat_register(domain, "sel_fail_scan_lost_reservation",
+				  _ipmi_mc_name(mc),
+				  &sel->sel_fail_scan_lost_reservation);
+	ipmi_domain_stat_register(domain, "sel_received_events",
+				  _ipmi_mc_name(mc),
+				  &sel->sel_received_events);
+	ipmi_domain_stat_register(domain, "sel_fetch_errors",
+				  _ipmi_mc_name(mc),
+				  &sel->sel_fetch_errors);
+	ipmi_domain_stat_register(domain, "sel_good_clears",
+				  _ipmi_mc_name(mc),
+				  &sel->sel_good_clears);
+	ipmi_domain_stat_register(domain, "sel_clear_lost_reservation",
+				  _ipmi_mc_name(mc),
+				  &sel->sel_clear_lost_reservation);
+	ipmi_domain_stat_register(domain, "sel_clear_errors",
+				  _ipmi_mc_name(mc),
+				  &sel->sel_clear_errors);
+	ipmi_domain_stat_register(domain, "sel_good_deletes",
+				  _ipmi_mc_name(mc),
+				  &sel->sel_good_deletes);
+	ipmi_domain_stat_register(domain, "sel_delete_lost_reservation",
+				  _ipmi_mc_name(mc),
+				  &sel->sel_delete_lost_reservation);
+	ipmi_domain_stat_register(domain, "sel_fail_delete_lost_reservation",
+				  _ipmi_mc_name(mc),
+				  &sel->sel_fail_delete_lost_reservation);
+	ipmi_domain_stat_register(domain, "sel_delete_errors",
+				  _ipmi_mc_name(mc),
+				  &sel->sel_delete_errors);
 	*new_sel = sel;
     }
     return rv;
@@ -297,6 +363,31 @@ internal_destroy_sel(ipmi_sel_info_t *sel)
     if (sel->sel_lock)
 	sel->os_hnd->destroy_lock(sel->os_hnd, sel->sel_lock);
 
+    if (sel->sel_good_scans)
+	ipmi_domain_stat_put(sel->sel_good_scans);
+    if (sel->sel_scan_lost_reservation)
+	ipmi_domain_stat_put(sel->sel_scan_lost_reservation);
+    if (sel->sel_fail_scan_lost_reservation)
+	ipmi_domain_stat_put(sel->sel_fail_scan_lost_reservation);
+    if (sel->sel_received_events)
+	ipmi_domain_stat_put(sel->sel_received_events);
+    if (sel->sel_fetch_errors)
+	ipmi_domain_stat_put(sel->sel_fetch_errors);
+    if (sel->sel_good_clears)
+	ipmi_domain_stat_put(sel->sel_good_clears);
+    if (sel->sel_clear_lost_reservation)
+	ipmi_domain_stat_put(sel->sel_clear_lost_reservation);
+    if (sel->sel_clear_errors)
+	ipmi_domain_stat_put(sel->sel_clear_errors);
+    if (sel->sel_good_deletes)
+	ipmi_domain_stat_put(sel->sel_good_deletes);
+    if (sel->sel_delete_lost_reservation)
+	ipmi_domain_stat_put(sel->sel_delete_lost_reservation);
+    if (sel->sel_fail_delete_lost_reservation)
+	ipmi_domain_stat_put(sel->sel_fail_delete_lost_reservation);
+    if (sel->sel_delete_errors)
+	ipmi_domain_stat_put(sel->sel_delete_errors);
+
     /* Do this after we have gotten rid of all external dependencies,
        but before it is free. */
     if (sel->destroy_handler)
@@ -339,34 +430,43 @@ static void
 fetch_complete(ipmi_sel_info_t *sel, int err)
 {
     sel_fetch_handler_t *elem, *next;
+    int                 sels_changed;
+    unsigned int        num_sels;
 
     if (sel->in_destroy)
 	goto out;
 
+    sels_changed = sel->sels_changed;
+    num_sels = sel->num_sels;
+
     elem = sel->fetch_handlers;
     sel->fetch_handlers = NULL;
     sel->fetched = 1;
     sel->in_fetch = 0;
+    sel_unlock(sel);
+
     while (elem) {
 	next = elem->next;
 	elem->next = NULL;
 	if (elem->handler)
 	    elem->handler(sel,
 		          err,
-		          sel->sels_changed,
-		          sel->num_sels,
+		          sels_changed,
+		          num_sels,
 		          elem->cb_data);
 	ipmi_mem_free(elem);
 	elem = next;
     }
 
     if (sel->destroyed) {
+	sel_lock(sel);
 	internal_destroy_sel(sel);
 	/* Previous call releases lock. */
 	return;
     }
 
     opq_op_done(sel->opq);
+    return;
 
  out:
     sel_unlock(sel);
@@ -400,23 +500,35 @@ handle_sel_clear(ipmi_mc_t  *mc,
     sel_lock(sel);
     if (sel->destroyed) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
+		 "%ssel.c(handle_sel_clear): "
 		 "SEL info was destroyed while an operation was in"
-		 " progress(1)");
+		 " progress(1)", sel->name);
 	fetch_complete(sel, ECANCELED);
 	goto out;
     }
 
     if (!mc) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
-		 "MC went away while SEL op was in progress");
-        fetch_complete(sel, ENXIO);
+		 "%ssel.c(handle_sel_clear): "
+		 "MC went away while SEL op was in progress",
+		 sel->name);
+        fetch_complete(sel, ECANCELED);
 	goto out;
     }
 
     if (rsp->data[0] == 0) {
+	if (sel->sel_good_clears)
+	    ipmi_domain_stat_add(sel->sel_good_clears, 1);
+
 	/* Success!  We can free the data. */
 	free_deleted_events(sel->events);
 	sel->del_sels = 0;
+    } else if (rsp->data[0] == IPMI_INVALID_RESERVATION_CC) {
+	if (sel->sel_clear_lost_reservation)
+	    ipmi_domain_stat_add(sel->sel_clear_lost_reservation, 1);
+    } else {
+	if (sel->sel_clear_errors)
+	    ipmi_domain_stat_add(sel->sel_clear_errors, 1);
     }
 
     fetch_complete(sel, 0);
@@ -467,29 +579,41 @@ handle_sel_data(ipmi_mc_t  *mc,
     sel_lock(sel);
     if (sel->destroyed) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
+		 "%ssel.c(handle_sel_data): "
 		 "SEL info was destroyed while an operation was in"
-		 " progress(2)");
+		 " progress(2)", sel->name);
 	fetch_complete(sel, ECANCELED);
 	goto out;
     }
 
     if (!mc) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
-		 "handle_sel_data: MC went away while SEL op was in progress");
-        fetch_complete(sel, ENXIO);
+		 "%ssel.c(handle_sel_data): "
+		 "handle_sel_data: MC went away while SEL op was in progress",
+		 sel->name);
+        fetch_complete(sel, ECANCELED);
 	goto out;
     }
-
+	
     if (rsp->data[0] == IPMI_INVALID_RESERVATION_CC) {
 	/* We lost our reservation, restart the operation.  Only do
            this so many times, in order to guarantee that this
-           completes. */
+           completes.  Note that we do this so many times *per fetch*,
+           we do not reset the counter if we get a successful
+           operation.  This is because if this happens a lot during a
+           fetch, there is heavy contention for the SEL and someone
+           needs to drop out to allow everyone else to continue. */
 	sel->fetch_retry_count++;
+	if (sel->sel_scan_lost_reservation)
+	    ipmi_domain_stat_add(sel->sel_scan_lost_reservation, 1);
 	if (sel->fetch_retry_count > MAX_SEL_FETCH_RETRIES) {
+	    if (sel->sel_fail_scan_lost_reservation)
+		ipmi_domain_stat_add(sel->sel_fail_scan_lost_reservation, 1);
 	    ipmi_log(IPMI_LOG_ERR_INFO,
-		     "handle_sel_data: "
-		     "Too many lost reservations in SEL fetch");
-	    fetch_complete(sel, EBUSY);
+		     "%ssel.c(handle_sel_data): "
+		     "Too many lost reservations in SEL fetch",
+		     sel->name);
+	    fetch_complete(sel, EAGAIN);
 	    goto out;
 	} else {
 	    start_fetch(elem, 0);
@@ -497,9 +621,21 @@ handle_sel_data(ipmi_mc_t  *mc,
 	}
     }
     if (rsp->data[0] != 0) {
+	if (sel->curr_rec_id != 0) {
+	    /* If we get a fetch error and it is not a lost
+	       reservation, it may be that another system deleted our
+	       "current" record.  Start over from the beginning of the
+	       SEL. */
+	    sel->curr_rec_id = 0;
+	    del_event = NULL;
+	    goto start_request_sel_data;
+	}
+	if (sel->sel_fetch_errors)
+	    ipmi_domain_stat_add(sel->sel_fetch_errors, 1);
 	ipmi_log(IPMI_LOG_ERR_INFO,
-		 "handle_sel_data: IPMI error from SEL fetch: %x",
-		 rsp->data[0]);
+		 "%ssel.c(handle_sel_data): "
+		 "IPMI error from SEL fetch: %x",
+		 sel->name, rsp->data[0]);
 	fetch_complete(sel, IPMI_IPMI_ERR_VAL(rsp->data[0]));
 	goto out;
     }
@@ -518,22 +654,35 @@ handle_sel_data(ipmi_mc_t  *mc,
 				 timestamp,
 				 rsp->data+6,
 				 13);
+    if (!del_event) {
+	ipmi_log(IPMI_LOG_ERR_INFO,
+		 "%ssel.c(handle_sel_data): "
+		 "Could not allocate event for SEL",
+		 sel->name);
+	fetch_complete(sel, ENOMEM);
+	goto out;
+    }
+
+    if ((timestamp > 0) && (timestamp < ipmi_mc_get_startup_SEL_time(mc)))
+	ipmi_event_set_is_old(del_event, 1);
 
     holder = find_event(sel->events, record_id);
     if (!holder) {
 	holder = ipmi_mem_alloc(sizeof(*holder));
 	if (!holder) {
 	    ipmi_log(IPMI_LOG_ERR_INFO,
-		     "handle_sel_data: "
-		     "Could not allocate log information for SEL");
+		     "%ssel.c(handle_sel_data): "
+		     "Could not allocate log information for SEL",
+		     sel->name);
 	    fetch_complete(sel, ENOMEM);
 	    goto out;
 	}
 	if (!ilist_add_tail(sel->events, holder, NULL)) {
 	    ipmi_mem_free(holder);
 	    ipmi_log(IPMI_LOG_ERR_INFO,
-		     "handle_sel_data: "
-		     "Could not link log onto the log linked list");
+		     "%ssel.c(handle_sel_data): "
+		     "Could not link log onto the log linked list",
+		     sel->name);
 	    fetch_complete(sel, ENOMEM);
 	    goto out;
 	}
@@ -541,6 +690,8 @@ handle_sel_data(ipmi_mc_t  *mc,
 	holder->deleted = 0;
 	event_is_new = 1;
 	sel->num_sels++;
+	if (sel->sel_received_events)
+	    ipmi_domain_stat_add(sel->sel_received_events, 1);
     } else if (event_cmp(del_event, holder->event) != 0) {
 	/* It's a new event in an old slot, so overwrite the old
            event. */
@@ -553,23 +704,37 @@ handle_sel_data(ipmi_mc_t  *mc,
 	    sel->del_sels--;
 	}
 	event_is_new = 1;
+	if (sel->sel_received_events)
+	    ipmi_domain_stat_add(sel->sel_received_events, 1);
     } else {
 	ipmi_event_free(del_event);
     }
 
     if (sel->next_rec_id == 0xFFFF) {
+	/* Only set the timestamps if the SEL fetch completed
+	   successfully.  If we were unsuccessful, we want to redo the
+	   operation so don't set the timestamps. */
+	sel->last_addition_timestamp = sel->curr_addition_timestamp;
+	sel->last_erase_timestamp = sel->curr_erase_timestamp;
+
 	/* To avoid confusion, deliver the event before we deliver fetch
            complete. */
-	if (event_is_new)
-	    if (sel->new_event_handler)
-		sel->new_event_handler(sel,
-				       mc,
-				       del_event,
-				       sel->new_event_cb_data);
+	if (event_is_new && sel->new_event_handler)
+	    sel->new_event_handler(sel,
+				   mc,
+				   del_event,
+				   sel->new_event_cb_data);
+
+	if (sel->sel_good_scans)
+	    ipmi_domain_stat_add(sel->sel_good_scans, 1);
+
 	/* If the operation completed successfully and everything in
-	   our SEL is deleted, then clear it with our old
-	   reservation. */
-	if ((sel->num_sels == 0) && (!ilist_empty(sel->events))) {
+	   our SEL is deleted, then clear it with our old reservation.
+	   We also do the clear if the overflow flag is set; on some
+	   systems this operation clears the overflow flag. */
+	if ((sel->num_sels == 0)
+	    && ((!ilist_empty(sel->events)) || sel->overflow))
+	{
 	    /* We don't care if this fails, because it will just
 	       happen again later if it does. */
 	    rv = send_sel_clear(elem, mc);
@@ -586,6 +751,7 @@ handle_sel_data(ipmi_mc_t  *mc,
     }
     sel->curr_rec_id = sel->next_rec_id;
 
+ start_request_sel_data:
     /* Request some more data. */
     cmd_msg.data = cmd_data;
     cmd_msg.netfn = IPMI_STORAGE_NETFN;
@@ -598,20 +764,34 @@ handle_sel_data(ipmi_mc_t  *mc,
     rv = ipmi_mc_send_command(mc, sel->lun, &cmd_msg, handle_sel_data, elem);
     if (rv) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
-		 "handle_sel_data: Could not send SEL fetch command: %x", rv);
+		 "%ssel.c(handle_sel_clear): "
+		 "Could not send SEL fetch command: %x", sel->name, rv);
 	fetch_complete(sel, rv);
 	goto out;
     }
 
-    if (event_is_new)
-	if (sel->new_event_handler)
-	    sel->new_event_handler(sel, mc, del_event, sel->new_event_cb_data);
+    if (event_is_new && sel->new_event_handler)
+	sel->new_event_handler(sel, mc, del_event, sel->new_event_cb_data);
  out_unlock:
     sel_unlock(sel);
  out:
     return;
 }
 
+/* Cheap hacks for broken hardware. */
+static void
+sel_fixups(ipmi_mc_t *mc, ipmi_sel_info_t *sel)
+{
+    unsigned int mfg_id, product_id;
+
+    /* Fixups */
+    mfg_id = ipmi_mc_manufacturer_id(mc);
+    product_id = ipmi_mc_product_id(mc);
+    if ((mfg_id == 0x157) && (product_id == 0x841))
+	/* Intel ATCA CMM mistakenly reports that it supports delete SEL */
+	sel->supports_delete_sel = 0;
+}
+
 static void
 handle_sel_info(ipmi_mc_t  *mc,
 		ipmi_msg_t *rsp,
@@ -629,29 +809,38 @@ handle_sel_info(ipmi_mc_t  *mc,
     sel_lock(sel);
     if (sel->destroyed) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
-		 "handle_sel_info: "
-		 "SEL info was destroyed while an operation was in progress");
+		 "%ssel.c(handle_sel_info): "
+		 "SEL info was destroyed while an operation was in progress",
+		 sel->name);
 	fetch_complete(sel, ECANCELED);
 	goto out;
     }
 
     if (!mc) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
-		 "handle_sel_info: MC went away while SEL op was in progress");
-        fetch_complete(sel, ENXIO);
+		 "%ssel.c(handle_sel_info): "
+		 "MC went away while SEL op was in progress",
+		 sel->name);
+        fetch_complete(sel, ECANCELED);
 	goto out;
     }
 	
     if (rsp->data[0] != 0) {
+	if (sel->sel_fetch_errors)
+	    ipmi_domain_stat_add(sel->sel_fetch_errors, 1);
 	ipmi_log(IPMI_LOG_ERR_INFO,
-		 "handle_sel_info: IPMI error from SEL info fetch: %x",
-		 rsp->data[0]);
+		 "%ssel.c(handle_sel_info): "
+		 "IPMI error from SEL info fetch: %x",
+		 sel->name, rsp->data[0]);
 	fetch_complete(sel, IPMI_IPMI_ERR_VAL(rsp->data[0]));
 	goto out;
     }
 
     if (rsp->data_len < 15) {
-	ipmi_log(IPMI_LOG_ERR_INFO, "handle_sel_info: SEL info too short");
+	if (sel->sel_fetch_errors)
+	    ipmi_domain_stat_add(sel->sel_fetch_errors, 1);
+	ipmi_log(IPMI_LOG_ERR_INFO,
+		 "%ssel.c(handle_sel_info): SEL info too short", sel->name);
 	fetch_complete(sel, EINVAL);
 	goto out;
     }
@@ -671,15 +860,20 @@ handle_sel_info(ipmi_mc_t  *mc,
     add_timestamp = ipmi_get_uint32(rsp->data + 6);
     erase_timestamp = ipmi_get_uint32(rsp->data + 10);
 
+    sel_fixups(mc, sel);
+
     /* If the timestamps still match, no need to re-fetch the repository */
     if (sel->fetched
 	&& (add_timestamp == sel->last_addition_timestamp)
 	&& (erase_timestamp == sel->last_erase_timestamp))
     {
 	/* If the operation completed successfully and everything in
-	   our SEL is deleted, then clear it with our old
-	   reservation. */
-	if ((sel->num_sels == 0) && (!ilist_empty(sel->events))) {
+	   our SEL is deleted, then clear it with our old reservation.
+	   We also do the clear if the overflow flag is set; on some
+	   systems this operation clears the overflow flag. */
+	if ((sel->num_sels == 0)
+	    && ((!ilist_empty(sel->events)) || sel->overflow))
+	{
 	    /* We don't care if this fails, because it will just
 	       happen again later if it does. */
 	    rv = send_sel_clear(elem, mc);
@@ -703,14 +897,20 @@ handle_sel_info(ipmi_mc_t  *mc,
 	sel->curr_rec_id = 0;
     }
 
-    sel->last_addition_timestamp = add_timestamp;
-    sel->last_erase_timestamp = erase_timestamp;
+    sel->curr_addition_timestamp = add_timestamp;
+    sel->curr_erase_timestamp = erase_timestamp;
 
     sel->sels_changed = 1;
     sel->next_rec_id = 0;
 
     if (fetched_num_sels == 0) {
 	/* No sels, so there's nothing to do. */
+
+	/* Set the timestamps here, because they are not the same, but
+	   there was nothing to do. */
+	sel->last_addition_timestamp = sel->curr_addition_timestamp;
+	sel->last_erase_timestamp = sel->curr_erase_timestamp;
+
 	fetch_complete(sel, 0);
 	goto out;
     }
@@ -727,8 +927,9 @@ handle_sel_info(ipmi_mc_t  *mc,
     rv = ipmi_mc_send_command(mc, sel->lun, &cmd_msg, handle_sel_data, elem);
     if (rv) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
-		 "handle_sel_info: Could not send first SEL fetch command: %x",
-		 rv);
+		 "%ssel.c(handle_sel_info): "
+		 "Could not send first SEL fetch command: %x",
+		 sel->name, rv);
 	fetch_complete(sel, rv);
 	goto out;
     }
@@ -752,28 +953,36 @@ sel_handle_reservation(ipmi_mc_t  *mc,
     sel_lock(sel);
     if (sel->destroyed) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
-		 "sel_handle_reservation: "
-		 "SEL info was destroyed while an operation was in progress");
+		 "%ssel.c(sel_handle_reservation): "
+		 "SEL info was destroyed while an operation was in progress",
+		 sel->name);
 	fetch_complete(sel, ECANCELED);
 	goto out;
     }
 
     if (!mc) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
-		 "sel_handle_reservation: "
-		 "MC went away while SEL op was in progress");
-        fetch_complete(sel, ENXIO);
+		 "%ssel.c(sel_handle_reservation): "
+		 "MC went away while SEL op was in progress",
+		 sel->name);
+        fetch_complete(sel, ECANCELED);
 	goto out;
     }
 	
     if (rsp->data[0] != 0) {
+	if (sel->sel_fetch_errors)
+	    ipmi_domain_stat_add(sel->sel_fetch_errors, 1);
 	ipmi_log(IPMI_LOG_ERR_INFO,
-		 "sel_handle_reservation: Failed getting reservation");
+		 "%ssel.c(sel_handle_reservation): "
+		 "Failed getting reservation", sel->name);
 	fetch_complete(sel, ENOSYS);
 	goto out;
     } else if (rsp->data_len < 3) {
+	if (sel->sel_fetch_errors)
+	    ipmi_domain_stat_add(sel->sel_fetch_errors, 1);
 	ipmi_log(IPMI_LOG_ERR_INFO,
-		 "sel_handle_reservation: got invalid reservation length");
+		 "%ssel.c(sel_handle_reservation): "
+		 "got invalid reservation length", sel->name);
 	fetch_complete(sel, EINVAL);
 	goto out;
     }
@@ -788,8 +997,8 @@ sel_handle_reservation(ipmi_mc_t  *mc,
     rv = ipmi_mc_send_command(mc, sel->lun, &cmd_msg, handle_sel_info, elem);
     if (rv) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
-		 "sel_handle_reservation: "
-		 "Could not send SEL info command: %x", rv);
+		 "%ssel.c(sel_handle_reservation): "
+		 "Could not send SEL info command: %x", sel->name, rv);
 	fetch_complete(sel, rv);
 	goto out;
     }
@@ -810,8 +1019,9 @@ start_fetch_cb(ipmi_mc_t *mc, void *cb_data)
     sel_lock(sel);
     if (sel->destroyed) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
-		 "start_fetch: "
-		 "SEL info was destroyed while an operation was in progress");
+		 "%ssel.c(start_fetch_cb): "
+		 "SEL info was destroyed while an operation was in progress",
+		 sel->name);
 	fetch_complete(sel, ECANCELED);
 	goto out;
     }
@@ -839,8 +1049,8 @@ start_fetch_cb(ipmi_mc_t *mc, void *cb_data)
 
     if (rv) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
-		 "start_fetch: could not send cmd: %x",
-		 rv);
+		 "%ssel.c(start_fetch_cb): could not send cmd: %x",
+		 sel->name, rv);
 	fetch_complete(sel, ECANCELED);
 	goto out;
     }
@@ -858,8 +1068,9 @@ start_fetch(void *cb_data, int shutdown)
 
     if (shutdown) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
-		 "start_fetch: "
-		 "SEL info was destroyed while an operation was in progress");
+		 "%ssel.c(start_fetch): "
+		 "SEL info was destroyed while an operation was in progress",
+		 elem->sel->name);
 	sel_lock(elem->sel);
 	fetch_complete(elem->sel, ECANCELED);
 	return;
@@ -869,7 +1080,8 @@ start_fetch(void *cb_data, int shutdown)
        deadlock. */
     rv = ipmi_mc_pointer_cb(elem->sel->mc, start_fetch_cb, elem);
     if (rv) {
-	ipmi_log(IPMI_LOG_ERR_INFO, "start_fetch: MC is not valid");
+	ipmi_log(IPMI_LOG_ERR_INFO,
+		 "%ssel.c(start_fetch): MC is not valid", elem->sel->name);
 	sel_lock(elem->sel);
 	fetch_complete(elem->sel, rv);
     }
@@ -892,7 +1104,8 @@ ipmi_sel_get_cb(ipmi_mc_t *mc, void *cb_data)
 
     if (!ipmi_mc_sel_device_support(mc)) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
-		 "ipmi_sel_get: No support for the system event log");
+		 "%ssel.c(ipmi_sel_get_cb): "
+		 "No support for the system event log", sel->name);
 	info->rv = ENOSYS;
 	return;
     }
@@ -906,12 +1119,14 @@ ipmi_sel_get_cb(ipmi_mc_t *mc, void *cb_data)
 	sel->in_fetch = 1;
 	sel->sels_changed = 0;
 
-	if (!opq_new_op(sel->opq, start_fetch, elem, 0)) {
-	    info->rv = ENOMEM;
-	    goto out_unlock;
-	}
 	elem->next = NULL;
 	sel->fetch_handlers = elem;
+	sel_unlock(sel);
+	if (!opq_new_op(sel->opq, start_fetch, elem, 0)) {
+	    sel->fetch_handlers = NULL;
+	    info->rv = ENOMEM;
+	}
+	goto out;
     } else if (elem->handler) {
 	/* Add it to the list of waiting fetch handlers, if it has a
 	   handler. */
@@ -921,8 +1136,9 @@ ipmi_sel_get_cb(ipmi_mc_t *mc, void *cb_data)
 	info->rv = EEXIST;
     }
 
- out_unlock:
     sel_unlock(sel);
+ out:
+    return;
 }
 
 int
@@ -937,7 +1153,8 @@ ipmi_sel_get(ipmi_sel_info_t     *sel,
     elem = ipmi_mem_alloc(sizeof(*elem));
     if (!elem) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
-		 "ipmi_sel_get: could not allocate the sel element");
+		 "%ssel.c(ipmi_sel_get): "
+		 "could not allocate the sel element", sel->name);
 	return ENOMEM;
     }
 
@@ -977,7 +1194,8 @@ typedef struct sel_cb_handler_data_s
     ipmi_event_t          *event;
 } sel_cb_handler_data_t;
 
-static int send_reserve_sel(sel_cb_handler_data_t *data, ipmi_mc_t *mc);
+static int send_reserve_sel_for_delete(sel_cb_handler_data_t *data,
+				       ipmi_mc_t *mc);
 
 static void
 sel_op_done(sel_cb_handler_data_t *data,
@@ -995,8 +1213,8 @@ sel_op_done(sel_cb_handler_data_t *data,
 	/* This will unlock the lock. */
 	internal_destroy_sel(sel);
     } else {
-	opq_op_done(sel->opq);
 	sel_unlock(sel);
+	opq_op_done(sel->opq);
     }
     if (data->event)
 	ipmi_event_free(data->event);
@@ -1015,24 +1233,28 @@ handle_sel_delete(ipmi_mc_t  *mc,
     sel_lock(sel);
     if (sel->destroyed) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
-		 "handle_sel_delete: "
-		 "SEL info was destroyed while an operation was in progress");
+		 "%ssel.c(handle_sel_delete): "
+		 "SEL info was destroyed while an operation was in progress",
+		 sel->name);
 	sel_op_done(data, ECANCELED);
 	goto out;
     }
 
     if (!mc) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
-		 "handle_sel_delete: "
-		 "MC went away while SEL fetch was in progress");
-	sel_op_done(data, ENXIO);
+		 "%ssel.c(handle_sel_delete): "
+		 "MC went away while SEL fetch was in progress",
+		 sel->name);
+	sel_op_done(data, ECANCELED);
 	goto out;
     }
 
     /* Special return codes. */
     if (rsp->data[0] == 0x80) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
-		 "handle_sel_delete: Operation not supported on SEL delete");
+		 "%ssel.c(handle_sel_delete): "
+		 "Operation not supported on SEL delete",
+		 sel->name);
 	rv = ENOSYS;
     } else if (rsp->data[0] == 0x81) {
 	/* The SEL is being erased, so by definition the log will be
@@ -1044,15 +1266,24 @@ handle_sel_delete(ipmi_mc_t  *mc,
     } else if ((data->count < MAX_DEL_RESERVE_RETRIES)
 	       && (rsp->data[0] == IPMI_INVALID_RESERVATION_CC))
     {
+	if (sel->sel_delete_lost_reservation)
+	    ipmi_domain_stat_add(sel->sel_delete_lost_reservation, 1);
 	/* Lost our reservation, retry the operation. */
 	data->count++;
-	rv = send_reserve_sel(data, mc);
+	rv = send_reserve_sel_for_delete(data, mc);
 	if (!rv)
 	    goto out_unlock;
     } else if (rsp->data[0]) {
+	if (rsp->data[0] == IPMI_INVALID_RESERVATION_CC) {
+	    if (sel->sel_fail_delete_lost_reservation)
+		ipmi_domain_stat_add(sel->sel_fail_delete_lost_reservation, 1);
+	} else {
+	    if (sel->sel_delete_errors)
+		ipmi_domain_stat_add(sel->sel_delete_errors, 1);
+	}
 	ipmi_log(IPMI_LOG_ERR_INFO,
-		 "handle_sel_delete: "
-		 "IPMI error from SEL delete: %x", rsp->data[0]);
+		 "%ssel.c(handle_sel_delete): "
+		 "IPMI error from SEL delete: %x", sel->name, rsp->data[0]);
 	rv = IPMI_IPMI_ERR_VAL(rsp->data[0]);
     } else {
 	/* We deleted the entry, so remove it from our database. */
@@ -1111,17 +1342,19 @@ handle_sel_check(ipmi_mc_t  *mc,
     sel_lock(sel);
     if (sel->destroyed) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
-		 "handle_sel_check: "
-		 "SEL info was destroyed while an operation was in progress");
+		 "%ssel.c(handle_sel_check): "
+		 "SEL info was destroyed while an operation was in progress",
+		 sel->name);
 	sel_op_done(data, ECANCELED);
 	goto out;
     }
 
     if (!mc) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
-		 "handle_sel_check: "
-		 "MC went away while SEL fetch was in progress");
-	sel_op_done(data, ENXIO);
+		 "%ssel.c(handle_sel_check): "
+		 "MC went away while SEL fetch was in progress",
+		 sel->name);
+	sel_op_done(data, ECANCELED);
 	goto out;
     }
 
@@ -1131,9 +1364,11 @@ handle_sel_check(ipmi_mc_t  *mc,
 	sel_op_done(data, 0);
 	goto out;
     } else if (rsp->data[0]) {
+	if (sel->sel_delete_errors)
+	    ipmi_domain_stat_add(sel->sel_delete_errors, 1);
 	ipmi_log(IPMI_LOG_ERR_INFO,
-		 "handle_sel_check: IPMI error from SEL get: %x",
-		 rsp->data[0]);
+		 "%ssel.c(handle_sel_check): IPMI error from SEL get: %x",
+		 sel->name, rsp->data[0]);
 	sel_op_done(data, IPMI_IPMI_ERR_VAL(rsp->data[0]));
 	goto out;
     } else {
@@ -1152,7 +1387,8 @@ handle_sel_check(ipmi_mc_t  *mc,
 				    13);
 	if (!ch_event) {
 	    ipmi_log(IPMI_LOG_ERR_INFO,
-		     "handle_sel_check: Could not allocate memory");
+		     "%ssel.c(handle_sel_check): Could not allocate memory",
+		     sel->name);
 	    sel_op_done(data, ENOMEM);
 	    goto out;
 	}
@@ -1168,11 +1404,18 @@ handle_sel_check(ipmi_mc_t  *mc,
 	rv = send_del_sel(data, mc);
 	if (rv) {
 	    ipmi_log(IPMI_LOG_ERR_INFO,
-		     "Could not send SEL delete command: %x", rv);
+		     "%ssel.c(handle_sel_check): "
+		     "Could not send SEL delete command: %x",
+		     sel->name, rv);
 	    sel_op_done(data, rv);
 	    goto out;
-	}
+	} else if (data->record_id == sel->curr_rec_id)
+	    /* We are deleting our "current" record (used for finding
+	       the next record), make sure we start again from
+	       scratch on the next fetch. */
+	    sel->curr_rec_id = 0;
     }
+	
 
     sel_unlock(sel);
 
@@ -1214,21 +1457,27 @@ sel_reserved_for_delete(ipmi_mc_t  *mc,
     sel_lock(sel);
     if (sel->destroyed) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
-		 "sel_reserved_for_delete: "
-		 "SEL info was destroyed while an operation was in progress");
+		 "%ssel.c(sel_reserved_for_delete): "
+		 "SEL info was destroyed while an operation was in progress",
+		 sel->name);
 	sel_op_done(data, ECANCELED);
 	goto out;
     }
     if (!mc) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
-		 "MC went away while SEL fetch was in progress");
-	sel_op_done(data, ENXIO);
+		 "%ssel.c(sel_reserved_for_delete): "
+		 "MC went away while SEL fetch was in progress", sel->name);
+	sel_op_done(data, ECANCELED);
 	goto out;
     }
 
     if (rsp->data[0] != 0) {
+	if (sel->sel_delete_errors)
+	    ipmi_domain_stat_add(sel->sel_delete_errors, 1);
 	ipmi_log(IPMI_LOG_ERR_INFO,
-		 "IPMI error from SEL delete reservation: %x", rsp->data[0]);
+		 "%ssel.c(sel_reserved_for_delete): "
+		 "IPMI error from SEL delete reservation: %x",
+		 sel->name, rsp->data[0]);
 	sel_op_done(data, IPMI_IPMI_ERR_VAL(rsp->data[0]));
 	goto out;
     }
@@ -1237,7 +1486,8 @@ sel_reserved_for_delete(ipmi_mc_t  *mc,
     rv = send_check_sel(data, mc);
     if (rv) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
-		 "Could not send SEL get command: %x", rv);
+		 "%ssel.c(sel_reserved_for_delete): "
+		 "Could not send SEL get command: %x", sel->name, rv);
 	sel_op_done(data, rv);
 	goto out;
     }
@@ -1248,7 +1498,7 @@ sel_reserved_for_delete(ipmi_mc_t  *mc,
 }
 
 static int
-send_reserve_sel(sel_cb_handler_data_t *data, ipmi_mc_t *mc)
+send_reserve_sel_for_delete(sel_cb_handler_data_t *data, ipmi_mc_t *mc)
 {
     unsigned char   cmd_data[MAX_IPMI_DATA_SIZE];
     ipmi_msg_t      cmd_msg;
@@ -1271,24 +1521,25 @@ start_del_sel_cb(ipmi_mc_t *mc, void *cb_data)
     ipmi_sel_info_t       *sel = data->sel;
     int                   rv;
 
-    sel_lock(sel);
+    /* Called with SEL lock held. */
     if (sel->destroyed) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
-		 "start_del_sel_cb: "
-		 "SEL info was destroyed while an operation was in progress");
+		 "%ssel.c(start_del_sel_cb): "
+		 "SEL info was destroyed while an operation was in progress",
+		 sel->name);
 	sel_op_done(data, ECANCELED);
 	goto out;
     }
 
     if (data->sel->supports_reserve_sel)
-	rv = send_reserve_sel(data, mc);
+	rv = send_reserve_sel_for_delete(data, mc);
     else
 	rv = send_check_sel(data, mc);
 
     if (rv) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
-		 "start_del_sel_cb: could not send cmd: %x",
-		 rv);
+		 "%ssel.c(start_del_sel_cb): could not send cmd: %x",
+		 sel->name, rv);
 	sel_op_done(data, rv);
 	goto out;
     }
@@ -1308,20 +1559,23 @@ start_del_sel(void *cb_data, int shutdown)
     sel_lock(sel);
     if (shutdown) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
-		 "start_del_sel: "
-		 "SEL info was destroyed while an operation was in progress");
+		 "%ssel.c(start_del_sel): "
+		 "SEL info was destroyed while an operation was in progress",
+		 sel->name);
 	sel_op_done(data, ECANCELED);
 	goto out;
     }
 
     rv = ipmi_mc_pointer_cb(sel->mc, start_del_sel_cb, data);
     if (rv) {
+	sel_unlock(sel);
 	ipmi_log(IPMI_LOG_ERR_INFO,
-		 "start_del_sel: MC went away during delete");
+		 "%ssel.c(start_del_sel_cb): MC went away during delete",
+		 sel->name);
 	sel_op_done(data, ECANCELED);
 	goto out;
     }
-    sel_unlock(sel);
+
  out:
     return;
 }
@@ -1399,13 +1653,16 @@ sel_del_event_cb(ipmi_mc_t *mc, void *cb_data)
 
 	/* We don't return a return code, because we don't really
            care.  If this fails, it will just be handled later. */
+	sel_unlock(sel);
 	if (!opq_new_op(sel->opq, start_del_sel, data, 0))
 	    ipmi_mem_free(data);
     } else {
+	sel_unlock(sel);
 	/* Don't really delete the event, but report is as done. */
 	info->handler(sel, info->cb_data, 0);
 	ipmi_event_free(event);
     }
+    return;
 
  out_unlock:
     sel_unlock(sel);
@@ -1920,8 +2177,8 @@ sel_add_op_done(sel_add_cb_handler_data_t *data,
 	/* This will unlock the lock. */
 	internal_destroy_sel(sel);
     } else {
-	opq_op_done(sel->opq);
 	sel_unlock(sel);
+	opq_op_done(sel->opq);
     }
     if (data->event)
 	ipmi_event_free(data->event);
@@ -1939,33 +2196,34 @@ sel_add_event_done(ipmi_mc_t  *mc,
     sel_lock(sel);
     if (sel->destroyed) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
-		 "sel.c(sel_add_event_done): "
-		 "SEL info was destroyed while an operation was in progress");
+		 "%ssel.c(sel_add_event_done): "
+		 "SEL info was destroyed while an operation was in progress",
+		 sel->name);
 	sel_add_op_done(info, ECANCELED);
 	goto out;
     }
 
     if (!mc) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
-		 "sel.c(sel_add_event_done): "
-		 "MC went away while SEL op was in progress");
-        sel_add_op_done(info, ENXIO);
+		 "%ssel.c(sel_add_event_done): "
+		 "MC went away while SEL op was in progress", sel->name);
+        sel_add_op_done(info, ECANCELED);
 	goto out;
     }
 	
     if (rsp->data[0] != 0) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
-		 "sel.c(sel_add_event_done): "
+		 "%ssel.c(sel_add_event_done): "
 		 "IPMI error from SEL info fetch: %x",
-		 rsp->data[0]);
+		 sel->name, rsp->data[0]);
 	sel_add_op_done(info, IPMI_IPMI_ERR_VAL(rsp->data[0]));
 	goto out;
     }
 
     if (rsp->data_len < 3) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
-		 "sel.c(sel_add_event_done): "
-		 "SEL add response too short");
+		 "%ssel.c(sel_add_event_done): SEL add response too short",
+		 sel->name);
 	sel_add_op_done(info, EINVAL);
 	goto out;
     }
@@ -2008,8 +2266,9 @@ sel_add_event_op(void *cb_data, int shutdown)
     sel_lock(sel);
     if (shutdown) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
-		 "sel.c(sel_add_event_op): "
-		 "SEL info was destroyed while an operation was in progress");
+		 "%ssel.c(sel_add_event_op): "
+		 "SEL info was destroyed while an operation was in progress",
+		 sel->name);
 	sel_add_op_done(info, ECANCELED);
 	goto out;
     }
@@ -2017,14 +2276,14 @@ sel_add_event_op(void *cb_data, int shutdown)
     rv = ipmi_mc_pointer_cb(sel->mc, sel_add_event_cb, info);
     if (rv) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
-		 "sel.c(sel_add_event_op): "
-		 "MC went away during delete");
+		 "%ssel.c(sel_add_event_op): MC went away during delete",
+		 sel->name);
 	sel_add_op_done(info, ECANCELED);
 	goto out;
     } else if (info->rv) {
 	ipmi_log(IPMI_LOG_ERR_INFO,
-		 "sel.c(sel_add_event_cb): could not send cmd: %x",
-		 rv);
+		 "%ssel.c(sel_add_event_cb): could not send cmd: %x",
+		 sel->name, rv);
 	sel_add_op_done(info, info->rv);
 	goto out;
     }
@@ -2059,14 +2318,19 @@ ipmi_sel_add_event_to_sel(ipmi_sel_info_t           *sel,
 	goto out_unlock;
     }
 
+    sel_unlock(sel);
+
     /* Schedule this to run at the end of the queue. */
     if (!opq_new_op(sel->opq, sel_add_event_op, info, 0)) {
 	rv = ENOMEM;
 	goto out_unlock;
     }
+    goto out;
 
  out_unlock:
     sel_unlock(sel);
+
+ out:
     if (rv)
 	ipmi_mem_free(info);
     return rv;
