@@ -32,9 +32,25 @@
  */
 
 #include <stdlib.h>
+#include <stdio.h>
 #include <OpenIPMI/ipmi_bits.h>
 #include <OpenIPMI/ipmi_msgbits.h>
+#include <OpenIPMI/ipmi_auth.h>
 #include <OpenIPMI/ipmiif.h>
+#include <OpenIPMI/ipmi_err.h>
+#include <string.h>
+#include <OpenIPMI/internal/ipmi_malloc.h>
+
+const char *
+ipmi_update_e_string(enum ipmi_update_e val)
+{
+    switch (val) {
+    case IPMI_ADDED: return "added";
+    case IPMI_DELETED: return "deleted";
+    case IPMI_CHANGED: return "changed";
+    default: return "invalid";
+    }
+}
 
 static char *hysteresis_support_types[] =
 {
@@ -44,7 +60,7 @@ static char *hysteresis_support_types[] =
     "fixed",
 };
 #define NUM_HYSTERESIS_SUPPORT_TYPES (sizeof(hysteresis_support_types)/sizeof(char *))
-char *
+const char *
 ipmi_get_hysteresis_support_string(unsigned int val)
 {
     if (val >= NUM_HYSTERESIS_SUPPORT_TYPES)
@@ -60,7 +76,7 @@ static char *threshold_access_support_types[] =
     "fixed",
 };
 #define NUM_THRESHOLD_ACCESS_SUPPORT_TYPES (sizeof(threshold_access_support_types)/sizeof(char *))
-char *
+const char *
 ipmi_get_threshold_access_support_string(unsigned int val)
 {
     if (val >= NUM_THRESHOLD_ACCESS_SUPPORT_TYPES)
@@ -76,7 +92,7 @@ static char *event_support_types[] =
     "none",
 };
 #define NUM_EVENT_SUPPORT_TYPES (sizeof(event_support_types)/sizeof(char *))
-char *
+const char *
 ipmi_get_event_support_string(unsigned int val)
 {
     if (val >= NUM_EVENT_SUPPORT_TYPES)
@@ -130,7 +146,7 @@ static char *sensor_types[] =
     "battery",
 };
 #define NUM_SENSOR_TYPES (sizeof(sensor_types)/sizeof(char *))
-char *
+const char *
 ipmi_get_sensor_type_string(unsigned int val)
 {
     if (val >= NUM_SENSOR_TYPES)
@@ -155,7 +171,7 @@ static char *event_reading_types[] =
     "discrete_acpi_power",
 };
 #define NUM_EVENT_READING_TYPES (sizeof(event_reading_types)/sizeof(char *))
-char *
+const char *
 ipmi_get_event_reading_type_string(unsigned int val)
 {
     if (val == IPMI_EVENT_READING_TYPE_SENSOR_SPECIFIC)
@@ -166,6 +182,21 @@ ipmi_get_event_reading_type_string(unsigned int val)
     return event_reading_types[val];
 }
 
+static char *sensor_directions[] =
+{
+    "unspecified",
+    "input",
+    "output",
+};
+#define NUM_SENSOR_DIRECTIONS (sizeof(sensor_directions)/sizeof(char *))
+const char *
+ipmi_get_sensor_direction_string(unsigned int val)
+{
+    if (val >= NUM_SENSOR_DIRECTIONS)
+	return "invalid";
+    return sensor_directions[val];
+}
+
 static char *rate_unit_types[] =
 {
     "",
@@ -177,7 +208,7 @@ static char *rate_unit_types[] =
     "/day",
 };
 #define NUM_RATE_UNIT_TYPES (sizeof(rate_unit_types)/sizeof(char *))
-char *
+const char *
 ipmi_get_rate_unit_string(unsigned int val)
 {
     if (val >= NUM_RATE_UNIT_TYPES)
@@ -277,10 +308,12 @@ static char *unit_types[] =
     "characters",
     "errors",
     "correctable_errors",
-    "uncorrectable_errors"
+    "uncorrectable_errors",
+    "fatal errors",
+    "grams",
 };
 #define NUM_UNIT_TYPES (sizeof(unit_types)/sizeof(char *))
-char *
+const char *
 ipmi_get_unit_type_string(unsigned int val)
 {
     if (val >= NUM_UNIT_TYPES)
@@ -290,32 +323,32 @@ ipmi_get_unit_type_string(unsigned int val)
 
 static char *threshold_types[] =
 {
-    "lower non critical",
+    "lower non-critical",
     "lower critical",
-    "lower non recoverable",
-    "upper non critical",
+    "lower non-recoverable",
+    "upper non-critical",
     "upper critical",
-    "upper non recoverable"
+    "upper non-recoverable"
 };
 #define NUM_THRESHOLD_TYPES (sizeof(threshold_types)/sizeof(char *))
-char *
-ipmi_get_threshold_string(unsigned int val)
+const char *
+ipmi_get_threshold_string(enum ipmi_thresh_e val)
 {
-    if (val >= NUM_THRESHOLD_TYPES)
+    if (val > NUM_THRESHOLD_TYPES)
 	return "invalid";
     return threshold_types[val];
 }
 
 static char *value_dir_types[] =
 {
-    "going low",
-    "going high"
+    "going-low",
+    "going-high"
 };
 #define NUM_VALUE_DIR_TYPES (sizeof(value_dir_types)/sizeof(char *))
-char *
-ipmi_get_value_dir_string(unsigned int val)
+const char *
+ipmi_get_value_dir_string(enum ipmi_event_value_dir_e val)
 {
-    if (val >= NUM_VALUE_DIR_TYPES)
+    if (val > NUM_VALUE_DIR_TYPES)
 	return "invalid";
     return value_dir_types[val];
 }
@@ -326,17 +359,17 @@ static char *event_dir_types[] =
     "deassertion"
 };
 #define NUM_EVENT_DIR_TYPES (sizeof(event_dir_types)/sizeof(char *))
-char *
-ipmi_get_event_dir_string(unsigned int val)
+const char *
+ipmi_get_event_dir_string(enum ipmi_event_dir_e val)
 {
-    if (val >= NUM_EVENT_DIR_TYPES)
+    if (val > NUM_EVENT_DIR_TYPES)
 	return "invalid";
     return event_dir_types[val];
 }
 
 static char *entity_id_types[] =
 {
-    "unspecified",
+    "unspecified",					/* 0 */
     "other",
     "unkown",
     "processor",
@@ -346,7 +379,7 @@ static char *entity_id_types[] =
     "system_board",
     "memory_module",
     "processor_module",
-    "power_supply",
+    "power_supply",					/* 10 */
     "add_in_card",
     "front_panel_board",
     "back_panel_board",
@@ -356,7 +389,7 @@ static char *entity_id_types[] =
     "other_system_board",
     "processor_board",
     "power_unit",
-    "power_module",
+    "power_module",					/* 20 */
     "power_management_board",
     "chassis_back_panel_board",
     "system_chassis",
@@ -366,7 +399,7 @@ static char *entity_id_types[] =
     "peripheral_bay",
     "device_bay",
     "fan_cooling",
-    "cooling_unit",
+    "cooling_unit",					/* 30 */
     "cable_interconnect",
     "memory_device",
     "system_management_software",
@@ -376,16 +409,22 @@ static char *entity_id_types[] =
     "group",
     "remote_mgmt_comm_device",
     "external_environment",
-    "battery",
+    "battery",						/* 40 */
     "processing blade",
     "connectivity switch",
     "processor/memory module",
     "I/O module",
     "processor I/O module",
     "management controller firmware",
+    "ipmi channel",
+    "pci bus",
+    "pci express bus",
+    "scsi bus",						/* 50 */
+    "sata/sas bus",
+    "processor front side bus",
 };
 #define NUM_ENTITY_ID_TYPES (sizeof(entity_id_types)/sizeof(char *))
-char *
+const char *
 ipmi_get_entity_id_string(unsigned int val)
 {
     if (val >= NUM_ENTITY_ID_TYPES)
@@ -510,8 +549,22 @@ static char *sensor_states[256][16] =
 	"memory scrub failed (stuck bit)",
 	"memory device disabled",
 	"correctable ECC log limit reached",
+	"presence detected",
+	"configuration error",
+	"spare",
+	"memory automatically throttled",
+    },
+    { /* 0x0d */
+	"drive presence",
+	"drive fault",
+	"predictive failure",
+	"hot spare",
+	"consistancy/parity check in progress",
+	"in critical array",
+	"in failed array",
+	"rebuild/remap in progress",
+	"rebuild/remap aborted",
     },
-    {}, /* 0x0d */
     {}, /* 0x0e */
     { /* 0x0f */
 	"system firmware error",
@@ -564,7 +617,10 @@ static char *sensor_states[256][16] =
     {}, /* 0x18 */
     {}, /* 0x19 */
     {}, /* 0x1a */
-    {}, /* 0x1b */
+    { /* 0x1b */
+	"cable/interconnect is connected",
+	"configuration error",
+    },
     {}, /* 0x1c */
     { /* 0x1d */
 	"initiated by power up",
@@ -661,7 +717,7 @@ static char *sensor_states[256][16] =
     },
 };
 
-char *
+const char *
 ipmi_get_reading_name(unsigned int event_reading_type,
 		      unsigned int sensor_type,
 		      unsigned int val)
@@ -697,7 +753,7 @@ static char *control_types[] =
     "one-shot output",
 };
 #define NUM_CONTROL_TYPES (sizeof(control_types)/sizeof(char *))
-char *
+const char *
 ipmi_get_control_type_string(unsigned int val)
 {
     if (val >= NUM_CONTROL_TYPES)
@@ -716,7 +772,7 @@ static char *colors[] =
     "orange",
 };
 #define NUM_COLORS (sizeof(colors)/sizeof(char *))
-char *
+const char *
 ipmi_get_color_string(unsigned int val)
 {
     if (val >= NUM_COLORS)
@@ -753,14 +809,13 @@ static char *ipmi_netfns[] =
  * Convert a NetFN code into a string
  */
 char *
-ipmi_get_netfn_string(unsigned int  netfn)
+ipmi_get_netfn_string(unsigned int netfn,
+		      char         *buffer,
+		      unsigned int buf_len)
 {
-    int		len;
-    char*	netfn_fs;
-    static char netfn_buffer[32];
-    extern int sprintf(char *str, const char *format, ...);
+    char *netfn_fs;
 
-    netfn &= 0x3f;	/* low 6 bits only thanks! */
+    netfn &= 0x3f;
 
     if ( netfn > 0x2f )
 	netfn_fs = ipmi_netfns[(netfn & 0x01) + 20];
@@ -773,11 +828,9 @@ ipmi_get_netfn_string(unsigned int  netfn)
     else
 	netfn_fs = ipmi_netfns[netfn];
 
-    if ( (len = sprintf(netfn_buffer, netfn_fs, netfn)) > (sizeof(netfn_buffer) - 1))
-	len = sizeof(netfn_buffer) - 1;
-    netfn_buffer[len] = '\0';
+    snprintf(buffer, buf_len, netfn_fs, netfn);
 
-    return(netfn_buffer);
+    return buffer;
 }
 
 static char * ipmi_app_cmds[] =
@@ -820,8 +873,23 @@ static char * ipmi_app_cmds[] =
     "GetUsrAccCmd:%02x",	// +35: 0x44
     "SetUsrName:%02x",		// +36: 0x45
     "GetUsrName:%02x",		// +37: 0x46
-    "SetUsrPasswd:%02x",	// +38: 0x47 (0x48..0x51 unassigned)
-    "MstrWrRd:%02x",		// +39: 0x52
+    "SetUsrPasswd:%02x",	// +38: 0x47
+    "ActPayload:%02x",		// +39: 0x48
+    "DeactPayload:%02x",	// +40: 0x49
+    "GetPayloadActStat:%02x",	// +41: 0x4a
+    "GetPayloadInstInfo:%02x",	// +42: 0x4b
+    "SetUserPayloadAcc:%02x",	// +43: 0x4c
+    "GetUserPayloadAcc:%02x",	// +44: 0x4d
+    "GetChanPayloadSup:%02x",	// +45: 0x4e
+    "GetChanPayloadVer:%02x",	// +46: 0x4f
+    "GetChanOEMPayloadInf:%02x",// +47: 0x50
+    "unassigned:%02x",		// +48: 0x51 unassigned
+    "MstrWrRd:%02x",		// +49: 0x52
+    "unassigned:%02x",		// +50: 0x53 unassigned
+    "GetChanCipherSuites:%02x",	// +51: 0x54
+    "SuspResPayloadEnc:%02x",	// +52: 0x55
+    "SetChanSecKey:%02x",	// +53: 0x56
+    "GetSysIntfCap:%02x",	// +54: 0x57
 };
 
 static char * ipmi_chassis_cmds[] =
@@ -923,6 +991,13 @@ static char * ipmi_transport_cmds[] =
     "Callback:%02x",		// +14: 0x19
     "SetUsrCallOpts:%02x",	// +15: 0x1a
     "GetUsrCallOpts:%02x",	// +16: 0x1b
+    "unassigned:%02x",		// +17: 0x1c
+    "unassigned:%02x",		// +18: 0x1d
+    "unassigned:%02x",		// +19: 0x1e
+    "unassigned:%02x",		// +20: 0x1f
+    "SOLActivating:%02x",	// +21: 0x20
+    "SetSOLConfParm:%02x",	// +22: 0x21
+    "GetSOLConfParm:%02x",	// +23: 0x22
 };
 
 static char * ipmi_bridge_cmds[] =
@@ -956,23 +1031,22 @@ static char * ipmi_bridge_cmds[] =
     "OEMcommands:%02x",		// +26: 0xc0..0xfe
     "ErrorReport:%02x",		// +27: 0xff
 };
+
 /*
  * Convert a NetFN/Command code into a string
  */
 char *
-ipmi_get_command_string(unsigned int  netfn, unsigned int  cmd)
+ipmi_get_command_string(unsigned int netfn,
+			unsigned int cmd,
+			char         *buffer,
+			unsigned int buf_len)
 {
-    int		len;
-    char**	netfn_table;
-    char*	cmd_fs;
-    static char cmd_buffer[32];
-    extern int sprintf(char *str, const char *format, ...);
+    char **netfn_table;
+    char *cmd_fs = NULL;
+ 
+    netfn &= 0x3f;
 
-    netfn &= 0x3f;	/* low 6 bits only thanks! */
-
-    if ( netfn > 0x0d )
-	cmd_fs = "unknown:%02x";
-    else {
+    if ( netfn <= 0x0d ) {
 	switch ( netfn ) {
 	    case IPMI_CHASSIS_NETFN:
 	    case IPMI_CHASSIS_NETFN | 1:
@@ -982,8 +1056,6 @@ ipmi_get_command_string(unsigned int  netfn, unsigned int  cmd)
 		    cmd_fs = netfn_table[cmd];
 		else if ( cmd == 0x0f )
 		    cmd_fs = netfn_table[11];
-		else
-		    cmd_fs = netfn_table[10];
 		break;
 	    }
 	    case IPMI_BRIDGE_NETFN:
@@ -1000,8 +1072,6 @@ ipmi_get_command_string(unsigned int  netfn, unsigned int  cmd)
 		    cmd_fs = netfn_table[cmd - 0x30 + 20];
 		else if ( (cmd > 0xbf) && (cmd < 0xff) )
 		    cmd_fs = netfn_table[26];
-		else
-		    cmd_fs = netfn_table[27];
 		break;
 	    }
 	    case IPMI_SENSOR_EVENT_NETFN:
@@ -1014,8 +1084,6 @@ ipmi_get_command_string(unsigned int  netfn, unsigned int  cmd)
 		    cmd_fs = netfn_table[cmd - 0x10 + 3];
 		else if ( (cmd > 0x1f) && (cmd < 0x30) )
 		    cmd_fs = netfn_table[cmd - 0x20 + 11];
-		else
-		    cmd_fs = netfn_table[23];
 		break;
 	    }
 	    case IPMI_APP_NETFN:
@@ -1026,12 +1094,8 @@ ipmi_get_command_string(unsigned int  netfn, unsigned int  cmd)
 		    cmd_fs = netfn_table[cmd];
 		else if ( (cmd > 0x21) && (cmd < 0x26) )
 		    cmd_fs = netfn_table[cmd - 0x22 + 9];
-		else if ( (cmd > 0x2d) && (cmd < 0x48) )
+		else if ( (cmd > 0x2d) && (cmd < 0x58) )
 		    cmd_fs = netfn_table[cmd - 0x2e + 13];
-		else if ( cmd == 0x52 )
-		    cmd_fs = netfn_table[39];
-		else
-		    cmd_fs = netfn_table[10];
 		break;
 	    }
 	    case IPMI_FIRMWARE_NETFN:
@@ -1054,8 +1118,6 @@ ipmi_get_command_string(unsigned int  netfn, unsigned int  cmd)
 		    cmd_fs = netfn_table[cmd - 0x40 + 17];
 		else if ( (cmd > 0x59) && (cmd < 0x5c) )
 		    cmd_fs = netfn_table[cmd - 0x5a + 27];
-		else
-		    cmd_fs = netfn_table[0];
 		break;
 	    }
 	    case IPMI_TRANSPORT_NETFN:
@@ -1064,23 +1126,22 @@ ipmi_get_command_string(unsigned int  netfn, unsigned int  cmd)
 		netfn_table=ipmi_transport_cmds;
 		if ( cmd < 0x05 )
 		    cmd_fs = netfn_table[cmd];
-		else if ( (cmd > 0x0f) && (cmd < 0x1c) )
+		else if ( (cmd > 0x0f) && (cmd < 0x23) )
 		    cmd_fs = netfn_table[cmd - 0x10 + 5];
-		else
-		    cmd_fs = netfn_table[0];
 		break;
 	    }
 	}
     }
-    /* Format the command into a human-readable string */
-    if ( (len = sprintf(cmd_buffer, cmd_fs, cmd)) > (sizeof(cmd_buffer) - 1))
-	len = sizeof(cmd_buffer) - 1;
-    cmd_buffer[len] = '\0';
 
-    return(cmd_buffer);
+    if (!cmd_fs)
+	cmd_fs = "unknown:%02x";
+
+    snprintf(buffer, buf_len, cmd_fs, cmd);
+
+    return buffer;
 }
 
-static char * ipmi_ccodes[] =
+static char *ipmi_ccodes[] =
 {
     "Normal:%02x",	// +0 : 0x00 (0x01..0xbf undefined)
     "NodeBusy:%02x",	// +1 : 0xc0
@@ -1106,18 +1167,16 @@ static char * ipmi_ccodes[] =
     "PrivError:%02x",	// +21: 0xd4
     "InvalState:%02x",	// +22: 0xd5 (0xd6..0xfe undefined)
     "Unspecified:%02x",	// +23: 0xff
-    "Unknown:%02x",	// +24: 0x??
 };
 /*
  * Convert a Completion Code into a string
  */
 char *
-ipmi_get_cc_string(unsigned int  cc)
+ipmi_get_cc_string(unsigned int cc,
+		   char         *buffer,
+		   unsigned int buf_len)
 {
-    int		len;
-    char*	cc_fs;
-    static char cc_buffer[32];
-    extern int sprintf(char *str, const char *format, ...);
+    char *cc_fs;
 
     if ( cc == 0x00 )
 	cc_fs = ipmi_ccodes[0];
@@ -1126,17 +1185,106 @@ ipmi_get_cc_string(unsigned int  cc)
     else if ( cc == 0xff )
 	cc_fs = ipmi_ccodes[23];
     else
-	cc_fs = ipmi_ccodes[24];
+	cc_fs = "Unknown:%02x";
 
-    if ( (len = sprintf(cc_buffer, cc_fs, cc)) > (sizeof(cc_buffer) - 1))
-	len = sizeof(cc_buffer) - 1;
-    cc_buffer[len] = '\0';
+    snprintf(buffer, buf_len, cc_fs, cc);
 
-    return(cc_buffer);
+    return buffer;
+}
+
+
+static const char *sol_error_codes[] =
+{
+	"SoLCharTransUnavail",
+	"SoLDeactivated",
+	"SoLNotAvailable",
+	"SoLDisconnected",
+	"SoLUnconfirmOp",
+	"SoLFlushed",
+	"SoLUnknown"
+};
+
+static const char *rmcpp_error_codes[] =
+{
+	"InsufResourForSess",
+	"InvalSessID",
+	"InvalPayloadType",
+	"InvalAuthAlg",
+	"InvalIntegAlg",
+	"NoMatchAuthPayld",
+	"NoMatchIntegPayld",
+	"InactSessID",
+	"InvalidRole",
+	"UnauthRoleOrPriv",
+	"InsufResourForRole",
+	"InvalNameLength",
+	"UnauthName",
+	"UnauthGUID",
+	"InvalIntegChkVal",
+	"InvalConfidAlg",
+	"NoCipherSuiteMatch",
+	"IllegalParam",
+	"RMCPPUnknown"
+};
+
+char *
+ipmi_get_error_string(unsigned int err,
+			char *buffer,
+			unsigned int buf_len)
+{
+	char *temp_buffer;
+	char *err_type;
+
+	if (err == 0)
+	{
+		strncpy(buffer, "Success (No error)", buf_len);
+		return buffer;
+	}
+
+	temp_buffer = ipmi_mem_alloc(buf_len);
+	if (!temp_buffer)
+		temp_buffer = buffer;
+
+	if (IPMI_IS_OS_ERR(err))
+	{
+		strncpy(temp_buffer, strerror(IPMI_GET_OS_ERR(err)), buf_len);
+		err_type = "OS";
+	} else if (IPMI_IS_IPMI_ERR(err))
+	{
+		ipmi_get_cc_string(IPMI_GET_IPMI_ERR(err), temp_buffer, buf_len);
+		err_type = "IPMI";
+	} else if (IPMI_IS_RMCPP_ERR(err))
+	{
+		int rmcpp_err = IPMI_GET_RMCPP_ERR(err);
+		if ((rmcpp_err <= 0) && (rmcpp_err > 0x12))
+			rmcpp_err = 0x13;
+		snprintf(temp_buffer, buf_len, "%s (0x%02x)", rmcpp_error_codes[rmcpp_err - 1], IPMI_GET_RMCPP_ERR(err));
+		err_type = "RMCP+";
+	} else if (IPMI_IS_SOL_ERR(err))
+	{
+		int sol_err = IPMI_GET_SOL_ERR(err);
+		if ((sol_err < 1) || (sol_err > 7))
+			sol_err = 7;
+		snprintf(temp_buffer, buf_len, sol_error_codes[sol_err - 1]);
+		err_type = "SoL";
+	}
+	else
+	{
+		strncpy(temp_buffer, "Unknown", buf_len);
+		err_type = "Unknown";
+	}
+
+	if (temp_buffer != buffer)
+	{
+		snprintf(buffer, buf_len, "%s: %s", err_type, temp_buffer);
+		ipmi_mem_free(temp_buffer);
+	}
+
+	return buffer;
 }
 
 /* Get a string name fo the hot swap state. */
-char *
+const char *
 ipmi_hot_swap_state_name(enum ipmi_hot_swap_states state)
 {
     switch (state) {
@@ -1151,3 +1299,61 @@ ipmi_hot_swap_state_name(enum ipmi_hot_swap_states state)
     default: return "invalid_state";
     }
 }
+
+static char *domain_types[] =
+{
+    "unknown",
+    "MXP",
+    "ATCA",
+};
+#define NUM_DOMAIN_TYPES (sizeof(domain_types)/sizeof(char *))
+
+const char *
+ipmi_domain_get_type_string(enum ipmi_domain_type dtype)
+{
+    if ((dtype < 0) || (dtype >= NUM_DOMAIN_TYPES))
+	return "invalid";
+    return domain_types[dtype];
+}
+
+const char *
+ipmi_authtype_string(int authtype)
+{
+    switch(authtype) {
+    case IPMI_AUTHTYPE_DEFAULT:
+	return "default";
+    case IPMI_AUTHTYPE_NONE:
+	return "none";
+    case IPMI_AUTHTYPE_MD2:
+	return "md2";
+    case IPMI_AUTHTYPE_MD5:
+	return "md5";
+    case IPMI_AUTHTYPE_STRAIGHT:
+	return "straight";
+    case IPMI_AUTHTYPE_OEM:
+	return "oem";
+    case IPMI_AUTHTYPE_RMCP_PLUS:
+	return "rmcp+";
+    default:
+	return "invalid";
+    }
+}
+
+const char *
+ipmi_privilege_string(int privilege)
+{
+    switch(privilege) {
+    case IPMI_PRIVILEGE_CALLBACK:
+	return "callback";
+    case IPMI_PRIVILEGE_USER:
+	return "user";
+    case IPMI_PRIVILEGE_OPERATOR:
+	return "operator";
+    case IPMI_PRIVILEGE_ADMIN:
+	return "admin";
+    case IPMI_PRIVILEGE_OEM:
+	return "oem";
+    default:
+	return "invalid";
+    }
+}
