@@ -48,21 +48,22 @@
 #include <popt.h> /* Option parsing made easy */
 
 #include <OpenIPMI/ipmi_conn.h>
-#include <OpenIPMI/ipmi_event.h>
 #include <OpenIPMI/ipmi_lan.h>
 #include <OpenIPMI/ipmi_smi.h>
 #include <OpenIPMI/ipmi_auth.h>
 #include <OpenIPMI/ipmi_msgbits.h>
 #include <OpenIPMI/ipmi_posix.h>
 #include <OpenIPMI/mxp.h>
+#include <OpenIPMI/ipmi_posix.h>
 
-#include <OpenIPMI/ipmi_int.h>
+#include <OpenIPMI/internal/ipmi_event.h>
 
 void ipmi_oem_force_conn_init(void);
 int ipmi_oem_motorola_mxp_init(void);
 
 static char* sOp		= NULL;
 static int   interactive        = 1;
+static int   interactive_done   = 0;
 
 struct poptOption poptOpts[]=
 {
@@ -97,6 +98,7 @@ struct poptOption poptOpts[]=
 selector_t *sel;
 os_handler_t *os_hnd;
 static ipmi_con_t *con;
+static int continue_operation = 1;
 
 /* We cobbled everything in the next section together to provide the
    things that the low-level handlers need. */
@@ -120,153 +122,25 @@ ipmi_read_unlock()
 {
 }
 
-unsigned int __ipmi_log_mask = 0;
-
 void
-ipmi_log(enum ipmi_log_type_e log_type, char *format, ...)
-{
-    va_list ap;
-
-    va_start(ap, format);
-    vfprintf(stderr, format, ap);
-    va_end(ap);
-}
-
-void
-ui_vlog(char *format, enum ipmi_log_type_e log_type, va_list ap)
+posix_vlog(char *format, enum ipmi_log_type_e log_type, va_list ap)
 {
     vfprintf(stderr, format, ap);
-}
-
-void
-ipmi_report_lock_error(os_handler_t *handler, char *str)
-{
-}
-
-void
-ipmi_check_lock(ipmi_lock_t *lock, char *str)
-{
-}
-
-void
-ipmi_register_ll(ll_ipmi_t *ll)
-{
-}
-
-struct ipmi_lock_s
-{
-    os_hnd_lock_t *ll_lock;
-    os_handler_t  *os_hnd;
-};
-int
-ipmi_create_lock_os_hnd(os_handler_t *os_hnd, ipmi_lock_t **new_lock)
-{
-    ipmi_lock_t *lock;
-    int         rv;
-
-    lock = ipmi_mem_alloc(sizeof(*lock));
-    if (!lock)
-	return ENOMEM;
-
-    lock->os_hnd = os_hnd;
-    if (lock->os_hnd && lock->os_hnd->create_lock) {
-	rv = lock->os_hnd->create_lock(lock->os_hnd, &(lock->ll_lock));
-	if (rv) {
-	    ipmi_mem_free(lock);
-	    return rv;
-	}
-    } else {
-	lock->ll_lock = NULL;
-    }
-
-    *new_lock = lock;
-
-    return 0;
-}
-
-int
-ipmi_create_global_lock(ipmi_lock_t **new_lock)
-{
-    ipmi_lock_t *lock;
-    int         rv;
-
-    lock = ipmi_mem_alloc(sizeof(*lock));
-    if (!lock)
-	return ENOMEM;
-
-    lock->os_hnd = os_hnd;
-    if (lock->os_hnd && lock->os_hnd->create_lock) {
-	rv = lock->os_hnd->create_lock(lock->os_hnd, &(lock->ll_lock));
-	if (rv) {
-	    ipmi_mem_free(lock);
-	    return rv;
-	}
-    } else {
-	lock->ll_lock = NULL;
-    }
-
-    *new_lock = lock;
-
-    return 0;
-}
-
-void ipmi_lock(ipmi_lock_t *lock)
-{
-    if (lock->ll_lock)
-	lock->os_hnd->lock(lock->os_hnd, lock->ll_lock);
-}
-
-void ipmi_unlock(ipmi_lock_t *lock)
-{
-    if (lock->ll_lock)
-	lock->os_hnd->unlock(lock->os_hnd, lock->ll_lock);
-}
-
-void ipmi_destroy_lock(ipmi_lock_t *lock)
-{
-    if (lock->ll_lock)
-	lock->os_hnd->destroy_lock(lock->os_hnd, lock->ll_lock);
-    ipmi_mem_free(lock);
+    fprintf(stderr, "\n");
 }
 
 static void
-real_quit(selector_t *sel, sel_timer_t *timer, void *data)
-{
-    int rv = (int)(*(int *)data);
-    sel_free_timer(timer);
-    if (con && con->close_connection) 
-	con->close_connection(con);
-    if (sel)
-	sel_free_selector(sel);
-    exit(rv);
-}
-
-void
 leave(int ret)
 {
-    int               rv;
-    sel_timer_t *timer = NULL;
-    struct timeval    timeout;
-    static int ret_code;
-
-    ret_code = ret;
-    rv = sel_alloc_timer(sel, real_quit, &ret_code, &timer);
-    if (rv) {
-	/* Cannot allocate timer, exit anyway */
-        if (sel)
-	    sel_free_selector(sel);
-	exit(1);
+    if (con && con->close_connection) {
+	con->close_connection(con);
+	con = NULL;
     }
-
-    timeout.tv_sec = 0;
-    timeout.tv_usec = 0;
-    rv = sel_start_timer(timer, &timeout);
-    if (rv) {
-	sel_free_timer(timer);
-        if (sel)
-	    sel_free_selector(sel);
-	exit(1);
+    if (sel) {
+	sel_free_selector(sel);
+	sel = NULL;
     }
+    exit(ret);
 }
 
 void printInfo( )
@@ -346,7 +220,7 @@ dump_msg_data(ipmi_msg_t *msg, ipmi_addr_t *addr, char *type)
 		   smi_addr->lun,
 		   msg->cmd);
 	}
-	else
+	else if (ipmb_addr)
 	{
 	    printf("%2.2x %2.2x %2.2x %2.2x ",
 		   addr->channel,
@@ -354,6 +228,12 @@ dump_msg_data(ipmi_msg_t *msg, ipmi_addr_t *addr, char *type)
 		   ipmb_addr->lun,
 		   msg->cmd);
 	}
+	else if (lan_addr)
+	    printf("    lan addr = %x,%x,%x,%x\n",
+		   lan_addr->session_handle,
+		   lan_addr->remote_SWID,
+		   lan_addr->local_SWID,
+		   lan_addr->lun);
     }
 
     for (i=0; i<msg->data_len; i++) {
@@ -379,21 +259,13 @@ cmd_handler(ipmi_con_t   *ipmi,
     printf("Command sequence = 0x%lx\n", sequence);
 }
 
-void
-rsp_handler(ipmi_con_t   *ipmi,
-	    ipmi_addr_t  *addr,
-	    unsigned int addr_len,
-	    ipmi_msg_t   *rsp,
-	    void         *data1,
-	    void         *data2,
-	    void         *data3,
-	    void         *data4)
+int
+rsp_handler(ipmi_con_t *ipmi, ipmi_msgi_t *rspi)
 {
-    dump_msg_data(rsp, addr, "response");
-    if (!interactive) {
-    	ipmi->close_connection(ipmi);
-	leave(0);
-    }
+    dump_msg_data(&rspi->msg, &rspi->addr, "response");
+    if (!interactive)
+	continue_operation = 0;
+    return IPMI_MSG_ITEM_NOT_USED;
 }
 
 void
@@ -401,8 +273,7 @@ event_handler(ipmi_con_t   *ipmi,
 	      ipmi_addr_t  *addr,
 	      unsigned int addr_len,
 	      ipmi_event_t *event,
-	      void         *data1,
-	      void         *data2)
+	      void         *cb_data)
 {
     unsigned int  record_id = ipmi_event_get_record_id(event);
     unsigned int  type = ipmi_event_get_type(event);
@@ -429,17 +300,10 @@ typedef struct timed_data_s
     unsigned int   total_count;
 } timed_data_t;
 
-void
-timed_rsp_handler(ipmi_con_t   *con,
-		  ipmi_addr_t  *addr,
-		  unsigned int addr_len,
-		  ipmi_msg_t   *rsp,
-		  void         *data1,
-		  void         *data2,
-		  void         *data3,
-		  void         *data4)
+int
+timed_rsp_handler(ipmi_con_t *con, ipmi_msgi_t *rspi)
 {
-    timed_data_t *data = data1;
+    timed_data_t *data = rspi->data1;
 
     if (data->count == 0) {
 	unsigned long  diff;
@@ -452,6 +316,7 @@ timed_rsp_handler(ipmi_con_t   *con,
 	       ((float) diff) / ((float)(data->total_count)),
 	       diff);
 	free(data);
+	free(rspi);
     } else {
 	int rv;
 
@@ -459,14 +324,15 @@ timed_rsp_handler(ipmi_con_t   *con,
 			       &data->addr,
 			       data->addr_len,
 			       &data->msg,
-			       timed_rsp_handler,
-			       data, NULL, NULL, NULL);
+			       timed_rsp_handler, rspi);
 	data->count--;
 	if (rv) {
-	    printf("Error sending command: %x\n", rv);
+	    fprintf(stderr, "Error sending command: %x\n", rv);
 	    free(data);
-	}
+	} else
+    	    return IPMI_MSG_ITEM_USED;
     }
+    return IPMI_MSG_ITEM_USED;
 }
 
 void
@@ -478,16 +344,23 @@ time_msgs(ipmi_con_t    *con,
 {
     timed_data_t *data;
     int          rv;
+    ipmi_msgi_t  *rspi;
 
     data = malloc(sizeof(*data));
     if (!data) {
-	printf("No memory to perform command\n");
+	fprintf(stderr, "No memory to perform command\n");
+	return;
+    }
+
+    rspi = ipmi_alloc_msg_item();
+    if (!rspi) {
+	free(data);
+	fprintf(stderr, "No memory to perform command\n");
 	return;
     }
 
     data->con = con;
     gettimeofday(&data->start_time, NULL);
-    data->count = 0;
     memcpy(&data->msg, msg, sizeof(data->msg));
     memcpy(data->data, msg->data, msg->data_len);
     data->msg.data = data->data;
@@ -496,16 +369,17 @@ time_msgs(ipmi_con_t    *con,
     data->count = count;
     data->total_count = count;
 
+    rspi->data1 = data;
     rv = con->send_command(data->con,
 			   &data->addr,
 			   data->addr_len,
 			   &data->msg,
-			   timed_rsp_handler,
-			   data, NULL, NULL, NULL);
+			   timed_rsp_handler, rspi);
     data->count--;
     if (rv) {
-	printf("Error sending command: %x\n", rv);
+	fprintf(stderr, "Error sending command: %x\n", rv);
 	free(data);
+	ipmi_free_msg_item(rspi);
     }
 }
 
@@ -550,7 +424,7 @@ process_input_line(char *buf)
     }
 
     if (strcmp(v, "quit") == 0) {
-	leave(0);
+	continue_operation = 0;
 	return 0;
     }
 
@@ -558,13 +432,13 @@ process_input_line(char *buf)
 	unsigned char netfn, cmd;
 	v = strtok_r(NULL, " \t\r\n,.", &strtok_data);
 	if (!v) {
-	    printf("No netfn for regcmd\n");
+	    fprintf(stderr, "No netfn for regcmd\n");
 	    return -1;
 	}
 	netfn = strtoul(v, &endptr, 16);
 	v = strtok_r(NULL, " \t\r\n,.", &strtok_data);
 	if (!v) {
-	    printf("No cmd for regcmd\n");
+	    fprintf(stderr, "No cmd for regcmd\n");
 	    return -1;
 	}
 	cmd = strtoul(v, &endptr, 16);
@@ -582,20 +456,20 @@ process_input_line(char *buf)
 	unsigned char netfn, cmd;
 	v = strtok_r(NULL, " \t\r\n,.", &strtok_data);
 	if (!v) {
-	    printf("No netfn for regcmd\n");
+	    fprintf(stderr, "No netfn for regcmd\n");
 	    return -1;
 	}
 	netfn = strtoul(v, &endptr, 16);
 	v = strtok_r(NULL, " \t\r\n,.", &strtok_data);
 	if (!v) {
-	    printf("No cmd for regcmd\n");
+	    fprintf(stderr, "No cmd for regcmd\n");
 	    return -1;
 	}
 	cmd = strtoul(v, &endptr, 16);
 	
 	rv = con->deregister_for_command(con, netfn, cmd);
 	if (rv) {
-	    printf("Could not set to get receive command: %x", rv);
+	    fprintf(stderr, "Could not set to get receive command: %x", rv);
 	    return -1;
 	}
 	return 0;
@@ -604,7 +478,7 @@ process_input_line(char *buf)
     if (strcmp(v, "test_lat") == 0) {
 	v = strtok_r(NULL, " \t\r\n,.", &strtok_data);
 	if (!v) {
-	    printf("No count for test_lat\n");
+	    fprintf(stderr, "No count for test_lat\n");
 	    return -1;
 	}
 	time_count = strtoul(v, &endptr, 16);
@@ -622,26 +496,31 @@ process_input_line(char *buf)
 
     while (v != NULL) {
 	if (pos >= sizeof(outbuf)) {
-	    printf("Message too long");
+	    fprintf(stderr, "Message too long");
 	    return -1;
 	}
 
 	outbuf[pos] = strtoul(v, &endptr, 16);
 	if (*endptr != '\0') {
-	    printf("Value %d was invalid\n", pos+1);
+	    fprintf(stderr, "Value %d was invalid\n", pos+1);
 	    return -1;
 	}
 	pos++;
 	v = strtok_r(NULL, " \t\r\n,.", &strtok_data);
     }
 
+    if (pos <= 0) {
+	fprintf(stderr, "No channel specified\n");
+	return -1;
+    }
+
     start = 0;
     channel = outbuf[start]; start++;
 
     if (channel == IPMI_BMC_CHANNEL) {
 	struct ipmi_system_interface_addr *si = (void *) &addr;
 	if ((pos-start) < 1) {
-	    printf("No IPMB address specified\n");
+	    fprintf(stderr, "No LUN specified\n");
 	    return -1;
 	}
 	si->lun = outbuf[start]; start++;
@@ -653,7 +532,7 @@ process_input_line(char *buf)
 	struct ipmi_lan_addr *lan = (void *) &addr;
 
 	if ((pos-start) < 3) {
-	    printf("No LAN address specified\n");
+	    fprintf(stderr, "No LAN address specified\n");
 	    return -1;
 	}
 
@@ -669,7 +548,7 @@ process_input_line(char *buf)
 	struct ipmi_ipmb_addr *ipmb = (void *) &addr;
 
 	if ((pos-start) < 2) {
-	    printf("No IPMB address specified\n");
+	    fprintf(stderr, "No IPMB address specified\n");
 	    return -1;
 	}
 
@@ -688,7 +567,7 @@ process_input_line(char *buf)
 
     if (msg.netfn & 1) {
 	if ((pos-start) < 1) {
-	    printf("No sequence for response\n");
+	    fprintf(stderr, "No sequence for response\n");
 	    return -1;
 	}
 
@@ -696,7 +575,7 @@ process_input_line(char *buf)
     }
 
     if ((pos-start) < 1) {
-	printf("Message too short\n");
+	fprintf(stderr, "Message too short\n");
 	return -1;
     }
 
@@ -710,10 +589,10 @@ process_input_line(char *buf)
 	if (msg.netfn & 1)
 	    rv = con->send_response(con, &addr, addr_len, &msg, seq);
 	else
-	    rv = con->send_command(con, &addr, addr_len, &msg, rsp_handler,
-				   NULL, NULL, NULL, NULL);
+	    rv = con->send_command(con, &addr, addr_len, &msg,
+				   rsp_handler, NULL);
 	if (rv) {
-	    printf("Error sending command: %x\n", rv);
+	    fprintf(stderr, "Error sending command: %x\n", rv);
 	}
     }
 
@@ -738,7 +617,8 @@ user_input_ready(int fd, void *data)
 	if( interactive)
 	    printf("\n"); 
     	con->close_connection(con);
-	leave(0);
+	continue_operation = 0;
+	return;
     }
     
     for (i=0; count > 0; i++, count--) {
@@ -758,7 +638,7 @@ user_input_ready(int fd, void *data)
     }
 
     if (pos >= 255) {
-	printf("Input line too long\n");
+	fprintf(stderr, "Input line too long\n");
 	pos = 0;
 	if (interactive)
 	    printf("=> ");
@@ -780,7 +660,10 @@ con_changed_handler(ipmi_con_t   *ipmi,
 	    fprintf(stderr, "Unable to setup connection: %x\n", err);
 	    leave(1);
 	}
-	process_input_line(buf);
+	if (!interactive_done) {
+	    interactive_done = 1;
+	    process_input_line(buf);
+	}
     } else {
 	if (err)
 	    fprintf(stderr, "Connection failed to port %d: %x\n", port_num,
@@ -793,7 +676,7 @@ con_changed_handler(ipmi_con_t   *ipmi,
 }
 
 int
-main(int argc, const char *oargv[])
+main(int argc, char *argv[])
 {
     int         rv;
     int         pos;
@@ -801,10 +684,10 @@ main(int argc, const char *oargv[])
     char        *bufline = NULL;
     int         curr_arg;
     ipmi_args_t *args;
-    const char  **argv;
+    const char  **oargv = (const char **) argv;
 
 
-    poptContext poptCtx = poptGetContext("ipmicmd", argc, oargv,poptOpts,0);
+    poptContext poptCtx = poptGetContext("ipmicmd", argc, oargv, poptOpts, 0);
 
     while (( o = poptGetNextOpt(poptCtx)) >= 0)
     {   
@@ -828,7 +711,7 @@ main(int argc, const char *oargv[])
 	}
     }
 
-    argv = poptGetArgs(poptCtx);
+    argv = (char **) poptGetArgs(poptCtx);
 
     if (!argv) {
 	fprintf(stderr, "Not enough arguments\n");
@@ -843,22 +726,10 @@ main(int argc, const char *oargv[])
 	exit(1);
     }
 
-    /* Initialize the OEM handlers. */
-    rv = _ipmi_conn_init();
-    if (rv) {
-	fprintf(stderr, "Error initializing connections: 0x%x\n", rv);
-	exit(1);
-    }
-    ipmi_oem_force_conn_init();
-    ipmi_oem_motorola_mxp_init();
-
-
-    curr_arg = 0;
-
     /* OS handler allocated first. */
     os_hnd = ipmi_posix_get_os_handler();
     if (!os_hnd) {
-	printf("ipmi_smi_setup_con: Unable to allocate os handler\n");
+	fprintf(stderr, "ipmi_smi_setup_con: Unable to allocate os handler\n");
 	exit(1);
     }
 
@@ -868,6 +739,15 @@ main(int argc, const char *oargv[])
     /* The OS handler has to know about the selector. */
     ipmi_posix_os_handler_set_sel(os_hnd, sel);
 
+    /* Initialize the OEM handlers. */
+    rv = ipmi_init(os_hnd);
+    if (rv) {
+	fprintf(stderr, "Error initializing connections: 0x%x\n", rv);
+	exit(1);
+    }
+
+    curr_arg = 0;
+
     rv = ipmi_parse_args(&curr_arg, argc, argv, &args);
     if (rv) {
 	fprintf(stderr, "Error parsing command arguments, argument %d: %s\n",
@@ -882,7 +762,7 @@ main(int argc, const char *oargv[])
     }
 
     if (interactive) {
-	rv = con->register_for_events(con, event_handler, NULL, NULL, NULL);
+	rv = con->add_event_handler(con, event_handler, NULL);
 	if (rv) {
 	    fprintf(stderr, "Could not set to get events: %x\n", rv);
 	}
@@ -892,7 +772,7 @@ main(int argc, const char *oargv[])
 	sel_set_fd_read_handler(sel, 0, SEL_FD_HANDLER_ENABLED);
     }
 
-    con->set_con_change_handler(con, con_changed_handler, NULL);
+    con->add_con_change_handler(con, con_changed_handler, NULL);
 
     rv = con->start_con(con);
     if (rv) {
@@ -905,54 +785,13 @@ main(int argc, const char *oargv[])
 	printf("=> ");
     fflush(stdout);
 
-    sel_select_loop(sel, NULL, 0, NULL);
-
-    return rv;
-}
-
-void
-posix_vlog(char *format, enum ipmi_log_type_e log_type, va_list ap)
-{
-    int do_nl = 1;
-
-    switch(log_type)
-    {
-	case IPMI_LOG_INFO:
-	    printf("INFO: ");
-	    break;
-
-	case IPMI_LOG_WARNING:
-	    printf("WARN: ");
-	    break;
-
-	case IPMI_LOG_SEVERE:
-	    printf("SEVR: ");
-	    break;
-
-	case IPMI_LOG_FATAL:
-	    printf("FATL: ");
-	    break;
-
-	case IPMI_LOG_ERR_INFO:
-	    printf("EINF: ");
-	    break;
-
-	case IPMI_LOG_DEBUG_START:
-	    do_nl = 0;
-	    /* FALLTHROUGH */
-	case IPMI_LOG_DEBUG:
-	    printf("DEBG: ");
-	    break;
-
-	case IPMI_LOG_DEBUG_CONT:
-	    do_nl = 0;
-	    /* FALLTHROUGH */
-	case IPMI_LOG_DEBUG_END:
+    while (continue_operation) {
+	rv = os_hnd->perform_one_op(os_hnd, NULL);
+	if (rv)
 	    break;
     }
 
-    vprintf(format, ap);
+    leave(rv);
 
-    if (do_nl)
-	printf("\n");
+    return rv;
 }
