@@ -46,11 +46,6 @@
 #include <OpenIPMI/ipmi_auth.h>
 #include <OpenIPMI/ipmi_lan.h>
 #include <OpenIPMI/ipmi_posix.h>
-#include <OpenIPMI/ipmi_event.h>
-
-/* This file should not normally be included by the user, but is here
-   to demonstrate an internal domain function. */
-#include <OpenIPMI/ipmi_domain.h>
 
 /* This sample application demostrates a very simple method to use
    OpenIPMI. It just search all sensors in the system.  From this
@@ -67,7 +62,8 @@ usage(void)
 	   "  %s [options] smi <smi #>\n"
 	   "     Make a connection to a local system management interface.\n"
 	   "     smi # is generally 0.\n"
-	   "  %s [options] lan <host> <port> <authtype> <privilege> <username> <password>\n"
+	   "  %s [options] lan <host> <port> <authtype> <privilege>"
+	   " <username> <password>\n"
 	   "     Make a connection to a IPMI 1.5 LAN interface.\n"
 	   "     Host and port specify where to connect to (port is\n"
 	   "     generally 623).  authtype is none, md2, md5, or straight.\n"
@@ -389,7 +385,7 @@ entity_change(enum ipmi_update_e op,
 	    printf("Entity added: %d.%d\n", id, instance);
 	    /* Register callback so that when the status of a
 	       sensor changes, sensor_change is called */
-	    rv = ipmi_entity_set_sensor_update_handler(entity,
+	    rv = ipmi_entity_add_sensor_update_handler(entity,
 						       sensor_change,
 						       entity);
 	    if (rv) {
@@ -397,7 +393,7 @@ entity_change(enum ipmi_update_e op,
 		exit(1);
 	    }
 
-            rv = ipmi_entity_set_fru_update_handler(entity,
+            rv = ipmi_entity_add_fru_update_handler(entity,
                                                     fru_change,
                                                     NULL);
 	    if (rv) {
@@ -407,22 +403,6 @@ entity_change(enum ipmi_update_e op,
     }
 }
 
-static void reread_sels_done(ipmi_domain_t *domain, int err, void *cb_data)
-{
-    int count;
-    ipmi_domain_sel_entries_used(domain, &count);
-    printf("reread sel done. entries = %d\n", count);
-}
-
-static void bus_scan_done(ipmi_domain_t *domain, int err, void *cb_data)
-{
-    int count;
-    ipmi_domain_reread_sels(domain, reread_sels_done, &count);
-    printf("bus scan done.\n");
-    return;
-}
-
-
 /* After we have established connection to domain, this function get called
    At this time, we can do whatever things we want to do. Herr we want to
    search all entities in the system */ 
@@ -438,25 +418,18 @@ setup_done(ipmi_domain_t *domain,
 
     /* Register a callback functin entity_change. When a new entities 
        is created, entity_change is called */
-    rv = ipmi_domain_set_entity_update_handler(domain, entity_change, domain);
+    rv = ipmi_domain_add_entity_update_handler(domain, entity_change, domain);
     if (rv) {      
-	printf("ipmi_domain_set_entity_update_handler return error: %d\n", rv);
+	printf("ipmi_domain_add_entity_update_handler return error: %d\n", rv);
 	return;
     }
 
-    rv = ipmi_domain_set_bus_scan_handler(domain, bus_scan_done, NULL);
-    if (rv) {
-	printf("ipmi_domain_set_bus_scan_handler return error: %d\n", rv);
-	return;
-    }
-    
 }
 
 static os_handler_t *os_hnd;
-static selector_t *sel;
 
 int
-main(int argc, const char *argv[])
+main(int argc, char *argv[])
 {
     int         rv;
     int         curr_arg = 1;
@@ -466,18 +439,12 @@ main(int argc, const char *argv[])
     progname = argv[0];
 
     /* OS handler allocated first. */
-    os_hnd = ipmi_posix_get_os_handler();
+    os_hnd = ipmi_posix_setup_os_handler();
     if (!os_hnd) {
 	printf("ipmi_smi_setup_con: Unable to allocate os handler\n");
 	exit(1);
     }
 
-    /* Create selector with os handler. */
-    sel_alloc_selector(os_hnd, &sel);
-
-    /* The OS handler has to know about the selector. */
-    ipmi_posix_os_handler_set_sel(os_hnd, sel);
-
     /* Initialize the OpenIPMI library. */
     ipmi_init(os_hnd);
 
@@ -490,7 +457,7 @@ main(int argc, const char *argv[])
        descriptor in selector is changed and predefined callback is
        called. After the connection is established, setup_done will be
        called. */
-    rv = ipmi_smi_setup_con(0, os_hnd, sel, &con);
+    rv = ipmi_smi_setup_con(0, os_hnd, NULL, &con);
     if (rv) {
 	printf("ipmi_smi_setup_con: %s", strerror(rv));
 	exit(1);
@@ -506,14 +473,15 @@ main(int argc, const char *argv[])
 	exit(1);
     }
 
-    rv = ipmi_args_setup_con(args, os_hnd, sel, &con);
+    rv = ipmi_args_setup_con(args, os_hnd, NULL, &con);
     if (rv) {
         fprintf(stderr, "ipmi_ip_setup_con: %s", strerror(rv));
 	exit(1);
     }
 #endif
 
-    rv = ipmi_init_domain(&con, 1, setup_done, NULL, NULL, NULL);
+    rv = ipmi_open_domain("", &con, 1, setup_done, NULL, NULL, NULL,
+			  NULL, 0, NULL);
     if (rv) {
 	fprintf(stderr, "ipmi_init_domain: %s\n", strerror(rv));
 	exit(1);
@@ -525,12 +493,15 @@ main(int argc, const char *argv[])
     /* We run the select loop here, this shows how you can use
        sel_select.  You could add your own processing in this loop. */
     while (1) {
-	sel_select(sel, NULL, 0, NULL, NULL);
+	os_hnd->perform_one_op(os_hnd, NULL);
     }
 #else
     /* Let the selector code run the select loop. */
-    sel_select_loop(sel, NULL, 0, NULL);
+    os_hnd->operation_loop(os_hnd);
 #endif
+
+    /* Technically, we can't get here, but this is an example. */
+    os_hnd->free_os_handler(os_hnd);
 }
 
 void
