@@ -37,18 +37,37 @@
 #include <stdlib.h>
 #include <errno.h>
 #include <stdio.h>
-#include <OpenIPMI/os_handler.h>
-#include <OpenIPMI/selector.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <fcntl.h>
 #include <unistd.h>
 #include <pthread.h>
 #include <string.h>
+#include <signal.h>
 
-#include <OpenIPMI/ipmi_int.h>
+#ifdef HAVE_GDBM
+#include <gdbm.h>
+#endif
+
+#include <OpenIPMI/os_handler.h>
+#include <OpenIPMI/selector.h>
 #include <OpenIPMI/ipmi_posix.h>
 
+#include <OpenIPMI/internal/ipmi_int.h>
+
+typedef struct pt_os_hnd_data_s
+{
+    selector_t       *sel;
+    int              wake_sig;
+    struct sigaction oldact;
+#ifdef HAVE_GDBM
+    char *gdbm_filename;
+    GDBM_FILE gdbmf;
+    pthread_mutex_t gdbm_lock;
+#endif
+} pt_os_hnd_data_t;
+
+
 struct os_hnd_fd_id_s
 {
     int             fd;
@@ -58,20 +77,16 @@ struct os_hnd_fd_id_s
     os_fd_data_freed_t freed;
 };
 
-static pthread_mutex_t fd_lock = PTHREAD_MUTEX_INITIALIZER;
-
 static void
 fd_handler(int fd, void *data)
 {
     os_hnd_fd_id_t *fd_data = (os_hnd_fd_id_t *) data;
     void            *cb_data;
-    os_handler_t    *handler;
+    os_data_ready_t handler;
 
-    pthread_mutex_lock(&fd_lock);
-    handler = fd_data->handler;
+    handler = fd_data->data_ready;
     cb_data = fd_data->cb_data;
-    pthread_mutex_unlock(&fd_lock);
-    fd_data->data_ready(fd, fd_data->cb_data, fd_data);
+    handler(fd, cb_data, fd_data);
 }
 
 static void
@@ -81,7 +96,7 @@ free_fd_data(int fd, void *data)
 
     if (fd_data->freed)
         fd_data->freed(fd, fd_data->cb_data);
-    ipmi_mem_free(data);
+    free(data);
 }
 
 static int
@@ -92,11 +107,12 @@ add_fd(os_handler_t       *handler,
        os_fd_data_freed_t freed,
        os_hnd_fd_id_t     **id)
 {
-    os_hnd_fd_id_t *fd_data;
-    int            rv;
-    selector_t     *posix_sel = handler->internal_data;
+    os_hnd_fd_id_t   *fd_data;
+    int              rv;
+    pt_os_hnd_data_t *info = handler->internal_data;
+    selector_t       *posix_sel = info->sel;
 
-    fd_data = ipmi_mem_alloc(sizeof(*fd_data));
+    fd_data = malloc(sizeof(*fd_data));
     if (!fd_data)
 	return ENOMEM;
 
@@ -110,7 +126,7 @@ add_fd(os_handler_t       *handler,
     rv = sel_set_fd_handlers(posix_sel, fd, fd_data, fd_handler, NULL, NULL,
 			     free_fd_data);
     if (rv) {
-	ipmi_mem_free(fd_data);
+	free(fd_data);
 	return rv;
     }
     sel_set_fd_read_handler(posix_sel, fd, SEL_FD_HANDLER_ENABLED);
@@ -122,7 +138,8 @@ add_fd(os_handler_t       *handler,
 static int
 remove_fd(os_handler_t *handler, os_hnd_fd_id_t *fd_data)
 {
-    selector_t *posix_sel = handler->internal_data;
+    pt_os_hnd_data_t *info = handler->internal_data;
+    selector_t       *posix_sel = info->sel;
 
     sel_set_fd_read_handler(posix_sel, fd_data->fd, SEL_FD_HANDLER_DISABLED);
     sel_clear_fd_handlers(posix_sel, fd_data->fd);
@@ -196,9 +213,10 @@ alloc_timer(os_handler_t      *handler,
 {
     os_hnd_timer_id_t *timer_data;
     int               rv;
-    selector_t        *posix_sel = handler->internal_data;
+    pt_os_hnd_data_t  *info = handler->internal_data;
+    selector_t        *posix_sel = info->sel;
 
-    timer_data = ipmi_mem_alloc(sizeof(*timer_data));
+    timer_data = malloc(sizeof(*timer_data));
     if (!timer_data)
 	return ENOMEM;
 
@@ -209,7 +227,7 @@ alloc_timer(os_handler_t      *handler,
     rv = sel_alloc_timer(posix_sel, timer_handler, timer_data,
 			 &(timer_data->timer));
     if (rv) {
-	ipmi_mem_free(timer_data);
+	free(timer_data);
 	return rv;
     }
 
@@ -221,7 +239,7 @@ static int
 free_timer(os_handler_t *handler, os_hnd_timer_id_t *timer_data)
 {
     sel_free_timer(timer_data->timer);
-    ipmi_mem_free(timer_data);
+    free(timer_data);
     return 0;
 }
 
@@ -289,12 +307,14 @@ create_lock(os_handler_t  *handler,
     os_hnd_lock_t *lock;
     int           rv;
 
-    lock = ipmi_mem_alloc(sizeof(*lock));
+    lock = malloc(sizeof(*lock));
     if (!lock)
 	return ENOMEM;
     rv = pthread_mutex_init(&lock->mutex, NULL);
-    if (rv)
+    if (rv) {
+	free(lock);
 	return rv;
+    }
     lock->lock_count = 0;
     *id = lock;
     return 0;
@@ -311,7 +331,7 @@ destroy_lock(os_handler_t  *handler,
     rv = pthread_mutex_destroy(&id->mutex);
     if (rv)
 	return rv;
-    ipmi_mem_free(id);
+    free(id);
     return 0;
 }
 
@@ -352,169 +372,6 @@ unlock(os_handler_t  *handler,
     return 0;
 }
 
-static int
-is_locked(os_handler_t  *handler,
-	  os_hnd_lock_t *id)
-{
-    return id->lock_count != 0;
-}
-
-struct os_hnd_rwlock_s
-{
-    pthread_rwlock_t rwlock;
-    pthread_mutex_t  read_lock_lock;
-    int              read_lock_count;
-    int              write_lock_count;
-
-    /* This is volatile and we always set the owner before we set the count.
-       That avoids race conditions checking the count and owner. */
-    volatile pthread_t write_owner;
-};
-
-static int
-create_rwlock(os_handler_t    *handler,
-	      os_hnd_rwlock_t **id)
-{
-    os_hnd_rwlock_t *lock;
-    int             rv;
-
-    lock = ipmi_mem_alloc(sizeof(*lock));
-    if (!lock)
-	return ENOMEM;
-    rv = pthread_rwlock_init(&lock->rwlock, NULL);
-    if (rv) {
-	ipmi_mem_free(lock);
-	return rv;
-    }
-    rv = pthread_mutex_init(&lock->read_lock_lock, NULL);
-    if (rv) {
-	pthread_rwlock_destroy(&lock->rwlock);
-	ipmi_mem_free(lock);
-	return rv;
-    }
-    lock->read_lock_count = 0;
-    lock->write_lock_count = 0;
-    *id = lock;
-    return 0;
-}
-
-static int
-destroy_rwlock(os_handler_t    *handler,
-	       os_hnd_rwlock_t *id)
-{
-    int rv;
-
-    if ((id->read_lock_count != 0) || (id->write_lock_count != 0))
-	ipmi_log(IPMI_LOG_FATAL, "Release of rwlock when count is not zero");
-    rv = pthread_rwlock_destroy(&id->rwlock);
-    if (rv)
-	return rv;
-    pthread_mutex_destroy(&id->read_lock_lock);
-    ipmi_mem_free(id);
-    return 0;
-}
-
-static int
-read_lock(os_handler_t    *handler,
-	  os_hnd_rwlock_t *id)
-{
-    int rv;
-
-    if ((id->write_lock_count > 0) && (id->write_owner == pthread_self())) {
-	id->read_lock_count++;
-	return 0;
-    }
-
-    rv = pthread_rwlock_rdlock(&id->rwlock);
-    if (rv)
-	return rv;
-    pthread_mutex_lock(&id->read_lock_lock);
-    id->read_lock_count++;
-    pthread_mutex_unlock(&id->read_lock_lock);
-    return 0;
-}
-
-static int
-read_unlock(os_handler_t    *handler,
-	    os_hnd_rwlock_t *id)
-{
-    int rv;
-
-    if ((id->write_lock_count > 0) && (id->write_owner == pthread_self())) {
-	if (id->read_lock_count == 0)
-	    ipmi_log(IPMI_LOG_FATAL, "read lock count went negative");
-	id->read_lock_count--;
-	return 0;
-    }
-
-    pthread_mutex_lock(&id->read_lock_lock);
-    if (id->read_lock_count == 0)
-	ipmi_log(IPMI_LOG_FATAL, "read lock count went negative");
-    id->read_lock_count--;
-    pthread_mutex_unlock(&id->read_lock_lock);
-    rv = pthread_rwlock_unlock(&id->rwlock);
-    if (rv) {
-	pthread_mutex_lock(&id->read_lock_lock);
-	id->read_lock_count++;
-	pthread_mutex_unlock(&id->read_lock_lock);
-	return rv;
-    }
-    return 0;
-}
-
-static int
-write_lock(os_handler_t    *handler,
-	   os_hnd_rwlock_t *id)
-{
-    int rv;
-
-    if ((id->write_lock_count == 0) || (id->write_owner != pthread_self())) {
-	rv = pthread_rwlock_wrlock(&id->rwlock);
-	if (rv)
-	    return rv;
-    }
-    id->write_owner = pthread_self();
-    id->write_lock_count++;
-    return 0;
-}
-
-static int
-write_unlock(os_handler_t    *handler,
-	     os_hnd_rwlock_t *id)
-{
-    int rv;
-
-    if (id->write_lock_count == 0)
-	ipmi_log(IPMI_LOG_FATAL, "write lock count went negative");
-    if (pthread_self() != id->write_owner)
-	ipmi_log(IPMI_LOG_FATAL, "lock release by non-owner");
-    id->write_lock_count--;
-    if (id->write_lock_count == 0) {
-	if (id->read_lock_count != 0)
-	    ipmi_log(IPMI_LOG_FATAL,"read lock count not zero on write unlock");
-	rv = pthread_rwlock_unlock(&id->rwlock);
-	if (rv) {
-	    id->write_lock_count++;
-	    return rv;
-	}
-    }
-    return 0;
-}
-
-static int
-is_readlocked(os_handler_t    *handler,
-	      os_hnd_rwlock_t *id)
-{
-    return ((id->write_lock_count != 0) || (id->read_lock_count != 0));
-}
-
-static int
-is_writelocked(os_handler_t    *handler,
-	       os_hnd_rwlock_t *id)
-{
-    return (id->write_lock_count != 0);
-}
-
 struct os_hnd_cond_s
 {
     pthread_cond_t cond;
@@ -525,12 +382,21 @@ create_cond(os_handler_t  *handler,
 	    os_hnd_cond_t **new_cond)
 {
     os_hnd_cond_t *cond;
+    int           rv;
 
-    cond = ipmi_mem_alloc(sizeof(*cond));
+    cond = malloc(sizeof(*cond));
     if (!cond)
 	return ENOMEM;
 
-    return pthread_cond_init(&cond->cond, NULL);
+    rv = pthread_cond_init(&cond->cond, NULL);
+
+    if (rv) {
+	free(cond);
+	return rv;
+    }
+
+    *new_cond = cond;
+    return 0;
 }
 
 static int
@@ -542,7 +408,7 @@ destroy_cond(os_handler_t  *handler,
     rv = pthread_cond_destroy(&cond->cond);
     if (rv)
 	return rv;
-    ipmi_mem_free(cond);
+    free(cond);
     return 0;
 }
 
@@ -596,6 +462,7 @@ create_thread(os_handler_t       *handler,
     pthread_attr_t     attr, *pattr = NULL;
     struct sched_param param;
     int                rv;
+    pthread_t          tid;
 
     if (priority) {
 	rv = pthread_attr_init(&attr);
@@ -611,7 +478,7 @@ create_thread(os_handler_t       *handler,
 	pattr = &attr;
     }
 
-    rv = pthread_create(NULL, pattr, (void *(*)(void *)) startup, data);
+    rv = pthread_create(&tid, pattr, (void *(*)(void *)) startup, data);
 
  out:
     if (pattr)
@@ -626,8 +493,216 @@ thread_exit(os_handler_t *handler)
 }
 
 
+void
+ipmi_posix_thread_free_os_handler(os_handler_t *os_hnd)
+{
+    pt_os_hnd_data_t *info = os_hnd->internal_data;
+
+#ifdef HAVE_GDBM
+    pthread_mutex_destroy(&info->gdbm_lock);
+    if (info->gdbm_filename)
+	free(info->gdbm_filename);
+    if (info->gdbmf)
+	gdbm_close(info->gdbmf);
+#endif
+    free(info);
+    free(os_hnd);
+}
+
+void
+ipmi_posix_thread_os_handler_set_sel(os_handler_t *os_hnd, selector_t *sel)
+{
+    pt_os_hnd_data_t  *info = os_hnd->internal_data;
+
+    info->sel = sel;
+}
+
+selector_t *
+ipmi_posix_thread_os_handler_get_sel(os_handler_t *os_hnd)
+{
+    pt_os_hnd_data_t  *info = os_hnd->internal_data;
+
+    return info->sel;
+}
+
+static void
+posix_thread_sighandler(int sig)
+{
+    /* Nothing to do, sending the sig just wakes up select(). */
+}
+
+static void
+posix_thread_send_sig(long thread_id, void *cb_data)
+{
+    pthread_t        *id = (void *) thread_id;
+    pt_os_hnd_data_t *info = cb_data;
+
+    pthread_kill(*id, info->wake_sig);
+}
+
+static int
+perform_one_op(os_handler_t   *os_hnd,
+	       struct timeval *timeout)
+{
+    pthread_t        self = pthread_self();
+    pt_os_hnd_data_t *info = os_hnd->internal_data;
+    int              rv;
+
+    rv = sel_select(info->sel, posix_thread_send_sig, (long) &self, info,
+		    timeout);
+    if (rv == -1)
+	return errno;
+    return 0;
+}
+
+static void
+operation_loop(os_handler_t *os_hnd)
+{
+    pthread_t        self = pthread_self();
+    pt_os_hnd_data_t *info = os_hnd->internal_data;
+
+    sel_select_loop(info->sel, posix_thread_send_sig, (long) &self, info);
+}
+
+static void
+free_os_handler(os_handler_t *os_hnd)
+{
+    pt_os_hnd_data_t *info = os_hnd->internal_data;
+
+    sigaction(info->wake_sig, &info->oldact, NULL);
+    sel_free_selector(info->sel);
+    ipmi_posix_thread_free_os_handler(os_hnd);
+}
+
+static void *
+posix_malloc(int size)
+{
+    return malloc(size);
+}
+
+static void
+posix_free(void *data)
+{
+    free(data);
+}
+
+#ifdef HAVE_GDBM
+#define GDBM_FILE ".OpenIPMI_db"
+
+static void
+init_gdbm(pt_os_hnd_data_t *info)
+{
+    if (!info->gdbm_filename) {
+	char *home = getenv("HOME");
+	if (!home)
+	    return;
+	info->gdbm_filename = malloc(strlen(home)+strlen(GDBM_FILE)+2);
+	if (!info->gdbm_filename)
+	    return;
+	strcpy(info->gdbm_filename, home);
+	strcat(info->gdbm_filename, "/");
+	strcat(info->gdbm_filename, GDBM_FILE);
+    }
+
+    info->gdbmf = gdbm_open(info->gdbm_filename, 512, GDBM_WRCREAT, 0600,
+			    NULL);
+    /* gdbmf will be NULL on error, which is what reports an error. */
+}
+
+static int
+database_store(os_handler_t  *handler,
+	       char          *key,
+	       unsigned char *data,
+	       unsigned int  data_len)
+{
+    pt_os_hnd_data_t *info = handler->internal_data;
+    datum            gkey, gdata;
+    int              rv;
+
+    pthread_mutex_lock(&info->gdbm_lock);
+    if (!info->gdbmf) {
+	init_gdbm(info);
+	if (!info->gdbmf) {
+	    pthread_mutex_unlock(&info->gdbm_lock);
+	    return EINVAL;
+	}
+    }
+
+    gkey.dptr = key;
+    gkey.dsize = strlen(key);
+    gdata.dptr = data;
+    gdata.dsize = data_len;
+
+    rv = gdbm_store(info->gdbmf, gkey, gdata, GDBM_REPLACE);
+    pthread_mutex_unlock(&info->gdbm_lock);
+    if (rv)
+	return EINVAL;
+    return 0;
+}
+
+static int
+database_find(os_handler_t  *handler,
+	      char          *key,
+	      unsigned int  *fetch_completed,
+	      unsigned char **data,
+	      unsigned int  *data_len,
+	      void (*got_data)(void          *cb_data,
+			       int           err,
+			       unsigned char *data,
+			       unsigned int  data_len),
+	      void *cb_data)
+{
+    pt_os_hnd_data_t *info = handler->internal_data;
+    datum            gkey, gdata;
+
+    pthread_mutex_lock(&info->gdbm_lock);
+    if (!info->gdbmf) {
+	init_gdbm(info);
+	if (!info->gdbmf) {
+	    pthread_mutex_unlock(&info->gdbm_lock);
+	    return EINVAL;
+	}
+    }
+
+    gkey.dptr = key;
+    gkey.dsize = strlen(key);
+    gdata = gdbm_fetch(info->gdbmf, gkey);
+    pthread_mutex_unlock(&info->gdbm_lock);
+    if (!gdata.dptr)
+	return EINVAL;
+    *data = gdata.dptr;
+    *data_len = gdata.dsize;
+    *fetch_completed = 1;
+    return 0;
+}
+
+static void
+database_free(os_handler_t  *handler,
+	      unsigned char *data)
+{
+    free(data);
+}
+
+static int
+set_gdbm_filename(os_handler_t *os_hnd, char *name)
+{
+    pt_os_hnd_data_t *info = os_hnd->internal_data;
+    char             *nname;
+
+    nname = strdup(name);
+    if (!nname)
+	return ENOMEM;
+    if (info->gdbm_filename)
+	free(info->gdbm_filename);
+    info->gdbm_filename = name;
+    return 0;
+}
+#endif
+
 static os_handler_t ipmi_posix_thread_os_handler =
 {
+    .mem_alloc = posix_malloc,
+    .mem_free = posix_free,
     .add_fd_to_wait_for = add_fd,
     .remove_fd_to_wait_for = remove_fd,
     .start_timer = start_timer,
@@ -636,17 +711,8 @@ static os_handler_t ipmi_posix_thread_os_handler =
     .free_timer = free_timer,
     .create_lock = create_lock,
     .destroy_lock = destroy_lock,
-    .is_locked = is_locked,
     .lock = lock,
     .unlock = unlock,
-    .create_rwlock = create_rwlock,
-    .destroy_rwlock = destroy_rwlock,
-    .read_lock = read_lock,
-    .write_lock = write_lock,
-    .read_unlock = read_unlock,
-    .write_unlock = write_unlock,
-    .is_readlocked = is_readlocked,
-    .is_writelocked = is_writelocked,
     .get_random = get_random,
     .log = sposix_log,
     .vlog = sposix_vlog,
@@ -658,23 +724,143 @@ static os_handler_t ipmi_posix_thread_os_handler =
     .cond_broadcast = cond_broadcast,
     .create_thread = create_thread,
     .thread_exit = thread_exit,
+    .free_os_handler = free_os_handler,
+    .perform_one_op = perform_one_op,
+    .operation_loop = operation_loop,
+#ifdef HAVE_GDBM
+    .database_store = database_store,
+    .database_find = database_find,
+    .database_free = database_free,
+    .database_set_filename = set_gdbm_filename,
+#endif
 };
 
 os_handler_t *
 ipmi_posix_thread_get_os_handler(void)
 {
-    os_handler_t *rv;
+    os_handler_t     *rv;
+    pt_os_hnd_data_t *info;
+#if HAVE_GDBM
+    int              err;
+#endif
 
-    rv = ipmi_mem_alloc(sizeof(*rv));
+    rv = malloc(sizeof(*rv));
     if (!rv)
 	return NULL;
 
     memcpy(rv, &ipmi_posix_thread_os_handler, sizeof(*rv));
+
+    info = malloc(sizeof(pt_os_hnd_data_t));
+    if (!info) {
+	free(rv);
+	return NULL;
+    }
+    memset(info, 0, sizeof(*info));
+    rv->internal_data = info;
+
+#ifdef HAVE_GDBM
+    err = pthread_mutex_init(&info->gdbm_lock, NULL);
+    if (err) {
+	free(info);
+	free(rv);
+	return NULL;
+    }
+#endif
+
     return rv;
 }
 
-void
-ipmi_posix_thread_os_handler_set_sel(os_handler_t *os_hnd, selector_t *sel)
+os_handler_t *
+ipmi_posix_thread_setup_os_handler(int wake_sig)
 {
-    os_hnd->internal_data = sel;
+    os_handler_t     *os_hnd;
+    pt_os_hnd_data_t *info;
+    struct sigaction act;
+    int              rv;
+
+    os_hnd = ipmi_posix_thread_get_os_handler();
+    if (!os_hnd)
+	return NULL;
+
+    info = os_hnd->internal_data;
+    info->wake_sig = wake_sig;
+
+    rv = sel_alloc_selector(os_hnd, &info->sel);
+    if (rv) {
+	ipmi_posix_thread_free_os_handler(os_hnd);
+	os_hnd = NULL;
+	goto out;
+    }
+
+    act.sa_handler = posix_thread_sighandler;
+    sigemptyset(&act.sa_mask);
+    act.sa_flags = 0;
+    rv = sigaction(wake_sig, &act, &info->oldact);
+    if (rv) {
+	ipmi_posix_thread_free_os_handler(os_hnd);
+	os_hnd = NULL;
+	goto out;
+    }
+
+ out:
+    return os_hnd;
+}
+
+/*
+ * Support for the selector code.
+ */
+
+int
+posix_mutex_alloc(void **val)
+{
+    void *m = malloc(sizeof(pthread_mutex_t));
+    int  rv;
+
+    if (!m)
+	return ENOMEM;
+
+    rv = pthread_mutex_init(m, NULL);         
+    if (rv) {
+	free(m);
+	return rv;
+    }
+    *val = m;
+    return 0;
+}
+
+void posix_mutex_free(void *val)
+{
+    free(val);
+}
+
+void posix_mutex_lock(void *val)
+{
+    pthread_mutex_lock(val);
+}
+
+void posix_mutex_unlock(void *val)
+{
+    pthread_mutex_unlock(val);
+}
+
+/*
+ * Cruft below, do not use these any more.
+ */
+int
+ipmi_posix_thread_sel_select(os_handler_t   *os_hnd,
+			     struct timeval *timeout)
+{
+    return perform_one_op(os_hnd, timeout);
+}
+
+void
+ipmi_posix_thread_sel_select_loop(os_handler_t *os_hnd)
+{
+    operation_loop(os_hnd);
+}
+
+void
+ipmi_posix_thread_cleanup_os_handler(os_handler_t *os_hnd)
+{
+    free_os_handler(os_hnd);
 }
